.ad l
.nh
.TH combine 1 "10 January 1993" "ImageMagick"
.SH NAME
combine - combine images to create new images.
.SH SYNOPSIS
.B "combine"
[ \fIoptions\fP ... ] \fIimage\fP \fIcomposite\fP [ \fImask\fP ] \fIcombined\fP
.SH DESCRIPTION
\fBcombine\fP combine images to create new images.
.SH EXAMPLES
.PP
To combine a image of a cockatoo with a perch, use
.PP
.B
     combine cockatoo.miff perch.ras composite.miff
.PP
To compute the difference between images in a series, use
.PP
.B
     combine -compose difference series.1 series.2 difference.miff
.PP
To combine a image of a cockatoo with a perch starting at location (100,150),
use
.PP
.B
     combine -geometry +100+150 cockatoo.miff perch.ras composite.miff
.PP
To tile a logo across your image of a cockatoo, use
.PP
.B
     convert +shade 30x60 cockatoo.miff mask.miff
     combine -compose bumpmap -tile logo.gif cockatoo.miff mask.miff composite.miff
.SH OPTIONS
.TP
.B "-blend \fIvalue\fP"
blend the two images a given percent.
.TP
.B "-colors \fIvalue\fP"
preferred number of colors in the image.

The actual number of colors in the image may be less than your request,
but never more.  Note, this is a color reduction option.  Images with
less unique colors than specified with this option will remain unchanged.
Refer to \fBquantize(9)\fP for more details.

Note, options \fB-dither\fP, \fB-colorspace\fP, and \fB-treedepth\fP affect
the color reduction algorithm.
.TP
.B "-colorspace \fIvalue\fP"
the type of colorspace: \fBGRAY\fP, \fBOHTA\fP, \fBRGB\fP,
\fBTransparent\fP, \fBXYZ\fP, \fBYCbCr\fP, \fBYIQ\fP, \fBYPbPr\fP, or
\fBYUV\fP.

Color reduction, by default, takes place in the RGB color space.
Empirical evidence suggests that distances in color spaces such as YUV
or YIQ correspond to perceptual color differences more closely
than do distances in RGB space.  These color spaces may give better
results when color reducing an image.  Refer to \fBquantize(9)\fP for
more details.
 
The \fBTransparent\fP color space behaves uniquely in that it preserves
the matte channel of the image if it exists.

The \fB-colors\fP or \fB-monochrome\fP option is required for this option
to take effect.
.TP
.B "-comment \fIstring\fP"
annotate an image with a comment.

By default, each image is commented with its file name.  Use this
option to assign a specific comment to the image.  Optionally you can
include the image filename, type, width, height, or scene number by
embedding special format characters.  Embed \fB%f\fP for filename,
\fB%d\fP for directory, \fB%e\fP for filename extention, \fB%t\fP for
top of filename, \fB%m\fP for magick, \fB%w\fP for width, \fB%h\fP for
height, \fB%s\fP for scene number, \fB%b\fP for file size, or \fB\\n\fP
for newline.  For example,

.nf
     -comment "%m:%f %wx%h"
.fi

produces an image comment of \fBMIFF:bird.miff 512x480\fP for an image
titled \fBbird.miff\fP and whose width is 512 and height is 480.

If the first character of \fIstring\fP is \fB@\fP, the image comment is read
from a file titled by the remaining characters in the string.
.TP
.B "-compose \fIoperator\fP"
the type of image composition.

By default, each of the composite image pixels are replaced by the
cooresponding image tile pixel. You can choose an alternate composite
operation:

    over
    in
    out
    atop
    xor
    plus
    minus
    add
    subtract
    difference
    bumpmap
    replace
.PP
How each operator behaves is described below.
.TP 9
.B over
The result will be the union of the two image shapes, with
\fIcomposite image\fP obscuring \fIimage\fP in the region of overlap.
.TP 9
.B in
The result is simply \fIcomposite image\fP cut by the shape of \fIimage\fP.
None of the image data of \fIimage\fP will be in the result.
.TP 9
.B out
The resulting image is \fIcomposite image\fP with the shape of \fIimage\fP cut
out.
.TP 9
.B atop
The result is the same shape as image \fIimage\fP, with \fIcomposite image\fP
obscuring \fIimage\fP where the image shapes overlap.  Note this differs from
\fBover\fP because the portion of \fIcomposite image\fP outside \fIimage\fP's
shape does not appear in the result.
.TP 9
.B xor
The result is the image data from both \fIcomposite image\fP and \fIimage\fP
that is outside the overlap region.  The overlap region will be blank.
.TP 9
.B plus
The result is just the sum of the image data.  Output values are
cropped to 255 (no overflow).  This operation is independent of the
matte channels.
.TP 9
.B minus
The result of \fIcomposite image\fP \- \fIimage\fP, with underflow cropped to
zero.  The matte channel is ignored (set to 255, full coverage).
.TP 9
.B add
The result of \fIcomposite image\fP + \fIimage\fP, with overflow wrapping
around (\fImod\fP 256).
.TP 9
.B subtract
The result of \fIcomposite image\fP - \fIimage\fP, with underflow wrapping
around (\fImod\fP 256).  The \fBadd\fP and \fBsubtract\fP operators can be
used to perform reversible transformations.
.TP 9
.B difference
The result of abs(\fIcomposite image\fP \- \fIimage\fP).  This is useful for
comparing two very similar images.
.TP 9
.B bumpmap
The result \fIimage\fP shaded by \fIcomposite image\fP.
.TP 9
.B replace
The resulting image is \fIimage\fP replaced with \fIcomposite image\fP.  Here
the matte information is ignored.
.TP
The image compositor requires a matte, or alpha channel in the image
for some operations.  This extra channel usually defines a mask which
represents a sort of a cookie-cutter for the image.  This is the case
when matte is 255 (full coverage) for pixels inside the shape, zero
outside, and between zero and 255 on the boundary.  For certain
operations, if \fIimage\fP does not have an matte channel, it is initialized
with 0 for any pixel matching in color to pixel location (0,0), otherwise
255 (to work properly \fBborderwidth\fP must be 0).
.TP
.B "-compress \fItype\fP"
the type of image compression: \fINo\fP, \fIRunlengthEncoded\fP, or \fIZip\fP.

Specify \fB\+compress\fP to store the binary image in an uncompressed format.
The default is the compression type of the specified image file.
.TP
.B "-density \fI<width>x<height>\fP
vertical and horizontal resolution in pixels of the image.

This option specifies an image density when decoding a Postscript or Portable
Document page.  The default is 72 pixels per inch in the horizontal and
vertical direction.
.TP
.B "-displace \fI<horizontal scale>x<vertical scale>\fP"
shift image pixels as defined by a displacement map.
.PP
With this option, \fIcomposite image\fP is used as a displacement map.  Black,
within the displacement map, is a maximum positive displacement.  White is a
maximum negative displacement and middle gray is neutral.  The displacement
is scaled to determine the pixel shift.  By default, the displacement applies
in both the horizontal and vertical directions.  However, if you specify
\fImask\fP, \fIcomposite image\fP is the horizontal X displacement and
\fImask\fP the vertical Y displacement.
.TP
.B "-display \fIhost:display[.screen]\fP"
specifies the X server to contact; see \fBX(1)\fP.
.TP
.B "-dispose \fImethod\fP"
GIF disposal method.

Here are the valid methods:

.nf
  0     No disposal specified.
  1     Do not dispose.      
  2     Restore to background color.
  3     Restore to previous.
.fi
.TP
.B "-dither"
apply Floyd/Steinberg error diffusion to the image.

The basic strategy of dithering is to trade intensity resolution for
spatial resolution by averaging the intensities of several neighboring
pixels.  Images which suffer from severe contouring when reducing colors
can be improved with this option.

The \fB-colors\fP option is required for dithering to take effect.
.TP
.B "-font \fIname\fP"
This option specifies the font to be used  for displaying normal text.
The default is \fBfixed\fP.
.TP
.B "-geometry \fI<width>{%}x<height>{%}{\+-}<x offset>{\+-}<y offset>{!}{<}{>}\fP"
the width and height of the image.

By default, the width and height are maximum values.  That is, the
image is expanded or contracted to fit the width and height value while
maintaining the aspect ratio of the image.  Append an exclamation point
to the geometry to force the image size to exactly the size you
specify.  For example, if you specify \fB640x480!\fP the image width is
set to 640 pixels and height to 480.  If only one factor is
specified, both the width and height assume the value.

To specify a percentage width or height instead, append \fB%\fP.  The
image size is multiplied by the width and height percentages to obtain
the final image dimensions.  To increase the size of an image, use a
value greater than 100 (e.g. 125%).  To decrease an image's size, use a
percentage less than 100.

Use \fB>\fP to change the dimensions of the image \fIonly\fP
if its size exceeds the geometry specification.  \fB>\fP resizes
the image \fIonly\fP if its dimensions is less than the geometry
specification.  For example, if you specify \fB640x480>\fP and the
image size is 512x512, the image size does not change.  However, if
the image is 1024x1024, it is resized to 640x480.

By default the images are combined relative to the image gravity
(see \fB-gravity\fP).  Use \fI<x offset>\fP and
\fI<y offset>\fP to specify a particular location to combine the images.
.TP
.B "-gravity \fIdirection\fP"
direction image gravitates to within the composite.  See \fBX(1)\fP for
details about the gravity specification.

The image may not fill the composite completely (see
\fB-geometry\fP).  The direction you choose specifies where to
position the image within the composite.  For example \fICenter\fP gravity
forces the image to be centered within the composite.  A gravity of
\fIForget\fP stretches the composite to the same size as the image.
By default, the image gravity is \fINorthWest\fP.
.TP
.B "-interlace \fItype\fP"
the type of interlacing scheme: \fBNo\fP, \fBLine\fP, \fBPlane\fP, or
\fBPartition\fP.  The default is \fBPlane\fP.

This option is used to specify the type of interlacing scheme for raw
image formats such as \fBRGB\fP or \fBYUV\fP.  \fBNo\fP means do not
interlace (RGBRGBRGBRGBRGBRGB...), \fBLine\fP uses scanline
interlacing (RRR...GGG...BBB...RRR...GGG...BBB...), and \fBPlane\fP uses
plane interlacing (RRRRRR...GGGGGG...BBBBBB...).  \fBPartition\fP is like
plane except the different planes are saved to individual files (e.g.
image.R, image.G, and image.B).

Use \fBLine\fP, or \fBPlane\fP to create an interlaced GIF or progressive
JPEG image.
.TP
.B "-label \fIname\fP"
assign a label to an image.

Use this option to assign a specific label to the image.  Optionally
you can include the image filename, type, width, height, or scene
number in the label by embedding special format characters.   Embed
\fB%f\fP for filename, \fB%d\fP for directory, \fB%e\fP for filename
extention, \fB%t\fP for top of filename, \fB%m\fP for magick, \fB%w\fP
for width, \fB%h\fP for height, \fB%s\fP for scene number, \fB%b\fP for
file size in kilobytes, or \fB\\n\fP for newline.  For example,
.nf
     -label "%m:%f %wx%h"
.fi
produces an image label of \fBMIFF:bird.miff 512x480\fP for an image
titled \fBbird.miff\fP and whose width is 512 and height is 480.

If the first character of \fIstring\fP is \fB@\fP, the image label is read
from a file titled by the remaining characters in the string.

When converting to Postscript, use this option to specify a header string
to print above the image.
.TP
.B "-matte"
store matte channel if the image has one.
.TP
.B "-monochrome"
transform the image to black and white.
.TP
.B "-negate"
apply color inversion to image.
 
The red, green, and blue intensities of an image are negated.  Use
\fB+negate\fP to only negate the grayscale pixels of the image.
.TP
.B "-page \fI<width>x<height>{\+-}<x offset>{\+-}<y offset>\fP"
preferred size and location of the Postscript page.

Use this option to specify the dimensions of the Postscript page in
pixels per inch or a TEXT page in pixels.  The default for a Postscript page is
to center the image on a letter page 612 by 792 pixels. The
margins are 1/2" (i.e.  612x792+42+42).  Other common sizes are:

    Letter      612x 792
    Tabloid     792x1224
    Ledger     1224x 792
    Legal       612x1008
    Statement   396x 612
    Executive   540x 720
    A3          842x1190
    A4          595x 842
    A5          420x 595
    B4          729x1032
    B5          516x 729
    Folio       612x 936
    Quarto      610x 780
    10x14       720x1008
.PP
For convenience you can specify the page size by media (e.g.
A4, Ledger, etc.).
.PP
To place a Postscript image with a given size on a given location on a
page, use -page +HOFFSET+VOFFSET -geometry WIDTHxHEIGHT (fill in
numbers). Note: this is only for generating Postscript, not Encapsulated
Postscript.
.PP
To position a GIF image, use -page +LEFT+TOP (e.g. -page +100+200).
.PP
The default page dimensions for a TEXT image is 612x792.
.TP
.B "-quality \fIvalue\fP"
JPEG quality setting.

Quality is 0 (worst) to 100 (best). The default is 75.
.TP
.B "-scene \fIvalue\fP"
image scene number.
.TP
.B "-size \fI<width>{%}x<height>{%}{+offset}{!}"
width and height of the image.

Use this option to specify the width and height of raw images whose
dimensions are unknown such as \fBGRAY\fP, \fBRGB\fP, or \fBCMYK\fP.
In addition to width and height, use \fB-size\fP to skip any header
information in the image or  tell the number of colors in a \fBMAP\fP
image file, (e.g. -size 640x512+256).
.TP
.B "-stereo"
combine two images into a red-green stereo image.

The left side of the stereo pair is saved as the red channel of the output
image.  The right sife is saved as the green channel.  Red-blue stereo
glasses are required to properly view the stereo image.
.TP
.B "-tile"
repeat composite operation across image.
.TP
.B "-treedepth \fIvalue\fP"
Normally, this integer value is zero or one.  A zero or one tells
\fBcombine\fP to choose a optimal tree depth for the color reduction
algorithm.

An optimal depth generally allows the best representation of the source
image with the fastest computational speed and the least amount of
memory.  However, the default depth is inappropriate for some images.
To assure the best representation, try values between 2 and 8 for this
parameter.  Refer to \fBquantize(9)\fP for more details.

The \fB-colors\fP option is required for this option to take effect.
.TP
.B -verbose
print detailed information about the image.

This information is printed: image scene number;  image name;  combined
image name;  image size;  the image class (\fIDirectClass\fP or
\fIPseudoClass\fP);  the total number of unique colors;  and the number
of seconds to read and combine the image.
.PP
Options are processed in command line order.
Any option you specify on the command line remains in effect until it is
explicitly changed by specifying the option again with a different effect.
.PP
Change '-' to '+' in any option above to reverse its effect.  For
example, specify +matte to store the image without its matte channel.
.PP
By default, the image format is determined by its magic number. To
specify a particular image format, precede the filename with an image
format name and a colon (i.e. ps:image) or specify the image type as
the filename suffix (i.e. image.ps).  See \fBconvert(1)\fP for a list
of valid image formats.
.PP
When you specify \fBX\fP as your image type, the filename has special
meaning.  It specifies an X window by id, name, or \fBroot\fP.  If no
filename is specified, the window is selected by clicking the mouse in
the desired window.
.PP
Specify \fIimage\fP as \fI-\fP for standard input,
\fIcombined\fP as \fI-\fP for standard output.  If \fIimage\fP
has the extension \fB.Z\fP or \fB.gz\fP, the file is uncompressed with
\fBuncompress\fP or \fBgunzip\fP respectively.  If \fIcombined\fP
has the extension \fB.Z\fP or \fB.gz\fP, the file size is compressed
using with \fBcompress\fP or \fBgzip\fP respectively.  Finally, precede
the image file name with \fI|\fP to pipe to or from a system command.
.PP
Use an optional index enclosed in brackets after a file name to specify
a desired subimage of a multi-resolution image format like Photo CD
(e.g. img0001.pcd[4]) or a range for MPEG images (e.g. video.mpg[50-75]).
A subimage specification can be disjoint (e.g. image.tiff[2,7,4]).
For raw images, specify a subimage with a geometry (e.g.  -size 640x512
image.rgb[320x256+50+50]).
.PP
The optional \fBmask\fP can be used to provide matte information for
\fBcomposite\fP when it has none or if you want a different mask.
A mask image is typically grayscale and the same size as
\fBcomposite\fP.  If the image is not grayscale, it is converted to
grayscale and the resulting intensities are used as matte information.
.PP
If \fIcombined\fP already exists, you will be prompted as to whether
it should be overwritten.
.SH ENVIRONMENT
.PP
.TP
.B display
To get the default host, display number, and screen.
.SH SEE ALSO
.B
display(1), animate(1), import(1), montage(1), mogrify(1), convert(1),
xtp(1)
.SH COPYRIGHT
Copyright 1997 E. I. du Pont de Nemours and Company
.PP
Permission to use, copy, modify, distribute, and sell this software and
its documentation for any purpose is hereby granted without fee,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of E. I. du Pont de Nemours
and Company not be used in advertising or publicity pertaining to
distribution of the software without specific, written prior
permission.  E. I. du Pont de Nemours and Company makes no representations
about the suitability of this software for any purpose.  It is provided
"as is" without express or implied warranty.
.PP
E. I. du Pont de Nemours and Company disclaims all warranties with regard
to this software, including all implied warranties of merchantability
and fitness, in no event shall E. I. du Pont de Nemours and Company be
liable for any special, indirect or consequential damages or any
damages whatsoever resulting from loss of use, data or profits, whether
in an action of contract, negligence or other tortious action, arising
out of or in connection with the use or performance of this software.
.SH AUTHORS
John Cristy, E.I. du Pont De Nemours and Company Incorporated
