/*
 *  WindowMaker miscelaneous function library
 * 
 *  Copyright (c) 1997 Alfredo K. Kojima
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include "wwmlib.h"
#include <stdlib.h>
#include <sys/types.h>
#include <unistd.h>
#include <stdio.h>

#ifdef WMAKER
extern void wAbort();
#endif

static int Aborting=0; /* if we're in the middle of an emergency exit */

#ifdef DEBUG
static HashTable Blocks;
static int TotalUsed=0;
static int *Size;
static int Inited=0;

static void initdebug()
{
    table_init(&Blocks);
}
#endif


void *xmalloc(size_t size)
{
    void *tmp;
    
    tmp = malloc(size);
    if (tmp == NULL) {
	wWarning("malloc() failed. Retrying after 2s.");
	sleep(2);
	tmp = malloc(size);
	if (tmp == NULL) {
	    if (Aborting) {
		puts("Real Bad Error: recursive malloc() failure.");
		exit(-1);
	    } else {
		wFatal("virtual memory exhausted");
#ifdef WMAKER
		wAbort();
#endif
		Aborting=1;
	    }
	}
    }
    return tmp;
}


#ifdef DEBUG
void *wmalloc(size_t size)
{
    void *tmp;
    
    if (!Inited) {
	initdebug();
	Inited=1;
    }
    tmp = malloc(size);
    if (tmp == NULL) {
	wWarning("malloc() failed. Retrying after 1s.");
	/* retry after some time just to see what happens */
	sleep(1);
	tmp = malloc(size);
	if (tmp == NULL) {
	    wFatal("virtual memory exhausted");
	    wAbort();
	}
    }
    Size = malloc(sizeof(int));
    *Size = size;
    TotalUsed += size;
    table_iput(&Blocks, (int)tmp, Size);
    return tmp;
}
#endif

void *xrealloc(void *ptr, size_t newsize)
{
    void *nptr;

    nptr=realloc(ptr, newsize);
    if (nptr==NULL) {
	printf("Could not do realloc");
	return NULL;
    }
    return nptr;
}

#ifdef DEBUG
void *wrealloc(void *ptr, size_t newsize)
{
    void *nptr;

    if (!Inited) {
	initdebug();
	Inited=1;
    }
    nptr=realloc(ptr, newsize);
    if (nptr==NULL) {
	printf("Could not do realloc");
	return NULL;
    }
    if (ptr==NULL) {
	Size = malloc(sizeof(int));
	*Size = newsize;
	TotalUsed += newsize;
	table_iput(&Blocks, (int)nptr, Size);
    } else {
	Size = table_iget(&Blocks, (int)ptr);    
	if (Size==NULL) {
	    puts("error in memory debugging code");
	} else {
	    table_idelete(&Blocks, (int)ptr);
	    TotalUsed -= *Size;
	    *Size = newsize;
	    TotalUsed += newsize;
	    table_iput(&Blocks, (int)nptr, Size);
	}
    }
    return nptr;
}
#endif


#ifdef DEBUG
void wfree(void *ptr)
{
    Size = table_iget(&Blocks, (int)ptr);
    if (Size==NULL) {
	puts("error in memory debugging code for free()");
    } else {
	table_idelete(&Blocks, (int)ptr);
	TotalUsed -= *Size;
    }
    free(ptr);
}

void pusage()
{
    printf("USED MEMORY %i (%i k)\n",TotalUsed,TotalUsed/1024);
    printf("blocks allocated %i\n",Blocks.elements);
}
#endif
