/*
 *  WindowMaker window manager
 * 
 *  Copyright (c) 1997 Alfredo K. Kojima
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the wfree Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the wfree Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "wconfig.h"

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "generic/wwmlib.h"
#include "WindowMaker.h"
#include "raster.h"
#include "screen.h"
#include "texture.h"



static void
bevelImage(unsigned char *image, int width, int height, int relief)
{
    int x, y, ofs, i;
    unsigned char *data;

    data = image;
    switch (relief) {
     case WREL_MENUENTRY:
	data += 3;
	for (x=0; x<(width-2)*3; x++) {
	    *data = (*data<175 ? *data+80 : 255);
	    data++;
	}
	data = image;
	for (y=0; y<height; y++) {
	    ofs = y*width*3;
	    data[ofs++]>>=1;
	    data[ofs++]>>=1;
	    data[ofs++]>>=1;
	    data[ofs] = (data[ofs] <175 ? data[ofs]+80 : 255);
	    data[ofs+1] = (data[ofs+1] <175 ? data[ofs+1]+80 : 255);
	    data[ofs+2] = (data[ofs+2] <175 ? data[ofs+2]+80 : 255);
	    ofs = y*width*3+(width-2)*3;
	    data[ofs] = (data[ofs] >40 ? data[ofs]-40 : 0);
	    data[ofs+1] = (data[ofs+1] >40 ? data[ofs+1]-40 : 0);
	    data[ofs+2] = (data[ofs+2] >40 ? data[ofs+2]-40 : 0);
	    ofs+=3;
	    data[ofs++]=0;
	    data[ofs++]=0;
	    data[ofs++]=0;
	}
	data = image + (height-2)*width*3 + 3;
	for (x=3; x<width*3; x++) {
	    *data = (*data>40 ? *data-40 : 0);
	    data++;
	}
	memset(data, 0, width*3);
	break;

     case WREL_MENUTITLE:
	for (x=0; x<(width-1)*3; x++) {
	    *data = (*data>40 ? *data-40 : 0);
	    data++;
	}
	*(data++)=0;
	*(data++)=0;
	*(data++)=0;
	data = image;
	data += width*3+3;
	for (x=0; x<(width-2)*3; x++) {
	    *data = (*data<175 ? *data+80 : 255);
	    data++;
	}
	data = image;
	for (y=1; y<height-1; y++) {
	    ofs = y*width*3;
	    for (i=0; i<3; i++, ofs++) {
		data[ofs] = (data[ofs] >50 ? data[ofs]-50 : 0);
		data[ofs+3] = (data[ofs+3] <175 ? data[ofs+3]+80 : 255);
	    }
	    ofs = y*width*3+(width-2)*3;

	    data[ofs++]>>=1;
	    data[ofs++]>>=1;
	    data[ofs++]>>=1;
	    data[ofs++]=0;
	    data[ofs++]=0;
	    data[ofs++]=0;
	}
	data = image + (height-2)*width*3;
	for (x=0; x<(width-1)*3; x++) {
	    *data = (*data>40 ? *data-40 : 0);
	    data++;
	}
	data+=3;
	memset(data, 0, width*3);
	break;

     default:
	for (x=0; x<width*3; x++) {
	    *data = (*data<175 ? *data+80 : 255);
	    data++;
	}
	data = image;
	for (y=1; y<height-2; y++) {
	    ofs = y*width*3;
	    data[ofs] = (data[ofs] <175 ? data[ofs]+80 : 255);
	    data[ofs+1] = (data[ofs+1] <175 ? data[ofs+1]+80 : 255);
	    data[ofs+2] = (data[ofs+2] <175 ? data[ofs+2]+80 : 255);
	    ofs = y*width*3+(width-1)*3;
	    data[ofs] = (data[ofs] >40 ? data[ofs]-40 : 0);
	    data[ofs+1] = (data[ofs+1] >40 ? data[ofs+1]-40 : 0);
	    data[ofs+2] = (data[ofs+2] >40 ? data[ofs+2]-40 : 0);
	}
	data = image + (height-2)*width*3;
	for (x=0; x<width*3; x++) {
	    *data = (*data>40 ? *data-40 : 0);
	    data++;
	}
	memset(data, 0, width*3);
	break;
    }
}



/*
 *----------------------------------------------------------------------
 * RenderHGradient--
 * 	Renders a horizontal linear gradient of the specified size in the
 * WImage format with a border of the specified type. 
 * It must then be dithered and converted to pixmap to be displayed.
 * 
 * Returns:
 * 	A 24bit WImage with the gradient (no alpha channel).
 * 
 * Side effects:
 * 	None
 *---------------------------------------------------------------------- 
 */
WImage *
RenderHGradient(int width, int height, XColor from, XColor to, int relief)
{
    int x, y;
    float r, g, b, dr, dg, db;
    WImage *image;
    unsigned char *data;

    image = wmalloc(sizeof(WImage));
    data = wmalloc(width*height*3);
    image->data = data;
    image->width = width;
    image->height = height;
    image->alpha = NULL;
    r = (float)(from.red>>8);
    g = (float)(from.green>>8);
    b = (float)(from.blue>>8);
    
    dr = (float)((to.red-from.red)>>8)/(float)width;
    dg = (float)((to.green-from.green)>>8)/(float)width;
    db = (float)((to.blue-from.blue)>>8)/(float)width;
    /* render the first line */
    for (x=0; x<width; x++) {
	*(data++) = (unsigned char)r;
	*(data++) = (unsigned char)g;
	*(data++) = (unsigned char)b;
	r+=dr;
	g+=dg;
	b+=db;
    }
    /* rewind pointer */
    data = image->data;
    /* copy the first line to the other lines */
    for (y=1; y<height; y++) {
	memcpy(&(data[y*width*3]), data, width*3);
    }
    /* draw bevel */
    bevelImage(data, width, height, relief);

    return image;
}

/*
 *----------------------------------------------------------------------
 * RenderVGradient--
 *      Renders a vertical linear gradient of the specified size in the
 * WImage format with a border of the specified type.
 * It must then be dithered and converted to pixmap to be displayed.
 *
 * Returns:
 *      A 24bit WImage with the gradient (no alpha channel).
 *
 * Side effects:
 *      None
 *----------------------------------------------------------------------
 */
WImage *
RenderVGradient(int width, int height, XColor from, XColor to, int relief)
{
    int x, y;
    float r, g, b, dr, dg, db;
    WImage *image;
    unsigned char *data;

    image = wmalloc(sizeof(WImage));
    data = wmalloc(width*height*3);
    image->data = data;
    image->width = width;
    image->height = height;
    image->alpha = NULL;
    r = (float)(from.red>>8);
    g = (float)(from.green>>8);
    b = (float)(from.blue>>8);

    dr = (float)((to.red-from.red)>>8)/(float)height;
    dg = (float)((to.green-from.green)>>8)/(float)height;
    db = (float)((to.blue-from.blue)>>8)/(float)height;

    for (y=0; y<height; y++) {
        for (x=0; x<width; x++) {
            *(data++) = (unsigned char)r;
            *(data++) = (unsigned char)g;
            *(data++) = (unsigned char)b;
        }
        r+=dr;
        g+=dg;
        b+=db;
    }

    /* rewind pointer */
    data = image->data;
    /* draw bevel */
    bevelImage(data, width, height, relief);

    return image;
}


/*
 *----------------------------------------------------------------------
 * RenderDGradient--
 *      Renders a diagonal linear gradient of the specified size in the
 * WImage format with a border of the specified type.
 * It must then be dithered and converted to pixmap to be displayed.
 *
 * Returns:
 *      A 24bit WImage with the gradient (no alpha channel).
 *
 * Side effects:
 *      None
 *----------------------------------------------------------------------
 */
WImage *
RenderDGradient(int width, int height, XColor from, XColor to, int relief)
{
    int x, y;
    float from_red,from_green,from_blue;
    float to_red,to_green,to_blue;
    float t_r,t_g,t_b,d_r,d_g,d_b,r,g,b,dr,dg,db;

    WImage *image;
    unsigned char *data;

    image = wmalloc(sizeof(WImage));
    data = wmalloc(width*height*3);
    image->data = data;
    image->width = width;
    image->height = height;
    image->alpha = NULL;

    from_red = (float)(from.red>>8);
    from_green = (float)(from.green>>8);
    from_blue = (float)(from.blue>>8);

    to_red = (float)(to.red>>8);
    to_green = (float)(to.green>>8);
    to_blue = (float)(to.blue>>8);

    t_r = (from_red+to_red) / (float)2;
    t_g = (from_green+to_green) / (float)2;
    t_b = (from_blue+to_blue) / (float)2;

    d_r = (float)(height*2) / (float)(from_red+to_red);
    d_g = (float)(height*2) / (float)(from_green+to_green);
    d_b = (float)(height*2) / (float)(from_blue+to_blue);

    for (y=0; y<height; y++) {

        r = from_red;
        g = from_green;
        b = from_blue;

        dr = (t_r-r) / (float)width;
        dg = (t_g-g) / (float)width;
        db = (t_b-b) / (float)width;

        for (x=0; x<width; x++) {
            *(data++) = (unsigned char)r;
            *(data++) = (unsigned char)g;
            *(data++) = (unsigned char)b;
            r+=dr;
            g+=dg;
            b+=db;
        }
        from_red+=d_r;
        from_green+=d_g;
        from_blue+=d_b;
        t_r+=d_r;
        t_g+=d_g;
        t_b+=d_b;
    }

    /* rewind pointer */
    data = image->data;
    /* draw bevel */
    bevelImage(data, width, height, relief);

    return image;
}

