


#include <X11/Xlib.h>
#include <X11/Xproto.h>
#include "../src/GNUstep.h"

static unsigned char bits[] = {
   0x00, 0xfc, 0x00, 0xfe, 0x00, 0xff, 0x80, 0xff, 0xc0, 0xff, 0xe0, 0xff,
   0xf0, 0xff, 0xf8, 0xff, 0xfc, 0xff, 0xfe, 0xff};

static unsigned char mbits[] = {
   0xff, 0x03, 0xff, 0x01, 0xff, 0x00, 0x7f, 0x00, 0x3f, 0x00, 0x1f, 0x00,
   0x0f, 0x00, 0x07, 0x00, 0x03, 0x00, 0x01, 0x00};


Atom _XA_GNUSTEP_WM_ATTR;

void
PropWriteGNUstepWMAttr(Display *dpy, Window window, GNUstepWMAttributes *attr)
{
    attr->reserved=0;
    XChangeProperty(dpy, window, _XA_GNUSTEP_WM_ATTR, _XA_GNUSTEP_WM_ATTR, 
		    32, PropModeReplace,  (unsigned char *)attr, 
		    sizeof(GNUstepWMAttributes)/sizeof(CARD32));
}


int main()
{
    Display *dpy;
    Window win;
    Atom delete_win, miniaturize_win;
    Atom resize_win, resized_win;
    Atom prots[6];
    GNUstepWMAttributes attr;

    
    dpy = XOpenDisplay("");
    if (!dpy) {
	puts("could not open display!");
	exit(1);
    }
    _XA_GNUSTEP_WM_ATTR = XInternAtom(dpy, GNUSTEP_WM_ATTR_NAME, False);
    delete_win = XInternAtom(dpy, "WM_DELETE_WINDOW", False);
    miniaturize_win = XInternAtom(dpy, "_GNUSTEP_WM_MINIATURIZE_WINDOW", 
				  False);
    resize_win = XInternAtom(dpy, "_GNUSTEP_WM_RESIZE_WINDOW", 
				  False);
    resized_win = XInternAtom(dpy, "_GNUSTEP_WM_RESIZED_WINDOW", 
				  False);
    win = XCreateSimpleWindow(dpy, DefaultRootWindow(dpy), 10, 10, 100, 100,
			      0, 0, 0);
    prots[0] = delete_win;
    prots[1] = miniaturize_win;
    prots[2] = resize_win;
    XSetWMProtocols(dpy, win, prots, 3);
    XStoreName(dpy, win, "Test Window");
    attr.flags = GSMiniaturizePixmapAttr|GSMiniaturizeMaskAttr;
    
    attr.miniaturize_pixmap = 
      XCreateBitmapFromData(dpy, DefaultRootWindow(dpy), bits, 10, 10);
    
    attr.miniaturize_mask =
      XCreateBitmapFromData(dpy, DefaultRootWindow(dpy), mbits, 10, 10);
    PropWriteGNUstepWMAttr(dpy, win, &attr);

    XMapRaised(dpy, win);
    XFlush(dpy);

    while (1) {
	XEvent ev;
	int w, h;
	XNextEvent(dpy, &ev);
	if (ev.type==ClientMessage) {
	    if (ev.xclient.data.l[0]==delete_win) {
		puts("You've pushed the close window button");
	    } else if (ev.xclient.data.l[0]==miniaturize_win) {
		puts("You've pushed the maximize window button");
	    } else if (ev.xclient.data.l[0]==resize_win) {
		w = ev.xclient.data.l[2];
		h = ev.xclient.data.l[3];
		if (w<100 || h<60) {
		    printf("%i x %i is too small\n",w,h);
		} else if (w>300 || h > 200) {
		    printf("%i x %i is too big\n",w,h);
		} else {
		    printf("%i x %i is ok\n",w,h);
		}
		
	    } else if (ev.xclient.data.l[0]==resized_win) {
		w = ev.xclient.data.l[2];
		h = ev.xclient.data.l[3];
		if (w<100 || h<60) {
		    printf("%i x %i is too small\n",w,h);
		} else if (w>300 || h > 200) {
		    printf("%i x %i is too big\n",w,h);
		} else {
		    printf("%i x %i is ok\n",w,h);
		}
		if (w>100 && h>60 && w<300 && h < 200) {
		    XResizeWindow(dpy, win, w, h);
		}
	    }
	}
    }
}

