/*
 * wconfig.h- default configuration and definitions 
 * 
 *  WindowMaker window manager
 * 
 *  Copyright (c) 1997 Alfredo K. Kojima
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef WMCONFIG_H_
#define WMCONFIG_H_

#include "config.h"

/*
 *--------------------------------------------------------------------
 * 			Feature Selection
 * 
 * 	Comment out the following #defines if you want to 
 * disable a feature. 
 * 	Also check the features you can enable through configure.
 *--------------------------------------------------------------------
 */

/* undefine ANIMATIONS if you don't want animations for iconification,
 * shading, icon arrangement etc. */
#define ANIMATIONS

/* define NEWSTYLE if you want a different style of titlebar buttons.
 */
#undef NEWSTYLE

/*
 * undefine USECPP if you don't want your config files to be preprocessed
 * by cpp
 */
#define USECPP

/*
 * I18N support
 */
#undef I18N


/*
 *--------------------------------------------------------------------
 * 			Default Configuration
 * 
 * 	Some of the following options can be configured in
 * the preference files, but if for some reason, they can't
 * be used, these defaults will be.
 * 
 *--------------------------------------------------------------------
 */

/* list of paths to look for the config files, searched in order
 * of appearance */
#define DEF_CONFIG_PATHS \
"~/gnustep/Library/WindowMaker/:/usr/GNUstep/Library/WindowMaker/:/usr/local/GNUstep/Library/WindowMaker/"

#define DEF_USERPREF_PATH	"~/gnustep/Library/WindowMaker/"

/*
#define DEF_CONFIG_PATHS \
"~/gnustep/Library/WindowMaker/:/usr/X11R6/lib/X11/WindowMaker/sample.WindowMaker/"
*/

/* root menu definition */
#define DEF_MENUCONFIG		"menu"
/* preference file */
#define DEF_PREFERENCES		"preferences"

#define DEF_DOCKCONFIG		"dock"

/* script file to exec on startup */

/* pixmap path */
#define DEF_PIXMAP_PATHS \
"~/gnustep/Library/WindowMaker/Pixmaps/:/usr/X11R6/lib/X11/WindowMaker/pixmaps/:/usr/local/GNUstep/Library/WindowMaker/Pixmaps/"

/* window title to use for untitled windows */
#define DEF_WINDOW_TITLE "Untitled"

/* default style */
#define DEF_FRAME_COLOR "white"
#define DEF_FOCUSED_TEXTURE_COLOR "black"
#define DEF_PREV_FOCUSED_TEXTURE_COLOR "#616161"
#define DEF_UNFOCUSED_TEXTURE_COLOR "#b6b6b6"
#define DEF_FOCUSED_TEXT_COLOR "white"
#define DEF_UNFOCUSED_TEXT_COLOR "black"
#define DEF_ICON_TEXTURE_COLOR "#b6b6b6"

#define DEF_FOCUSED_TEXT_COLOR "white"
#define DEF_UNFOCUSED_TEXT_COLOR "black"
#define DEF_DISABLED_TEXT_COLOR "#777777"

#ifdef I18N
#define DEF_TITLE_FONT "-misc-fixed-*-*-*-*-*-130-75-75-*-*-*-*"
#define DEF_MENU_TITLE_FONT "-misc-fixed-*-*-*-*-*-130-75-75-*-*-*-*"
#define DEF_MENU_ENTRY_FONT "-misc-fixed-*-*-*-*-*-130-75-75-*-*-*-*"
#define DEF_ICON_TITLE_FONT "-misc-fixed-*-*-*-*-*-130-75-75-*-*-*-*"
#define DEF_PANEL_TITLE_FONT "-misc-fixed-*-*-*-*-*-130-75-75-*-*-*-*"
#define DEF_PANEL_TEXT_FONT "-misc-fixed-*-*-*-*-*-130-75-75-*-*-*-*"
#define DEF_INFO_TEXT_FONT "-misc-fixed-*-*-*-*-*-130-75-75-*-*-*-*"
#else /* I18N */
#define DEF_TITLE_FONT "-*-helvetica-bold-r-*-*-12-*-*-*-*-*-*-*"
#define DEF_MENU_TITLE_FONT "-*-helvetica-bold-r-*-*-12-*-*-*-*-*-*-*"
#define DEF_MENU_ENTRY_FONT "-*-helvetica-medium-r-*-*-12-*-*-*-*-*-*-*"
#define DEF_ICON_TITLE_FONT "-*-helvetica-medium-r-*-*-8-*-*-*-*-*-*-*"
#define DEF_PANEL_TITLE_FONT "-*-helvetica-medium-r-*-*-20-*-*-*-*-*-*-*"
#define DEF_PANEL_TEXT_FONT "-*-helvetica-medium-r-*-*-12-*-*-*-*-*-*-*"
#define DEF_INFO_TEXT_FONT "-*-helvetica-medium-r-*-*-12-*-*-*-*-*-*-*"
#endif /* !I18N */

#define DEF_FRAME_THICKNESS 1	       /* linewidth of the move/resize frame */

/* double click delay time in ms */
#define DEF_DBLCLICK_TIME 	250 

#define DEF_XPM_CLOSENESS	40000 


/* color component weights for PseudoColor displays 
 * (DEF_RED_WEIGHT*DEF_GREEN_WEIGHT*DEF_BLUE_WEIGHT must be < 2^depth)
 * each component must also be >=2 */
#define DEF_RED_WEIGHT 5
#define DEF_GREEN_WEIGHT 5
#define DEF_BLUE_WEIGHT 5

/* animation parameters */
#define SHADE_ANIMATION_STEPS	15
#define SHADE_ANIMATION_DELAY 	0
#define MINIATURIZE_ANIMATION_STEPS 20
#define MINIATURIZE_ANIMATION_DELAY 500
#define ARRANGE_ICONS_ANIMATION_STEPS 35
#define ARRANGE_ICONS_ANIMATION_DELAY 0

#define DOCK_SPRING_DELAY	200
#define DOCK_SPRING_STEPS	20

#define DOCK_EXTRA_SPACE	3
/* blink interval when invoking a menu item */
#define MENU_BLINK_DELAY	70000

#define MOVE_THRESHOLD	3 /* how many pixels to move before dragging windows */
/*
 * defined in config.h by configure
 * #define CPP_PATH	"/lib/cpp"
*/
/*
 *----------------------------------------------------------------------
 * You should not modify the following values, unless you know
 * what you're doing.
 *----------------------------------------------------------------------
 */

#define FRAME_BORDER_WIDTH 1	       /* width of window border for frames */

#define RESIZEBAR_HEIGHT 8	       /* height of the resizebar */
#define RESIZEBAR_MIN_WIDTH 20	       /* min. width of handles-corner_width */
#define RESIZEBAR_CORNER_WIDTH 28      /* width of the corner of resizebars */


/* minimum size for windows */
#define MIN_WSIZE	4

#define MIN_TITLEFONT_HEIGHT(h)   ((h)>14 ? (h) : 14)

/* button info */
#define MAX_LBUTTONS 	1	       /* max. no of buttons at left of tbar */
#define MAX_RBUTTONS 	1	       /* max. no of buttons at right */

#ifdef NEWSTYLE
#define PRED_BPIXMAPS	6
#else
#define PRED_BPIXMAPS	3	       /* number of predefined button pixmaps*/
#endif

#define ICON_WIDTH	64	       /* size of the icon window */
#define ICON_HEIGHT	64
#define ICON_BORDER_WIDTH 2

#define MAX_ICON_WIDTH	60	       /* size of the icon pixmap */
#define MAX_ICON_HEIGHT 48

#define MAX_WORKSPACES  120

#define MAX_MENU_TEXT_LENGTH 512

#define MAX_RESTART_ARGS	16

#define MAX_COMMAND_SIZE	1024

#define MAX_DEAD_PROCESSES	256

#define DITHERING

#define DEBUG0

#ifdef I18N
#define X_LOCALE
#endif

#define DOCK

#endif /* WMCONFIG_H_ */

