/* -*- C++ -*- */
/*******************************************************************/
/* random thoughts
   E apps on other nodes can send sound to the server, 
   via environment variable ESPEAKER=node:port
   SIGHUP should reset ownership of daemon

[raster!raster@trode.redhat.com] actuyally dump like 1k oir
          whatever - tell soudncard to start at the first 1kb
          -white readihg this (the soudncard) write the 2nd kb -
          then the 3rd etc... u just have to keep in step witht
          he card and a little ahead of it... but only just enuf
          to minimise lag.
[raster!raster@trode.redhat.com] www.4front.com i thing.. the
          4front peoepl wrote ythe linxu soudn drivers.. they
          ahve programing docs..
*/

#include "esd.h"

/*******************************************************************/
/* globals */
esd_player_t *esd_players_list = NULL;
esd_player_t *esd_recorder = NULL;
esd_player_t *esd_monitor = NULL;

/*******************************************************************/
/* for debugging purposes, dump the list of the clients and data */
void dump_players()
{
    esd_player_t *player = esd_players_list;

    while ( player != NULL ) {
	printf( "player: (%d) [0x%08x]\n", 
		player->source_id, player );
	player = player->next;
    }
    return;
}

/*******************************************************************/
/* deallocate memory for the player */
void free_player( esd_player_t *player )
{
    /* free any memory allocated with the player */
    free( player->data_buffer );

    /* free the player memory itself */
    free( player );
    return;
}

/*******************************************************************/
/* add a complete new player into the list of players at head */
void add_player( esd_player_t *player )
{
    /* printf ( "adding player 0x%08x\n", new_player ); */
    player->next = esd_players_list;
    esd_players_list = player;
    return;
}

/*******************************************************************/
/* remove a player from the player list */
void remove_player( esd_player_t *player )
{
    esd_player_t *previous = NULL;
    esd_player_t *current = esd_players_list;

    /* iterate until we hit a NULL */
    while ( current != NULL )
    {
	/* see if we hit the target player */
	if ( current == player ) {
	    if( previous != NULL ){
		/* we are deleting in the middle of the list */
		previous->next = current->next;
	    } else { 
		/* we are deleting the head of the list */
		esd_players_list = current->next;
	    }

	    /* TODO: remove client should be independent of players */
	    if ( ( player->format & ESD_MASK_MODE ) == ESD_STREAM )
		remove_client( player->parent );
	    free_player( player );

	    return;
	}

	/* iterate through the list */
	previous = current;
	current = current->next;
    }

    /* hmm, we didn't find the desired player, just get on with life */
    printf( "player not found (%d)\n", player->source_id );
    return;
}


/*******************************************************************/
/* read block of data from client, return < 0 to have it removed */
int read_player( esd_player_t *player )
{
    fd_set rd_fds;
    int actual = 0, can_read = 0;
    struct timeval timeout;
    char message[ 100 ];

    switch( player->format & ESD_MASK_MODE ) {
    case ESD_STREAM:
	/* use select to prevent blocking clients that are ready */
	timeout.tv_sec = 0;
	timeout.tv_usec = 0;
	FD_ZERO( &rd_fds );
	FD_SET( player->source_id, &rd_fds );

	/* if the data is ready, read a block */
	if ( select( player->source_id + 1, 
		     &rd_fds, NULL, NULL, &timeout ) ) 
	{
	    actual = read( player->source_id, 
			   player->data_buffer, 
			   player->buffer_length );
	    player->actual_length = actual;
	} 

	/* check for data (actual>0), timeout (==0), or error(<0) */
	if ( actual > 0 ) {
	    player->last_read = time(NULL);
	} else if ( 0 == actual ) {
	    if ( time(NULL) > player->last_read + ESD_CLIENT_TIMEOUT )
	    {
		printf( "client (%d) timed out, closing\n", 
			player->source_id );
		actual = -1;
	    }
	} else if ( actual < 0 ) {
	    sprintf( message, "error reading client (%d)\n", 
		     player->source_id );
	    perror( message );
	}

	break;

    case ESD_SAMPLE:

	actual = ( ((esd_sample_t*)player->parent)->sample_length 
		   - player->last_pos > player->buffer_length )
	    ? player->buffer_length 
	    : ((esd_sample_t*)player->parent)->sample_length - player->last_pos;
	if ( actual > 0 ) {
	    memcpy( player->data_buffer, 
		    ((esd_sample_t*)player->parent)->data_buffer 
		    + player->last_pos, actual );
	    player->last_pos += actual;
	} else {
	    return -1;
	}

	break;

    default:
	printf( "read_player: format %08x not supported\n", 
		player->format );
	return -1;
    }

    return actual;
}

/*******************************************************************/
/* send the players buffer to it's associated socket, remove if EOF */
void monitor_write() {
    int length;

    /* make sure we have a monitor connected */
    if ( !esd_monitor ) 
	return;

    /* write the data buffer to the socket */
    length = write ( esd_monitor->source_id, 
		     esd_monitor->data_buffer, 
		     esd_monitor->buffer_length );

    if ( length <= 0 ) {
	/* error on write, close it down */
	printf( "closing monitor (%d)\n", esd_monitor->source_id );
	remove_client( esd_monitor->parent );
	free_player( esd_monitor );
	esd_monitor = NULL;
    }

    return;
}

void recorder_write() {
    int length;

    /* make sure we have a recorder connected */
    if ( !esd_recorder ) 
	return;

    /* write the data buffer to the socket */
    length = write ( esd_recorder->source_id, 
		     esd_recorder->data_buffer, 
		     esd_recorder->buffer_length );

    if ( length <= 0 ) {
	/* couldn't send anything, close it down */
	printf( "closing recorder (%d)\n", esd_recorder->source_id );

	/* stop recording */
	audio_close();
	esd_audio_format &= ~ESD_RECORD;
	audio_open();

	/* clear the recorder */
	remove_client( esd_recorder->parent );
	free_player( esd_recorder );
	esd_recorder = NULL;
    }

    return;
}

/*******************************************************************/
/* allocate and initialize a player from client stream */
esd_player_t *new_stream_player( int client_fd )
{
    esd_player_t *player;

    /* make sure we have the memory to save the client... */
    player = (esd_player_t*) malloc( sizeof(esd_player_t) );
    if ( player == NULL ) {
	return NULL;
    }
    
    /* and initialize the player */
    player->next = NULL;
    player->parent = NULL;
    read( client_fd, &player->format, sizeof(player->format) );
    player->format &= ~ESD_MASK_MODE; /* force to ESD_STREAM */
    read( client_fd, &player->rate, sizeof(player->rate) );
    player->source_id = client_fd;

    printf( "connection format: 0x%08x at %d Hz\n", 
	    player->format, player->rate );

    /* calculate buffer length to match the mix buffer duration */
    player->buffer_length = ESD_BUF_SIZE * player->rate / 44100;
    if ( (player->format & ESD_MASK_BITS) == ESD_BITS8 )
	player->buffer_length /= 2;
    if ( (player->format & ESD_MASK_CHAN) == ESD_MONO )
	player->buffer_length /= 2;

    /* force to an even multiple of 4 */
    player->buffer_length += player->buffer_length % 4;

    player->data_buffer
	= (void *) malloc( player->buffer_length );

    /* if not enough room for data buffer, clean up, and return NULL */
    if ( player->data_buffer == NULL ) {
	free( player );
	return NULL;
    }

    /* everything's ok, return the allocated player */
    player->last_read = time(NULL);

    printf( "player: (%d) [0x%08x]\n", 
	    player->source_id, player );

    return player;
}

/*******************************************************************/
/* allocate and initialize a player from client stream */
esd_player_t *new_sample_player( int sample_id )
{
    esd_player_t *player;
    esd_sample_t *sample;

    /* find the sample we want to play */
    for ( sample = esd_samples_list ; sample != NULL 
	      ; sample = sample->next )
    {
	if ( sample->sample_id == sample_id ) {
	    break;
	}
    }
    /* if we didn't find it, return NULL */
    if ( sample == NULL ) {
	return NULL;
    }
        
    /* make sure we have the memory to save the client... */
    player = (esd_player_t*) malloc( sizeof(esd_player_t) );
    if ( player == NULL ) {
	return NULL;
    }
    
    /* and initialize the player */
    player->next = NULL;
    player->parent = sample;
    player->format = sample->format | ESD_SAMPLE;
    player->rate = sample->rate;
    player->source_id = sample->sample_id;

    printf( "connection format: 0x%08x at %d Hz\n", 
	    player->format, player->rate );

    /* calculate buffer length to match the mix buffer duration */
    player->buffer_length = ESD_BUF_SIZE * player->rate / 44100;
    if ( (player->format & ESD_MASK_BITS) == ESD_BITS8 )
	player->buffer_length /= 2;
    if ( (player->format & ESD_MASK_CHAN) == ESD_MONO )
	player->buffer_length /= 2;

    /* force to an even multiple of 4 */
    player->buffer_length += player->buffer_length % 4;

    player->data_buffer
	= (void *) malloc( player->buffer_length );

    /* if not enough room for data buffer, clean up, and return NULL */
    if ( player->data_buffer == NULL ) {
	free( player );
	return NULL;
    }

    /* everything's ok, return the allocated player */
    player->last_pos = 0;

    printf( "player: (%d) [0x%08x]\n", 
	    player->source_id, player );

    return player;
}
