
#include "esd.h"

/*******************************************************************/
/* globals */
esd_sample_t *esd_samples_list = NULL;

static int esd_next_sample_id = 1;	/* sample id = 0 is an error */

/*******************************************************************/
/* for debugging purposes, dump the list of the clients and data */
void dump_samples()
{
    esd_sample_t *sample = esd_samples_list;

    while ( sample != NULL ) {
	printf( "sample: (%d) [0x%08x] - %d bytes\n", 
		sample->sample_id, sample, sample->sample_length );
	sample = sample->next;
    }
    return;
}

/*******************************************************************/
/* deallocate memory for the client */
void free_sample( esd_sample_t *sample )
{
    /* free any memory allocated with the sample */
    free( sample->data_buffer );

    /* free the sample memory itself */
    free( sample );
    return;
}

/*******************************************************************/
/* add a complete new client into the list of samples at head */
void add_sample( esd_sample_t *new_sample )
{
    /* printf ( "adding sample 0x%08x\n", new_sample ); */
    new_sample->next = esd_samples_list;
    esd_samples_list = new_sample;
    return;
}

/*******************************************************************/
/* remove a sample from the sample list */
void remove_sample( int id )
{
    esd_sample_t *previous = NULL;
    esd_sample_t *current = esd_samples_list;

    /* iterate until we hit a NULL */
    while ( current != NULL )
    {
	/* see if we hit the target sample */
	if ( current->sample_id == id ) {
	    if( previous != NULL ){
		/* we are deleting in the middle of the list */
		previous->next = current->next;
	    } else { 
		/* we are deleting the head of the list */
		esd_samples_list = current->next;
	    }

	    /* remove last traces of sample from existence */
	    free_sample( current );

	    return;
	}

	/* iterate through the list */
	previous = current;
	current = current->next;
    }

    /* hmm, we didn't find the desired sample, just get on with life */
    printf( "sample not found (%d)\n", id );
    return;
}

/*******************************************************************/
/* allocate and initialize a sample from client stream */
#define min(a,b) ( ( (a)<(b) ) ? (a) : (b) )
int read_sample( esd_sample_t *sample )
{
    int actual, total = 0;

    while ( ( total < sample->sample_length ) && ( actual != 0 ) ) {
	actual = read( sample->parent->fd, sample->data_buffer + total,
		       min( ESD_BUF_SIZE, sample->sample_length-total) );
	total += actual;
    }

    printf( "%d bytes total\n", total );
    return total;
}

/*******************************************************************/
/* allocate and initialize a sample from client stream */
esd_sample_t *new_sample( int client_fd )
{
    esd_sample_t *sample;

    /* make sure we have the memory to save the client... */
    sample = (esd_sample_t*) malloc( sizeof(esd_sample_t) );
    if ( sample == NULL ) {
	return NULL;
    }
    
    /* and initialize the sample */
    sample->next = NULL;
    sample->parent = NULL;
    read( client_fd, &sample->format, sizeof(sample->format) );
    read( client_fd, &sample->rate, sizeof(sample->rate) );
    read( client_fd, &sample->sample_length, sizeof(sample->sample_length) );
    sample->sample_id = esd_next_sample_id++;

    printf( "connection format: 0x%08x at %d Hz\n", 
	    sample->format, sample->rate );

    /* force to an even multiple of 4, do it in the player */
    sample->data_buffer
	= (void *) malloc( sample->sample_length );

    /* if not enough room for data buffer, clean up, and return NULL */
    if ( sample->data_buffer == NULL ) {
	free( sample );
	return NULL;
    }

    printf( "sample: <%d> [0x%08x] - %d bytes\n", 
	    sample->sample_id, sample, sample->sample_length );
    write( client_fd, &sample->sample_id, sizeof(sample->sample_id) );

    return sample;
}

/*******************************************************************/
/* spawn a player for this sample */
int play_sample( int sample_id )
{
    esd_player_t *player = NULL;

    player = new_sample_player( sample_id );
    if ( player == NULL )
	return 0;

    add_player( player );
    return 1;
}
