;; text-circle a script for The GIMP
;; Shuji Narazaki (narazaki@InetQ.or.jp)
;; Time-stamp: <97/03/01 09:49:10 narazaki@InetQ.or.jp>

(define (script-fu-circle-logo text radius font font-size slant antialias)
  (let* ((width (* radius 2.5))
	 (height (* radius 2.5))
	 (img (car (gimp-image-new width height RGB)))
	 (drawable (car (gimp-layer-new img width height RGBA_IMAGE
					"text-circle" 100 NORMAL)))
	 (char-num (string-length text))
	 (radians (/ (* 2 *pi*) char-num))
	 (rad-90 (/ *pi* 2))
	 (center-x (/ width 2))
	 (center-y (/ height 2))
	 (fixed-pole "]")
	 (font-infos (gimp-text-get-extents fixed-pole font-size PIXELS
					    "*" font "*" slant "*" "*"))
	 (extra (max 0 (- (nth 0 font-infos) 5))) ; why 4? See text_tool.c.
	 (desc (nth 3 font-infos))
	 (letter "")
	 (new-layer #f)
	 (index 0))
    (gimp-image-disable-undo img)
    (gimp-image-add-layer img drawable 0)
    (gimp-edit-fill img drawable)
    (while (< index char-num)
      (set! letter (substring text index (+ index 1)))
      (if (not (equal? " " letter))
	  ;; Running gimp-text with " " causes an error!
	  (let* ((new-layer (car (gimp-text img -1 0 0
					    (string-append letter fixed-pole)
					    1 antialias
					    font-size PIXELS
					    "*" font "*" slant "*" "*")))
		 (width (car (gimp-drawable-width new-layer)))
		 (height (car (gimp-drawable-height new-layer)))
		 (rotate-radius (- (/ height 2) desc))
		 (rad1 (- (* index radians) rad-90))
		 (rad2 (* index radians)))
	    (gimp-layer-resize new-layer (- width extra 1) height 0 0)
	    '(gimp-layer-translate new-layer (* index 34) 10)
	    (gimp-layer-translate new-layer
				  (+ center-x
				     (* radius (cos rad1))
				     (- (* rotate-radius (sin rad2)))
				     (- (/ width 2)))
				  (+ center-y
				     (* radius (sin rad1))
				     (- (* rotate-radius (cos rad2)))
				     (- (/ height 2))))
	    (gimp-rotate img new-layer 1 rad2)))
      (set! index (+ index 1)))
    (gimp-image-merge-visible-layers img CLIP-TO-IMAGE)
    (gimp-image-enable-undo img)
    (gimp-display-new img)))

(script-fu-register "script-fu-circle-logo"
		    "<Toolbox>/Xtns/Script-Fu/Logos/Text Circle"
		    "Render the specified text along the perimeter of a circle"
		    "Shuji Narazaki (narazaki@InetQ.or.jp)"
		    "Shuji Narazaki"
		    "1997"
		    ""
		    SF-VALUE "Text" "\"Ring Ring Ring, Tiger! Tiger! Tiger! \""
		    SF-VALUE "Radius" "80"
		    SF-VALUE "Family" "\"helvetica\""
		    SF-VALUE "Font Size" "18"
		    SF-VALUE "Slant" "\"r\""
		    SF-VALUE "Antialias(0,1)" "1"
)

;(text-circle "This is a sample string. " "Bodoni" 24 80)
;; end of text-circle.scm
