/* 
 * FILE:
 * Ledger.h 
 *
 * FUNCTION:
 * Definitions for a number of functions that interface 
 * between the transaction engine, and the register GUI.
 *
 * The xaccGetCurrentSplit() method returns the split at which 
 *    the cursor is curently located.
 *
 * The xaccLoadRegEntry() method will copy data from the indicated
 *    split to the table cursor; i.e. to the current table 
 *    register entry.  It does not refresh the GUI afte the
 *    copy.
 *
 * The xaccLoadRegister() subroutine will copy transaction
 *    information from a list of splits to the rows of the
 *    register GUI.  The third argument, default_source_acc,
 *    will be used to initialize the source account of a new,
 *    blank split appended to the tail end of the register.
 *    This "blank split" is the place where the user can 
 *    edit info to create new tranasactions.
 *
 * The xaccSaveRegEntry() method will copy the contents 
 *    from the cursor to a split.  The split/transaction
 *    that is updated is the one associated with the current 
 *    cursor (register entry) position.
 */

#ifndef __XACC_LEDGER_H__
#define __XACC_LEDGER_H__

#include "Group.h"
#include "register.h"
#include "Transaction.h"

Split * xaccGetCurrentSplit (BasicRegister *reg);
void    xaccLoadRegEntry (BasicRegister *reg, Split *split);
void    xaccLoadRegister (BasicRegister *reg, Split **slist,
                          Account *default_source_acc);

void    xaccSaveRegEntry (BasicRegister *reg);

void    xaccLoadXferCell (ComboCell *cell,  AccountGroup *grp);


#endif /* __XACC_LEDGER_H__ */
