/***********************************************************************
 *
 * Module: fl_edit.h
 * Author: Marc van Kempen (marc@bowtie.nl)
 * Desc:   include file for textedit object 
 *
 * Copyright (c) 1996, Marc van Kempen
 *
 * All rights reserved.
 *
 * This software may be used, modified, copied, distributed, and
 * sold, in both source and binary form provided that the above
 * copyright and these terms are retained, verbatim, as the first
 * lines of this file.  Under no circumstances is the author
 * responsible for the proper functioning of this software, nor does
 * the author assume any responsibility for damages incurred with
 * its use.
 *
 ***********************************************************************/

/*       $Id: fl_edit.h,v 1.14 1996/12/22 13:56:43 gena Exp $
 */

#ifndef __FL_EDIT_H__
#define __FL_EDIT_H__

#include <sys/param.h>
#include <forms.h>
#include "textbuf.h"

/***********************************************************************
 *
 * Defines
 *
 ***********************************************************************/

#define FL_TEXTEDIT		1001
#define FL_NORMAL_TEXTEDIT 	0

typedef struct {
    TextBuf tb;			/* a textbuffer */
    char name[MAXPATHLEN];	/* the name of the buffer */
    int r, c;			/* the coordinates of the cursor */
    int cpos;			/* desired x coord of cursor */
    int topline;		/* the topline to be shown */
    int leftcol;		/* the leftmost column to be shown 
				   in characters */
    int text_style;		/* style of text */
    int text_size;		/* size of text */
    int ccol;			/* color of textcursor */
    FL_OBJECT *sb, *hsb;	/* scrollbar objects */
    int sselr, sselc;		/* row & column of selection start */
    int eselr, eselc;		/* row & column of selection end */
    int flags;			/* special flags */
#define	FL_TEXTEDIT_READONLY	0x01	/* widget is not editable */
#define	FL_TEXTEDIT_NOTDRAWN	0x02	/* no FL_DRAW received yet */
#define	FL_TEXTEDIT_VSBAR	0x04	/* vertical scrollbar */
#define	FL_TEXTEDIT_VSBAR_AUTO	0x08	/* vertical scrollbar autohide */
#define	FL_TEXTEDIT_HSBAR	0x10	/* horizontal scrollbar */
#define	FL_TEXTEDIT_HSBAR_AUTO	0x20	/* horizontal scrollbar autohide */
    char *exp;			/* text to search for */
    int oldh, oldw;		/* old object size */
} SPEC;

/***********************************************************************
 *
 * Prototypes
 *
 ***********************************************************************/

extern FL_OBJECT *fl_create_textedit(int, FL_Coord, FL_Coord, FL_Coord,FL_Coord,
				     const char *);
extern FL_OBJECT *fl_add_textedit(int, FL_Coord, FL_Coord, FL_Coord, FL_Coord,
				  const char *);
extern void	  fl_save_textedit(FL_OBJECT *ob, char *fname);
extern void	  fl_load_textedit(FL_OBJECT *ob, char *fname);
extern void	  fl_set_textedit_fontsize(FL_OBJECT *, int);
extern int	  fl_get_textedit_fontsize(FL_OBJECT *);
extern void	  fl_set_textedit_fontstyle(FL_OBJECT *, int);
extern int	  fl_get_textedit_fontstyle(FL_OBJECT *);
extern void	  fl_set_textedit_color(FL_OBJECT *, int, int, int, int);
extern void	  fl_get_textedit_color(FL_OBJECT *, int *, int *, int *);
extern void	  fl_set_textedit_bufname(FL_OBJECT *, char *);
extern char	 *fl_get_textedit_bufname(FL_OBJECT *);
extern void	  fl_set_textedit_bufchanged(FL_OBJECT *, int);
extern int	  fl_get_textedit_bufchanged(FL_OBJECT *);
extern void	  fl_set_textedit(FL_OBJECT *, char *buf, long len);
extern char	 *fl_get_textedit(FL_OBJECT *, u_long *len);
extern FL_OBJECT *fl_get_textedit_vscrollbar(FL_OBJECT *);
extern void	  fl_set_textedit_vscrollbar(FL_OBJECT *, int);
extern FL_OBJECT *fl_get_textedit_hscrollbar(FL_OBJECT *);
extern void	  fl_set_textedit_hscrollbar(FL_OBJECT *, int);
extern void	  fl_calc_cursorpos(FL_OBJECT *, u_long, int *, int *);
extern void	  fl_set_textedit_cursorpos(FL_OBJECT *, int, int, long);
extern void	  fl_get_textedit_cursorpos(FL_OBJECT *, int *, int *,u_long *);
extern void	  fl_clear_textedit(FL_OBJECT *);
extern int	  fl_set_textedit_wrap(FL_OBJECT *, int, int);
extern void	  fl_insert_textedit(FL_OBJECT *, char *);
extern void	  fl_insert_textedit_file(FL_OBJECT *, char *);
extern void	  fl_textedit_paste(FL_OBJECT *);
extern int	  fl_textedit_readonly(FL_OBJECT *, int);
extern int	  fl_get_textedit_topline(FL_OBJECT *);
extern int	  fl_get_textedit_maxline(FL_OBJECT *);
extern int	  fl_get_textedit_screenlines(FL_OBJECT *);
extern void	  fl_set_textedit_screenlines(FL_OBJECT *, int);
extern char	 *fl_get_textedit_curline(FL_OBJECT *);
extern char	 *fl_get_textedit_line(FL_OBJECT *, int);
extern void	  fl_add_textedit_line(FL_OBJECT *, char *);
extern void	  fl_addto_textedit(FL_OBJECT *, char *);
extern int	  fl_textedit_line_visible(FL_OBJECT *, int);
extern void	  fl_insert_textedit_line(FL_OBJECT *, int, char *);
extern void	  fl_delete_textedit_line(FL_OBJECT *, int);
extern void	  fl_replace_textedit_line(FL_OBJECT *, int, char *);
extern void	  fl_deselect_textedit(FL_OBJECT *);
extern int	  fl_isselected_textedit_line(FL_OBJECT *, int);
extern void	  fl_select_textedit_line(FL_OBJECT *, int);
extern void	  fl_deselect_textedit_line(FL_OBJECT *, int);
extern void	  fl_set_textedit_line_color(FL_OBJECT *, int, int, int);
extern void	  fl_get_textedit_line_color(FL_OBJECT *, int, int *, int *);
extern void	  fl_scroll_textedit(FL_OBJECT *, int);
#define	FL_TEXTEDITSCROLL_PGUP		0x01
#define	FL_TEXTEDITSCROLL_PGDOWN	0x02
#define	FL_TEXTEDITSCROLL_LINEUP	0x04
#define	FL_TEXTEDITSCROLL_LINEDOWN	0x08
#define	FL_TEXTEDITSCROLL_TOP		0x10
#define	FL_TEXTEDITSCROLL_BOTTOM	0x20

#endif /* __FL_EDIT_H__ */
