/*
 * Copyright(c) 1996 by Gennady B. Sorokopud (gena@NetVision.net.il)
 *
 * This software can be freely redistributed and modified for
 * non-commercial purposes as long as above copyright
 * message and this permission notice appear in all
 * copies of distributed source code and included as separate file
 * in binary distribution.
 *
 * Any commercial use of this software requires author's permission.
 *
 * This software is provided "as is" without expressed or implied
 * warranty of any kind.
 * Under no circumstances is the author responsible for the proper
 * functioning of this software, nor does the author assume any
 * responsibility for damages incurred with its use.
 *
 */

/* $Id: imap.c,v 1.37 1996/12/16 15:45:49 gena Exp $
 */

#include <fmail.h>
#include <umail.h>

static int imapsock = -1;	/* IMAP server socket 		*/
static FILE *imap_in, *imap_out;/* in/out file descriptors 	*/
static char *response = NULL;	/* responce from server 	*/
static char *plist = NULL;	/* parenthized list		*/
static char *elem = NULL;	/* parenthized list element	*/
static int icap = 0x00;		/* server capabilities 		*/
static int istate = 0x00;	/* server state 		*/
static struct _mail_folder	*ifold = NULL, /* selected folder           */
				*iinbox = NULL,/* INBOX                     */
				*itrash = NULL;/* TRASH                     */
static struct _mail_msg		*imsg = NULL;  /* "current" message"        */
static u_long *search_res = NULL; /* search results                         */
static int uid = -1;		/* UID */
static char *imap_h = NULL;	/* listed hierarchy */

#define	MAX_IMAP_STRING	524288	/* don't malloc more then this */

#define	IMAP_INBOX	"INBOX"
#define	IMAP_TRASH	"trash"

/* commands */
#define	ICOM_NOCOM	0x00	/* just wait for greeting */
#define	ICOM_CAPABILITY	0x01	/* CAPABILITY */
#define	ICOM_NOOP	0x02	/* NOOP */
#define	ICOM_LOGOUT	0x03	/* LOGOUT */
#define ICOM_AUTHENITCATE	0x04	/* AUTHENTICATE */
#define ICOM_LOGIN	0x05	/* LOGIN */
#define ICOM_SELECT	0x06	/* SELECT */
#define ICOM_EXAMINE	0x07	/* EXAMINE */
#define ICOM_CREATE	0x08	/* CREATE */
#define	ICOM_DELETE	0x09	/* DELETE */
#define	ICOM_RENAME	0x0a	/* RENAME */
#define ICOM_SUBSCRIBE	0x0b	/* SUBSCRIBE */
#define	ICOM_UNSUBSCRIBE	0x0c	/* UNSUBSCRIBE */
#define	ICOM_LIST	0x0d	/* LIST */
#define	ICOM_LSUB	0x0e	/* LSUB */
#define ICOM_STATUS	0x0f	/* STATUS */
#define ICOM_APPEND	0x10	/* APPEND */
#define ICOM_CHECK	0x11	/* CHECK */
#define ICOM_CLOSE	0x12	/* CLOSE */
#define	ICOM_EXPUNGE	0x13	/* EXPUNGE */
#define ICOM_SEARCH	0x14	/* SEARCH */
#define ICOM_FETCH	0x15	/* FETCH */
#define ICOM_STORE	0x16	/* STORE */
#define ICOM_COPY	0x17	/* COPY */
#define ICOM_UIDSEARCH	0x18	/* UID SEARCH */
#define ICOM_UIDCOPY	0x19	/* UID COPY */
#define ICOM_UIDFETCH	0x1a	/* UID FETCH */
#define ICOM_UIDSTORE	0x1b	/* UID STORE */

#define	ICOM_MAXCOMM	0x1c
/* Command flags */
#define	ICOM_DONTLOG	0x20	/* don't log output */
#define	ICOM_SILENT	0x40	/* don't display error messages */
#define	ICOM_DONTWAIT	0x80	/* don't wait for server responce */
#define	ICOM_ABORT	0x0100	/* abort command */

#define	ICOM_MATCHALL	0xff

/* command status */
#define	IMAP_FATAL	-2
#define	IMAP_ERROR	-1
#define	IMAP_OK		0
#define	IMAP_NO		1
#define	IMAP_BAD	2

#define	IMAP_TIMER	1500	/* inactivity timer */

/* capabilities */
#define	ICAP_IMAP2	0x01	/* IMAP2 */
#define	ICAP_IMAP4	0x02	/* IMAP4 */
#define	ICAP_IMAP4REV1	0x04	/* IMAP4rev1 */
#define	ICAP_STATUS	0x08	/* has STATUS command */
#define	ICAP_SCAN	0x10	/* has SCAN command */

/* states */
#define	ISTATE_NOAUTH	0x01	/* no authentication yet */
#define	ISTATE_AUTH	0x02	/* successfull authentication */
#define	ISTATE_SELECT	0x03	/* mailbox selected */
#define	ISTATE_NOCON	0x04	/* not connected */

/* FETCH message parts */
#define	IFETCH_BODY		0x00
#define	IFETCH_BODYSTRUCTURE	0x01
#define	IFETCH_ENVELOPE		0x02
#define	IFETCH_FLAGS		0x03
#define	IFETCH_INTERNALDATE	0x04
#define	IFETCH_RFC822		0x05
#define	IFETCH_RFC822HEADER	0x06
#define	IFETCH_RFC822SIZE	0x07
#define	IFETCH_RFC822TEXT	0x08
#define	IFETCH_UID		0x09

char *mparts[] = {
"BODY",
"BODYSTRUCTURE",
"ENVELOPE",
"FLAGS",
"INTERNALDATE",
"RFC822",
"RFC822.HEADER",
"RFC822.SIZE",
"RFC822.TEXT",
"UID",
NULL
};

char *plist_getnext(char *list, char **next);
char *plist_getnext_string(char *list, char **next);
void end_plist();
int start_plist();
char *get_imap_folder_domain(struct _mail_folder *folder);
char *get_imap_folder_domain_path(struct _mail_folder *folder);
struct _mail_folder *imap_folder_switch(struct _mail_folder *folder);
void expand_uid_range(struct _mail_folder *folder, struct _mail_msg *msg, int mask1, int mask2, int *uid_top, int *uid_bottom);
void imap_disconnect();
int imap_connect();
int imap_reconnect();
int delete_imap_message(struct _mail_msg *msg);
int move_to_imap_folder(struct _mail_msg *msg, struct _mail_folder *folder);
int move_to_imap_folder_range(struct _mail_msg *msg, struct _mail_folder *folder);
struct _mail_msg *copy_to_imap_folder(struct _mail_msg *msg, struct _mail_folder *folder);
struct _mail_msg *copy_to_imap_folder_range(struct _mail_msg *msg, struct _mail_folder *folder);
int update_imap_message(struct _mail_msg *msg);
void update_imap_message_range(struct _mail_msg *msg);
char *imap_string(char *str);

typedef	int (*imap_process)(int command, char *tag, char *resp, char *param);

int unk_process(int command, char *tag, char *resp, char *param);
int ok_process(int command, char *tag, char *resp, char *param);
int no_process(int command, char *tag, char *resp, char *param);
int bad_process(int command, char *tag, char *resp, char *param);
int bye_process(int command, char *tag, char *resp, char *param);
int disc_process(int command, char *tag, char *resp, char *param);
int cap_process(int command, char *tag, char *resp, char *param);
int list_process(int command, char *tag, char *resp, char *param);
int stat_process(int command, char *tag, char *resp, char *param);
int preauth_process(int command, char *tag, char *resp, char *param);
int flags_process(int command, char *tag, char *resp, char *param);
int exists_process(int command, char *tag, char *resp, char *param);
int recent_process(int command, char *tag, char *resp, char *param);
int search_process(int command, char *tag, char *resp, char *param);
int fetch_process(int command, char *tag, char *resp, char *param);
int expunge_process(int command, char *tag, char *resp, char *param);
int append_process(int command, char *tag, char *resp, char *param);

void process_respcode(int command, char *param);

typedef struct _imapresp {
int command;
char tag [6];
char resp[32];
char *param;
imap_process process;
} imapresp;

struct _imapresp imap_responces[] = {
{ ICOM_MATCHALL,	"*",	"OK",	"$all",	ok_process  },
{ ICOM_MATCHALL,	"*",	"NO",	"$all",	no_process  },
{ ICOM_MATCHALL,	"*",	"BAD",	"$all",	bad_process },
{ ICOM_LOGOUT,		"*",	"BYE",	"$all",	bye_process },
{ ICOM_MATCHALL,	"*",	"BYE",	"$all",	disc_process },
{ ICOM_LIST,		"*",	"LIST",	"$all",	list_process },
{ ICOM_LSUB,		"*",	"LSUB",	"$all",	list_process },
{ ICOM_CAPABILITY,	"*",	"CAPABILITY",	"$all",	cap_process },
{ ICOM_STATUS,		"*",	"STATUS",	"$all",	stat_process },
{ ICOM_MATCHALL,	"*",	"PREAUTH",	"$all",	preauth_process },
{ ICOM_SELECT,		"*",	"FLAGS",	"$all",	flags_process },
{ ICOM_EXAMINE,		"*",	"FLAGS",	"$all",	flags_process },
{ ICOM_SEARCH,		"*",	"SEARCH","$all",	search_process },
{ ICOM_UIDSEARCH,	"*",	"SEARCH","$all",	search_process },
{ ICOM_FETCH,		"*",	"$num",	"$FETCH (",	fetch_process },
{ ICOM_UIDFETCH,	"*",	"$num",	"$FETCH (",	fetch_process },
{ ICOM_STORE,		"*",	"$num",	"$FETCH (",	fetch_process },
{ ICOM_UIDSTORE,	"*",	"$num",	"$FETCH (",	fetch_process },
{ ICOM_EXPUNGE,		"*",	"$num",	"EXPUNGE",	expunge_process },
{ ICOM_APPEND,		"+",	"$all",	"$all",		append_process },

{ ICOM_MATCHALL,	"*",	"$num",	"EXISTS",	exists_process },
{ ICOM_MATCHALL,	"*",	"$num",	"RECENT",	recent_process },
{ ICOM_MATCHALL,	"$all",	"OK",	"$all",	ok_process  },
{ ICOM_MATCHALL,	"$all",	"NO",	"$all",	no_process  },
{ ICOM_MATCHALL,	"$all",	"BAD",	"$all",	bad_process },
{ ICOM_MATCHALL,	"$all",	"$all",	"$all",	unk_process },	/* unknown */
{ ICOM_NOCOM,		"$all",	"$all",	"$all",	unk_process }
};

char *imap_commands[] = {
NULL,
"CAPABILITY",
"NOOP",
"LOGOUT",
"AUTHENTICATE",
"LOGIN",
"SELECT",
"EXAMINE",
"CREATE",
"DELETE",
"RENAME",
"SUBSCRIBE",
"UNSUBSCRIBE",
"LIST",
"LSUB",
"STATUS",
"APPEND",
"CHECK",
"CLOSE",
"EXPUNGE",
"SEARCH",
"FETCH",
"STORE",
"COPY",
"UID SEARCH",
"UID COPY",
"UID FETCH",
"UID STORE",
NULL
};

void
process_respcode(command, param)
int command;
char *param;
{
char *p, *q, rcode[127];
int len;

 while (*param == ' ')
	param++;

 if (*param != '[')
	return;

 param++;
 if ((p = strchr(param, ']')) == NULL)
	return;

 len = p - param;
 if ((len < 2) || (len >= 126))
	return;

 strncpy(rcode, param, len);
 rcode[len] = '\0';
 p++;
 p[64] = '\0';
 while (*p == ' ')
	p++;
 if ((q = strchr(p, ' ')) != NULL) {
	*q = '\0';
	while (*q == ' ')
		q++;		   }

 if (!strcasecmp(rcode, "ALERT"))
	fl_show_alert("IMAP ALERT", "", p, 0);
 else
 if (!strcasecmp(rcode, "PARSE"))
	display_msg(MSG_LOG, "IMAP PARSE ERROR", "%.64s", p);
 else
 if (!strcasecmp(rcode, "READ-ONLY"))
	if (ifold && (command != ICOM_EXAMINE))
		ifold->flags |= FRONLY;
 else
 if (!strcasecmp(rcode, "READ-WRITE"))
	if (ifold) ifold->flags &= ~FRONLY;
 else
 if (!strcasecmp(rcode, "TRYCREATE"))
	display_msg(MSG_WARN, "IMAP - Try to create destination mailbox", "%.64s", p);
 else
 if (!strcasecmp(rcode, "NEWNAME"))
	display_msg(MSG_WARN, "The mailbox has been renamed", "%.32s to %.32s", p, q ? q : "unknown");

 return;
}

int
ok_process(command, tag, resp, param)
int command;
char *tag, *resp, *param;
{
 process_respcode(command, param);
 return IMAP_OK;
}

int
no_process(command, tag, resp, param)
int command;
char *tag, *resp, *param;
{
 process_respcode(command, param);
 return IMAP_OK;
}

int
bad_process(command, tag, resp, param)
int command;
char *tag, *resp, *param;
{
 process_respcode(command, param);
 return IMAP_OK;
}

int
bye_process(command, tag, resp, param)
int command;
char *tag, *resp, *param;
{
 return IMAP_OK;
}

int
disc_process(command, tag, resp, param)
int command;
char *tag, *resp, *param;
{
 display_msg(MSG_WARN, "IMAP", "Server closed connection");
 istate = ISTATE_NOCON;
 imap_close(0);
 return IMAP_FATAL;
}

int
append_process(command, tag, resp, param)
int command;
char *tag, *resp, *param;
{
FILE *afd;
char *mfile;
int res;

 if (!imsg)
	return IMAP_ERROR;

 if ((mfile = imsg->get_file(imsg)) == NULL)
	return IMAP_ERROR;

 if ((afd = fopen(mfile, "r")) == NULL)
	return IMAP_ERROR;

 if ((res = putdata(NULL, imsg->msg_len * 2, imap_out, afd)) < 0)	{
	fclose(afd);
	if (res == -2)
		imap_reconnect();
	else
	if (res == -1)
		imap_close(0);
	return IMAP_ERROR;						}

 fclose(afd);

 return IMAP_OK;
}

int
expunge_process(command, tag, resp, param)
int command;
char *tag, *resp, *param;
{
 return IMAP_OK;
}

int
cap_process(command, tag, resp, param)
int command;
char *tag, *resp, *param;
{
char *p;

 icap = 0;

 if ((p = strtok(param, " ")) == NULL) {
	display_msg(MSG_WARN, "IMAP", "Invalid CAPABILITY responce");
	return IMAP_ERROR;
					}

 do {
  if (!strcasecmp(p, "IMAP4"))
	icap |= ICAP_IMAP4;
  else
  if (!strcasecmp(p, "IMAP4rev1")) {
	icap |= ICAP_IMAP4;
	icap |= ICAP_IMAP4REV1;
	icap |= ICAP_STATUS;       }
  else
  if (!strcasecmp(p, "IMAP2"))
	icap |= ICAP_IMAP2;
  else
  if (!strcasecmp(p, "STATUS"))
	icap |= ICAP_STATUS;
  else
  if (!strcasecmp(p, "SCAN"))
	icap |= ICAP_SCAN;
} while ((p = strtok(NULL, " ")) != NULL);

 if (!(icap & ICAP_IMAP4) && !(icap & ICAP_IMAP2)) {
	display_msg(MSG_WARN, "IMAP", "Unsupported IMAP server version\ncan not proceed");
	return IMAP_ERROR;
						   }

 if (icap & ICAP_IMAP2)    {
	display_msg(MSG_WARN, "IMAP", "IMAP2 is not supported");
	return IMAP_ERROR; }

 return IMAP_OK;
}

int
flags_process(command, tag, resp, param)
int command;
char *tag, *resp, *param;
{
 return IMAP_OK;
}

int
exists_process(command, tag, resp, param)
int command;
char *tag, *resp, *param;
{
char *p;

 if (!ifold)
	return IMAP_OK;

 ifold->num_msg = strtoul(resp, &p, 10);

 if (*p) 		   {
	display_msg(MSG_WARN, "IMAP", "Invalid EXISTS responce");
	return IMAP_ERROR; }

 if (ifold->unread_num > ifold->num_msg)
	ifold->unread_num = ifold->num_msg;

 return IMAP_OK;
}

int
recent_process(command, tag, resp, param)
int command;
char *tag, *resp, *param;
{
char *p;

 strtoul(resp, &p, 10);
 if (*p) 		   {
	display_msg(MSG_WARN, "IMAP", "Invalid RECENT responce");
	return IMAP_ERROR; }

 return IMAP_OK;
}

long
is_literal(str)
char *str;
{
char *s;
u_long lvalue;

 if (*str != '{')
	return -1;
 str++;

 lvalue = strtoul(str, &s, 10);
 if (*s != '}')
	return -1;

 return lvalue;
}

void
skip_literal(len)
u_long len;
{
FILE *nul;
int res;

 if ((nul = fopen("/dev/null", "w")) != NULL) 	{
  if ((res = getdata(NULL, len, imap_in, nul)) < 0) {
    if (res == -2)
	imap_reconnect();
    else
    if (res == -1)
	imap_close(0);				    }
  fclose(nul);
						}

 return;
}

char *
get_imap_string(string, file)
char *string;
FILE *file;
{
char *res;
long llen;
int rres;

  if (!string)
	return NULL;

  if ((llen = is_literal(string)) != -1) {
    if (response)
	free(response);
    response = NULL;	
    plist = NULL;
    res = NULL;

    if (file) {
     if(llen >= MAX_MSG_LEN)   	{
	display_msg(MSG_WARN, "IMAP", "Server responce too long, skipping");
	skip_literal();		}
     else
     if ((rres = getdata(NULL, llen, imap_in, file)) < 0) {
	if (rres == -2)
		imap_reconnect();
	else
	if (rres == -1)        {
		display_msg(MSG_WARN, "IMAP", "Can not receive string");
		imap_close(0); }
	return "";				  	  }
     else
	res = "OK";
	      }
    else      {
     if(llen >= MAX_IMAP_STRING)   {
	display_msg(MSG_WARN, "IMAP", "Server responce too long, skipping");
	skip_literal();		   }

     if ((res = (char *)malloc(llen + 1)) == NULL)
	display_msg(MSG_FATAL, "IMAP", "Malloc failed");
     else
     if ((rres = getdata(res, llen, imap_in, NULL)) < 0) {
	free(res);
	if (rres == -2)
		imap_reconnect();
	else
	if (rres == -1)		{
	display_msg(MSG_WARN, "IMAP", "Can not receive string");
		imap_close(0);   }
	return NULL;				 	 }
		}

    if ((response = getline(NULL, -65535, imap_in)) == NULL) {
	display_msg(MSG_WARN, "IMAP", "Uncomplete FETCH responce");
	imap_close(0);
       	return NULL;		   		 	     }

    if (!*response) {
	imap_reconnect();
	return NULL;}

    plist = response;

    return res;
						}

  while ((*string == '"') || (*string == '\'') || (*string == ' '))
	string++;

  if ((llen = strlen(string)) == 0)
	return strdup("");

  while ((string[llen - 1] == '"') ||
	(string[llen - 1] == '\'') ||
	(string[llen - 1] == ' '))
	llen--;

  if ((llen == 0) || !strcasecmp(string, "NIL"))
	return strdup("");

  if ((res = (char *)malloc(llen + 1)) == NULL) 	{
	display_msg(MSG_FATAL, "IMAP", "Malloc failed");
	return "";					}

  memcpy(res, string, llen);
  res[llen] = '\0';

  return res;
}

char *
imap_string(str)
char *str;
{
static char imapstr[255];

 if (!str)
	return "NIL";

 if (!strpbrk(str, "%*(){ \\\""))
	return str;

 if ((strlen(str) + 2) >= sizeof(imapstr))
	return str;

 sprintf(imapstr, "\"%s\"", str);

 return imapstr;
}

time_t
get_imap_date(str)
char *str;
{
int year, mon, day, hour, min, sec;
int i, offt;
struct tm tmm;  
char month[5];

 if (!str || (strlen(str) < 24))
	return 0L;

 mon = year = hour = min = sec = -1;
 month[0] = '\0';
 offt = 0;

 sscanf(str, "%2d-%3s-%4d %2d:%2d:%2d %d", &day, month, &year, &hour, &min, &sec, &offt);

 for(i=0;i<12;i++) {
    if(!strncasecmp(month, months[i], 3)) {
        mon = i;
        break; 				  }
                   }

 if ((mon == -1) || (year == -1) || (hour == -1))
        return 0L;

 if (offt)
        offt = ((offt / 100) * 60 + (offt % 100)) * 60;

 if (year > 1900)
        year -= 1900;

 if (sec < 0)
	sec = 0L;

 tmm.tm_sec = sec;
 tmm.tm_min = min;
 tmm.tm_hour = hour;
 tmm.tm_mday = day;
 tmm.tm_mon = mon;
 tmm.tm_year = year;

 tmm.tm_yday = 0;
 tmm.tm_wday = 0;
 tmm.tm_isdst = -1;

#if defined(__FreeBSD__) || defined(__NetBSD__) || defined(SunOS)
 tmm.tm_gmtoff = offt;
 tmm.tm_zone = NULL;
#endif
 return mktime(&tmm);
}

int
imap_fetchbody(msg, arg)
struct _mail_msg *msg;
char *arg;
{
char *str, *next;
char typestr[64], subtypestr[64], buf[255];

 if (!*arg)
        return 0;
   
 if (start_plist() == -1)
        return -1;
 
 if ((str = plist_getnext_string(plist, &next)) == NULL) {
	display_msg(MSG_WARN, "IMAP", "Unknown body MIME type");
	return -1;					 }

 strncpy(typestr, (str[0] == '(') ? "MULTIPART" : str, 63);
 typestr[63] = '\0';
 free(str);

 subtypestr[0] = '\0';

 while ((str = plist_getnext_string(NULL, &next)) != NULL)  {
  if ((*subtypestr == '\0') && (*str != '(')) {
	strncpy(subtypestr, str, 63);
	subtypestr[63] = '\0';		      }	
  free(str);
							    }

 sprintf(buf, "%s/%s", typestr, subtypestr);
 add_field(msg, MIME_C_TYPE, buf);

 sprintf(buf, "%02d", MIME_VERS_SUPP);
 buf[2] = buf[1];
 buf[1] = '.';
 buf[3] = '\0'; 
 add_field(msg, MIME_VERSION, buf);

 return 0;
}

int
imap_fetchbodystruct(msg, arg)
struct _mail_msg *msg;
char *arg;
{
 return imap_fetchbody(msg, arg);
}

struct _mail_addr *
imap_fetchaddr(str)
char *str;
{
struct _mail_addr *addr;
char *next, *p;

 if (start_plist() == -1)
	return NULL;

 if ((p = plist_getnext_string(plist, &next)) == NULL) 	{
	display_msg(MSG_WARN, "IMAP", "Invalid address");
 	end_plist();
	return NULL;					}

 if ((addr = (struct _mail_addr *)malloc(sizeof(struct _mail_addr))) == NULL) {
	display_msg(MSG_FATAL, "IMAP", "Malloc failed");
 	end_plist();
	return NULL;							      }

 addr->num = 0;
 addr->comment = NULL;
 addr->next_addr = NULL;
 addr->name = *p ? strdup(p) : NULL;
 addr->addr = NULL;

 free(p);

 if ((p = plist_getnext_string(NULL, &next)) == NULL) 	{
	display_msg(MSG_WARN, "IMAP", "Invalid address");
	discard_address(addr);
 	end_plist();
	free(addr);
	return NULL;					}

 free(p);

 if ((p = plist_getnext_string(NULL, &next)) == NULL) 	{
	display_msg(MSG_WARN, "IMAP", "Invalid address");
	discard_address(addr);
 	end_plist();
	free(addr);
	return NULL;					}

 if (*p)
	addr->addr = strdup(p);

 free(p);

 if ((p = plist_getnext_string(NULL, &next)) == NULL) 	{
	display_msg(MSG_WARN, "IMAP", "Invalid address");
	discard_address(addr);
 	end_plist();
	free(addr);
	return NULL;					}

 if (*p) {
  if (addr->addr) {
    if ((next = (char *)malloc(strlen(addr->addr) + strlen(p) + 3)) == NULL) {
	display_msg(MSG_FATAL, "IMAP", "Malloc failed");
	discard_address(addr);
 	free(p);
	free(addr);
 	end_plist();
	return NULL;							     }

    sprintf(next, "%s@%s", addr->addr, p);
    free(addr->addr);
    addr->addr = next;
		 }
  else
    addr->addr = strdup(p);
	}

 free(p);

 end_plist();

 return addr; 
}

struct _mail_addr *
imap_fetchaddrlist(str)
char *str;
{
struct _mail_addr *addr, *addr1, *addr2;

 if (start_plist() == -1)
	return NULL;

 addr = NULL;
 addr1 = NULL;

 while ((addr2 = imap_fetchaddr(plist)) != NULL)
 {
  if (addr2->addr)   {
   if (addr)
	addr1->next_addr = addr2;
   else
	addr = addr2;
   addr1 = addr2;
   addr->num++;
		     }
  else
	discard_address(addr2);
 }

 return addr; 
}

int
imap_fetchenvelope(msg, arg)
struct _mail_msg *msg;
char *arg;
{
char *str, *next;

 if (!*arg)
	return 0;

 if (start_plist() == -1)
	return -1;

 if ((str = plist_getnext_string(plist, &next)) == NULL) 	{
	display_msg(MSG_WARN, "IMAP", "Invalid envelope date");
	end_plist();
	return -1;					}

 msg->header->snt_time = *str ? get_date(str) : 0L;
 if (msg->header->rcv_time == 0L)
	msg->header->rcv_time = msg->header->snt_time;
 replace_field(msg, "Date", str);

 free(str);

 if ((str = plist_getnext_string(NULL, &next)) == NULL) 	{
	display_msg(MSG_WARN, "IMAP", "Invalid envelope subject");
	end_plist();
	return -1;					}

 if (msg->header->Subject)
	free(msg->header->Subject);
 msg->header->Subject = *str ? strdup(str) : NULL;

 free(str);

 msg->header->From = imap_fetchaddrlist(plist);
 msg->header->Sender = imap_fetchaddrlist(plist);
 imap_fetchaddrlist(plist);
 msg->header->To = imap_fetchaddrlist(plist);
 msg->header->Cc = imap_fetchaddrlist(plist);
 msg->header->Bcc = imap_fetchaddrlist(plist);

 if ((str = plist_getnext_string(NULL, &next)) == NULL)	{
	display_msg(MSG_WARN, "IMAP", "Invalid envelope in-reply-to");
	end_plist();
	return -1;					}

 if (*str)
	replace_field(msg, "In-Reply-To", str);

 free(str);

 if ((str = plist_getnext_string(NULL, &next)) == NULL)	{
	display_msg(MSG_WARN, "IMAP", "Invalid envelope message-id");
	end_plist();
	return -1;					}

 if (*str)
	replace_field(msg, "Message-ID", str);

 free(str);

 end_plist();

 return 0;
}

int
imap_fetchflags(msg, arg)
struct _mail_msg *msg;
char *arg;
{
char *p;

 if (*arg != '(')  {
	display_msg(MSG_WARN, "IMAP", "Invalid FLAGS list");
	return -1; }

 arg++;
 if ((p = strchr(arg, ')')) == NULL) {
	display_msg(MSG_WARN, "IMAP", "Invalid FLAGS list");
	return -1; 		     }
 *p = '\0';

 msg->flags |= UNREAD;
 msg->header->flags |= UNREAD;

 if ((p = strtok(arg, " ")) == NULL)
	return 0;

 msg->flags &= ~ANSWERED;
 msg->header->flags &= ~ANSWERED;

 msg->flags &= ~MARKED;
 msg->header->flags &= ~MARKED;

 msg->flags &= ~DELETED;
 msg->header->flags &= ~DELETED;

 msg->flags &= ~RECENT;
 msg->header->flags &= ~RECENT;

 do {
  if (!strcasecmp(p, "\\Seen")) {
	msg->flags &= ~UNREAD;
	msg->header->flags &= ~UNREAD; }
  else
  if (!strcasecmp(p, "\\Answered"))
  {
	msg->flags |= ANSWERED;
	msg->header->flags |= ANSWERED;
  }
  else
  if (!strcasecmp(p, "\\Flagged"))    {
	msg->flags |= MARKED;
	msg->header->flags |= MARKED; }
  else
  if (!strcasecmp(p, "\\Deleted"))     {
	msg->flags &= ~UNREAD;
	msg->header->flags &= ~UNREAD; }
  else
  if (!strcasecmp(p, "\\Draft"))
  {
  }
  else
  if (!strcasecmp(p, "\\Recent"))    {
	msg->flags |= RECENT;
	msg->header->flags |= RECENT; }
  else
	display_msg(MSG_WARN, "IMAP", "Unknown flag %s", p);

 } while ((p = strtok(NULL, " ")) != NULL);

 return 0;
}

int
imap_fetchidate(msg, arg)
struct _mail_msg *msg;
char *arg;
{
 msg->header->rcv_time = *arg ? get_imap_date(arg) : 0L;
 replace_field(msg, "X-RDate", get_arpa_date(msg->header->rcv_time));

 return 0;
}

int
imap_fetchrfc822(msg, arg)
struct _mail_msg *msg;
char *arg;
{
struct _mail_msg *msg1;
char buf[255];
FILE *mfd;
int oflags;

 if (msg->num == -1) {
  if ((msg->num = get_new_name(fimap)) == -1) 	{
	display_msg(MSG_WARN, "IMAP", "No space in %s", fimap->fold_path);
	return -1;				}
		     }

 sprintf(buf, "%s/%d", fimap->fold_path, msg->num);
 if ((mfd = fopen(buf, "w")) == NULL) 	{
	display_msg(MSG_WARN, "IMAP", "Can not open %s", fimap->fold_path);
	msg->num = -1;
	return -1;			}

 if (strcmp(get_imap_string(arg, mfd), "OK")) 	{
	display_msg(MSG_WARN, "IMAP", "Failed to fetch message from server");
	msg->num = -1;
	fclose(mfd);
	unlink(buf);
	return -1;				}
 fclose(mfd);

 if ((msg1 = get_message(msg->num, fimap)) == NULL) {
	display_msg(MSG_WARN, "IMAP", "Can not parse message");
	msg->num = -1;
	unlink(buf);
	return -1;				    }

 oflags = msg->header->flags;
 discard_message_header(msg);
 msg->header = msg1->header;
 msg1->header = NULL;
 discard_message(msg1);
 msg->header->flags = oflags;
 msg->msg_len = msg1->msg_len;
 msg->flags &= ~H_ONLY;

 return 0;
}

int
imap_fetchrfc822hdr(msg, arg)
struct _mail_msg *msg;
char *arg;
{
struct _mail_msg *msg1;
char buf[255];
FILE *mfd;
int oflags;

 if (msg->num == -1) {
  if ((msg->num = get_new_name(fimap)) == -1) 	{
	display_msg(MSG_WARN, "IMAP", "No space in %s", fimap->fold_path);
	return -1;				}
		     }

 if (msg->flags & H_ONLY)
	sprintf(buf, "%s/%d", fimap->fold_path, msg->num);
 else	{
	if ((oflags = get_new_name(fimap)) == -1)   {
	display_msg(MSG_WARN, "IMAP", "No space in %s", fimap->fold_path);
	return -1;				}
	sprintf(buf, "%s/%d", fimap->fold_path, oflags);
	}

 if ((mfd = fopen(buf, "w")) == NULL) 	{
	display_msg(MSG_WARN, "IMAP", "Can not open %s", fimap->fold_path);
	return -1;			}

 if (strcmp(get_imap_string(arg, mfd), "OK")) 	{
	display_msg(MSG_WARN, "IMAP", "Failed to fetch message header from server");
	fclose(mfd);
	unlink(buf);
	return -1;				}

 if (!(msg->flags & H_ONLY))    {
	msg->print_body(msg, mfd);
	sprintf(buf, "%s/%d", fimap->fold_path, msg->num);
	unlink(buf);
	msg->num = oflags;
				}

 fclose(mfd);

 if ((msg1 = get_message(msg->num, fimap)) == NULL) {
	display_msg(MSG_WARN, "IMAP", "Can not parse message header");
	msg->num = -1;
	unlink(buf);
	return -1;				    }

 oflags = msg->header->flags;
 discard_message_header(msg);
 msg->header = msg1->header;
 msg1->header = NULL;
 discard_message(msg1);
 msg->header->flags = oflags;
 msg->msg_len = msg1->msg_len;

 return 0;
}

int
imap_fetchrfc822size(msg, arg)
struct _mail_msg *msg;
char *arg;
{
char *p;

 msg->msg_len = strtoul(arg, &p, 10);
 if (*p) {
	display_msg(MSG_WARN, "IMAP", "Invalid message size");
	msg->msg_len = 0;
	return -1;
	 }

 replace_field(msg, MIME_C_LENGTH, arg);

 return 0;
}

int
imap_fetchrfc822text(msg, arg)
struct _mail_msg *msg;
char *arg;
{
struct _mail_msg *msg1;
char buf[255];
FILE *mfd;
int oflags;

 if (msg->num == -1) {
  if ((msg->num = get_new_name(fimap)) == -1) 	{
	display_msg(MSG_WARN, "IMAP", "No space in %s", fimap->fold_path);
	return -1;				}
		     }

 if ((mfd = fopen(buf, "w")) == NULL) 	{
	display_msg(MSG_WARN, "IMAP", "Can not open %s", fimap->fold_path);
	return -1;			}

 print_message_header(msg, mfd);

 if (strcmp(get_imap_string(arg, mfd), "OK")) 	{
	display_msg(MSG_WARN, "IMAP", "Failed to fetch message header from server");
	fclose(mfd);
	return -1;				}

 fclose(mfd);

 if ((msg1 = get_message(msg->num, fimap)) == NULL) {
	display_msg(MSG_WARN, "IMAP", "Can not parse message header");
	return -1;				    }

 oflags = msg->header->flags;
 discard_message_header(msg);
 msg->header = msg1->header;
 msg1->header = NULL;
 discard_message(msg1);
 msg->header->flags = oflags;
 msg->msg_len = msg1->msg_len;
 msg->flags &= ~H_ONLY;

 return 0;
}

int
imap_fetchuid(msg, arg)
struct _mail_msg *msg;
char *arg;
{
char *p;

 msg->uid = strtoul(arg, &p, 10);
 if (*p) {
	display_msg(MSG_WARN, "IMAP", "Invalid UID");
	msg->uid = -1;
	return -1;
	 }

 return 0;
}

int
fetch_process(command, tag, resp, param)
int command;
char *tag, *resp, *param;
{
struct _mail_msg *msg, *msg1, *msg2;
char *p, *next, *value;
int mpart, res = 0;

 if (ifold == NULL) 	   {
	display_msg(MSG_WARN, "IMAP", "No folder to fetch into");
	return IMAP_ERROR; }

 res = strtoul(resp, &p, 10);
 if ((res <= 0) || *p) 		{
	display_msg(MSG_WARN, "IMAP", "Invalid message number in FETCH responce");
	return IMAP_ERROR;	}

 display_msg(MSG_STAT, NULL, "Fetching %d", res);

 if ((p = strchr(param, '(')) == NULL) 	{
	display_msg(MSG_WARN, "IMAP", "Invalid FETCH responce");
	return IMAP_ERROR;		}

 p++;

 if ((p = plist_getnext(p, &next)) == NULL) 
	return IMAP_OK;

 if (imsg)
	msg = imsg;
 else 	{
	if ((msg = (struct _mail_msg *)malloc(sizeof(struct _mail_msg))) == NULL) {
	 display_msg(MSG_FATAL, "IMAP", "Malloc failed");
	 return IMAP_FATAL;
  }

	if ((msg->header = (msg_header *)malloc(sizeof(msg_header))) == NULL) {
	 display_msg(MSG_FATAL, "IMAP", "Malloc failed");
	 return IMAP_FATAL;
									      }	

	msg->msg_len = 0;
	msg->flags = H_ONLY|UNREAD;
	msg->msg_body = NULL;
	msg->msg_body_len = 0;
	msg->next = NULL;  
	msg->mime = NULL;
	msg->folder = ifold;
	msg->num = -1;
	msg->uid = -1;
	msg->pdata = NULL;
	msg->ref = NULL;
	msg->refs = 0;
	msg->data = MSG_DAT_EMPTY;

	msg->header->other_fields = NULL;
	msg->header->flags = H_ONLY|UNREAD;
	msg->header->From = NULL;
	msg->header->To = NULL;
	msg->header->Cc = NULL;
	msg->header->Bcc = NULL;
	msg->header->Sender = NULL;
	msg->header->snt_time = 0L;
	msg->header->rcv_time = 0L;
	msg->header->Subject = NULL;

	imap_message(msg);
	}

 do {
  mpart = 0;
  while (mparts[mpart] != NULL)		{
	if (!strcasecmp(mparts[mpart], p))
		break;
	mpart++;			}

  if (elem)
	free(elem);
  elem = NULL;

  if ((mpart != IFETCH_ENVELOPE) && (mpart != IFETCH_BODY) &&
	(mpart != IFETCH_BODYSTRUCTURE)) {
  if ((p = plist_getnext(NULL, &next)) == NULL) {
	display_msg(MSG_WARN, "IMAP", "Invalid FETCH responce");
	if (!imsg)
		discard_message(msg);
	if (response)
		free(response);
	response = NULL;
	return IMAP_ERROR;			}
					 }

  value = NULL;

  switch (mpart) {
   case IFETCH_RFC822:
   case IFETCH_RFC822TEXT:
   case IFETCH_ENVELOPE:
   case IFETCH_BODY:
   case IFETCH_BODYSTRUCTURE:
   break;

   default:
    if ((value = get_imap_string(p, NULL)) == NULL) {
	display_msg(MSG_WARN, "IMAP", "Can not get value of %s", mparts[mpart] ? mparts[mpart] : "UNKNOWN");
	continue;				    }
   break;
		 }

  if (mparts[mpart] == NULL) 	{
	display_msg(MSG_WARN, "IMAP", "Unknown message part name in FETCH responce");
	continue;		}

  switch (mpart) {
   case IFETCH_BODY:
	res = imap_fetchbody(msg, plist);
   break;

   case IFETCH_BODYSTRUCTURE:
	res = imap_fetchbodystruct(msg, plist);
   break;

   case IFETCH_ENVELOPE:
	res = imap_fetchenvelope(msg, plist);
   break;

   case IFETCH_FLAGS:
	res = imap_fetchflags(msg, value);
   break;

   case IFETCH_INTERNALDATE:
	res = imap_fetchidate(msg, value);
   break;

   case IFETCH_RFC822:
	res = imap_fetchrfc822(msg, p);
   break;

   case IFETCH_RFC822HEADER:
	res = imap_fetchrfc822hdr(msg, p);
   break;

   case IFETCH_RFC822SIZE:
	res = imap_fetchrfc822size(msg, value);
   break;

   case IFETCH_RFC822TEXT:
	res = imap_fetchrfc822text(msg, p);
   break;

   case IFETCH_UID:
	res = imap_fetchuid(msg, value);
   break;
		 }

  if (value)
	free(value);
  value = NULL;

  if (elem)
	free(elem);
  elem = NULL;

  if (res < 0)
	display_msg(MSG_WARN, "IMAP", "Failed to fetch message part");

} while ((p = plist_getnext(NULL, &next)) != NULL);

 if (response)
	free(response);
 response = NULL;

 if (elem)
	free(elem);
 elem = NULL;

 if (msg->uid == -1) {
	if (!imsg)
		discard_message(msg);
	return IMAP_OK;
		    }

 if (imsg == NULL) {
  msg1 = msg2 = ifold->messages;
  while (msg1) {
	if (msg1->uid == msg->uid)
		break;
	msg2 = msg1;
	msg1 = msg1->next;
	       }

  if (msg1) {
	if (!(msg1->flags & LOCKED)) {
	msg->num = msg1->num;
	msg->next = msg1->next;
	if (ifold->messages == msg1)
		ifold->messages = msg;
	else
		msg2->next = msg;

	discard_message(msg1);
				     }
	    }
  else	{
	if (ifold->num_msg >= MAX_MSG_IN_FOLDER) {
	  discard_message(msg);
	  return IMAP_OK;			 }

	msg->next = ifold->messages;
	ifold->messages = msg;
	ifold->flags &= ~SORTED;
	ifold->num_msg++;
	if (msg->flags & UNREAD)
		ifold->unread_num++;
	}
		   }

 return IMAP_OK;
}

int
search_process(command, tag, resp, param)
int command;
char *tag, *resp, *param;
{
int scount;
char *p;

 if (search_res)
	free(search_res);
 search_res = NULL;

 if (!param || (strlen(param) < 1))
	return IMAP_OK;

 scount = 1;
 p = param;
 while ((p = strchr(p, ' ')) != NULL) 	{
	scount++;
	while (*p == ' ')
		p++;
					}

 if (scount > MAX_MSG_IN_FOLDER) {
	display_msg(MSG_WARN, "IMAP", "Warning: SEARCH list too long, truncating");
	scount = MAX_MSG_IN_FOLDER;
				 }

 if ((search_res = (u_long *)malloc((scount + 2) * sizeof(u_long))) == NULL) {
	display_msg(MSG_WARN, "IMAP", "malloc failed");
	return IMAP_FATAL;						     }

 search_res[0] = scount;

 p = param;
 scount = 1;
do {
 while (*p == ' ')
	p++;
 search_res[scount] = strtoul(p, NULL, 10);
 scount++;
} while ((p = strchr(p, ' ')) != NULL);

 return IMAP_OK;
}

int
preauth_process(command, tag, resp, param)
int command;
char *tag, *resp, *param;
{
 if ((istate == ISTATE_NOAUTH) || (istate == ISTATE_NOCON))
	istate = ISTATE_AUTH;

 return IMAP_OK;
}

int
stat_process(command, tag, resp, param)
int command;
char *tag, *resp, *param;
{
char fname[MAX_IMAP_FOLD_NAME_LEN], *p, *p1;
struct _mail_folder *folder;
int len;

 if (!param)
	return IMAP_ERROR;

 if (sscanf(param, "%127s", fname) != 1) {
	display_msg(MSG_WARN, "IMAP", "Invalid STATUS responce");
	return IMAP_ERROR;
					 }

 if ((folder = find_imap_folder(fname)) == NULL) {
	display_msg(MSG_WARN, "IMAP", "Unknown folder name in STATUS responce");
	return IMAP_OK;
						 }

 if ((p = strchr(param, ' ')) == NULL)   {
	display_msg(MSG_WARN, "IMAP", "Invalid STATUS responce");
	return IMAP_ERROR;
					 }

 while (*p == ' ')
	p++;

 if (*p != '(')	 {
	display_msg(MSG_WARN, "IMAP", "Invalid STATUS responce");
	return IMAP_ERROR;
		 }

 p++;
 if ((p1 = strchr(p, ')')) == NULL) 	{
	display_msg(MSG_WARN, "IMAP", "Unterminated STATUS list");
	return IMAP_ERROR;
					}

 len = p1 - p;
 if (len < 1)
	return IMAP_OK;

 if (len > 127) {
	display_msg(MSG_WARN, "IMAP", "STATUS list too long");
	return IMAP_ERROR;
		}

 strncpy(fname, p, len);
 fname[len] = '\0';
 if ((p = strtok(fname, " ")) == NULL)
	return IMAP_OK;

 do {
  if (!strcasecmp(p, "MESSAGES")) {
	if ((p = strtok(NULL, " ")) == NULL) 	{
		display_msg(MSG_WARN, "IMAP", "Invalid MESSAGES value in STATUS responce");
		return IMAP_ERROR;
						}
	folder->num_msg = strtoul(p, &p1, 10);
	if (*p1) {
		display_msg(MSG_WARN, "IMAP", "Invalid MESSAGES value in STATUS responce");
		return IMAP_ERROR;
		}
				 }
 else
  if (!strcasecmp(p, "UNSEEN")) {
	if ((p = strtok(NULL, " ")) == NULL) 	{
		display_msg(MSG_WARN, "IMAP", "Invalid UNSEEN value in STATUS responce");
		return IMAP_ERROR;
						}
	folder->unread_num = strtoul(p, &p1, 10);
	if (*p1) {
		display_msg(MSG_WARN, "IMAP", "Invalid UNSEEN value in STATUS responce");
		return IMAP_ERROR;
		}
				 }
 else
  if (!strcasecmp(p, "RECENT")) {
	if ((p = strtok(NULL, " ")) == NULL) 	{
		display_msg(MSG_WARN, "IMAP", "Invalid RECENT value in STATUS responce");
		return IMAP_ERROR;
						}
	strtoul(p, &p1, 10);
	if (*p1) {
		display_msg(MSG_WARN, "IMAP", "Invalid RECENT value in STATUS responce");
		return IMAP_ERROR;
		}
				 }
 else
  if (!strcasecmp(p, "UIDNEXT")) {
	if ((p = strtok(NULL, " ")) == NULL) 	{
		display_msg(MSG_WARN, "IMAP", "Invalid UIDNEXT value in STATUS responce");
		return IMAP_ERROR;
						}
	uid = strtoul(p, &p1, 10);
	if (*p1) 		   {
		uid = -1;
		display_msg(MSG_WARN, "IMAP", "Invalid RECENT value in STATUS responce");
		return IMAP_ERROR; }
				 }
 else
  if (!strcasecmp(p, "UIDVALIDITY")) {
	if ((p = strtok(NULL, " ")) == NULL) 	{
		display_msg(MSG_WARN, "IMAP", "Invalid UIDVALIDITY value in STATUS responce");
		return IMAP_ERROR;
						}
				 }
 else 	{
		display_msg(MSG_WARN, "IMAP", "Unknown STATUS parameter");
		return IMAP_ERROR;
	}

 } while ((p = strtok(NULL, " ")) != NULL);


 return IMAP_OK;
}

int
list_process(command, tag, resp, param)
int command;
char *tag, *resp, *param;
{
struct _mail_folder *fld;
char *p, *foldname;
char flags[127], hdelim[6], fname[255], delimchar;
int len, fflags;

 fflags = 0;

 if (*param != '(')  	{
	display_msg(MSG_WARN, "IMAP", "Missing flags in LIST responce");
	return IMAP_ERROR;
		 	}

 param++;
 if ((p = strchr(param, ')')) == NULL) 	{
	display_msg(MSG_WARN, "IMAP","Unterminated flag list in LIST responce");
	return IMAP_ERROR;
					}

 len = p - param;
 if (len > 126) {
	display_msg(MSG_WARN, "IMAP", "Flag list too long in LIST responce");
	return IMAP_ERROR;
		}

 strncpy(flags, param, len);
 flags[len] = '\0';
 param = p;

 if ((p = strtok(flags, " ")) != NULL)
 do {
  if (!strcasecmp(p, "\\Noinferiors"))
	fflags |= NOINFR;
  else
  if (!strcasecmp(p, "\\Noselect"))
	return IMAP_OK;
  else
  if (!strcasecmp(p, "\\Marked"))
	fflags |= SEARCH;
 } while ((p = strtok(NULL, " ")) != NULL);

 param++;
 while (*param == ' ')
	param++;

 if (sscanf(param, "%3s %127s", hdelim, fname) != 2) 	{
	display_msg(MSG_WARN, "IMAP", "Invalid LIST responce");
	return IMAP_ERROR;				}

 fname[127] = '\0';
 hdelim[3] = '\0';
 foldname = rem_tr_spacequotes(fname);

 p = hdelim;

 if (*p == '"')
	p++;

 if (!strcasecmp(p, "NIL"))
	delimchar = '\0';
 else
	delimchar = *p;

 p = foldname;
 while (*p) {
  if (!isgraph(*p))
	return IMAP_OK;
  p++;
	    }	

 if (imap_h && !strncmp(foldname, imap_h, strlen(imap_h)))
	p = foldname + strlen(imap_h);
 else
	p = foldname;

 /* we do not support nested folders (yet) */
 if (p && (delimchar != '\0') && (delimchar != '.')
			&& (strchr(p, delimchar) != NULL))
	return IMAP_OK;

 if ((fld = find_imap_folder(foldname)) == NULL) {
   if (folders_num >= MAX_FOLDERS)
	return IMAP_OK;

   if (strlen(foldname) > MAX_IMAP_FOLD_NAME_LEN)
	return IMAP_OK;

   display_msg(MSG_STAT, NULL, "Processing: %s", foldname);
   if ((fld = (struct _mail_folder *)malloc(sizeof(mail_folder))) == NULL) {
	display_msg(MSG_FATAL, "IMAP", "Malloc failed");
	return IMAP_FATAL;
									   }

   fld->messages = NULL;
   strcpy(fld->fold_path, foldname);
   fld->flags = 0;
   fld->num_msg = 0;
   fld->unread_num = 0;
   imap_folder(fld);

   mailbox[folders_num] = fld;
   folders_num++;

   if (folders_num >= MAX_FOLDERS)
	display_msg(MSG_WARN, "IMAP", "Too many folders");
					      }
 
 fld->flags |= fflags;
 fld->hdelim = delimchar;

 return IMAP_OK;
}

int
unk_process(command, tag, resp, param)
int command;
char *tag, *resp, *param;
{
 display_msg(MSG_WARN, "IMAP", "unknown responce [%s] from server", resp);
 return IMAP_ERROR;
}

int
token_comp(token, string)
char *token, *string;
{
char *s;

 if (!token || !string)
	return 0;

 if (!strcmp(token, "$all"))
	return 1;

 if (!strcmp(token, "$num")) 	{
  strtoul(string, &s, 10);
  return *s ? 0 : 1;
				}

 if (*token == '$') {
	token++;
	return !strncasecmp(token, string, strlen(token));
		     }

 return !strcasecmp(token, string);
}

#if __STDC__ || defined(USE_STDARG)
int imap_command(int command, char *fmt, ...)
#else
int imap_command(command, fmt, va_list)
int command;
char *fmt;
va_dcl
#endif
{
static int ntag = 0;
va_list ap;
char buf[255], tag[6], *p;
char rtag[16], rresp[16], rparam[64], rmsg[64];
int rnum, pstatus, comflags, res, lines;

#if __STDC__
 va_start(ap, fmt);
#else
 va_start(ap);
#endif

 if (!imap_isconnected())
	return IMAP_ERROR;

 fl_set_timer(imap_timer, IMAP_TIMER);

 comflags = command & 0xe0;
 command &= 0x1f;

 if (command == ICOM_NOCOM)  	{
	strcpy(tag, "*");
	goto rwait;
				}

 if (istate == ISTATE_NOCON)
	return IMAP_ERROR;

 if ((command < 0) || (command >= ICOM_MAXCOMM)) {
	display_msg(MSG_WARN, "IMAP", "unknown command");
	return IMAP_ERROR;
						 }
 ntag++;
 if (ntag > 999)
	ntag = 0;
 sprintf(tag, "A%03d", ntag);
 sprintf(buf, "%s %s", tag, imap_commands[command]);

 p = NULL;
 if (fmt) {
	p = buf + strlen(buf);
	*p = ' ';
	p++;
	vsprintf(p, fmt, ap);
	  }

 if (command != ICOM_LOGIN)
        display_msg(MSG_LOG, "imap", "-> %s %s", imap_commands[command], p ? p : "");
    else	{
	va_start(ap, fmt);
	p = va_arg(ap, char *);
        display_msg(MSG_LOG, "imap", "-> LOGIN %s ******", p);
	va_end(ap);
		}

 if ((res = putline(buf, imap_out)) < 0) {
   if (res == -2)
	imap_reconnect();
   else
   if (res == -1)
	imap_close(0);
   return IMAP_ERROR;	   		 }

rwait:
 if (response)
	free(response);
 response = NULL;
 pstatus = IMAP_OK;

 if (comflags & ICOM_DONTWAIT)
	return IMAP_OK;

 if ((response = getline(NULL, -65535, imap_in)) == NULL) {
	imap_close(0);
	return IMAP_ERROR;				  }

 lines = 0;
 do {

 lines++;
 if (!*response) {
   free(response);
   response = NULL;
   imap_reconnect();
   return IMAP_ERROR;
		 }

 rparam[0] = '\0';
 if (sscanf(response, "%s %s %s", rtag, rresp, rparam) < 2) {
   if (!(comflags & ICOM_SILENT))
	display_msg(MSG_WARN, "IMAP", "Invalid reponse from server");
   free(response);
   response = NULL;
   imap_reconnect();
   return IMAP_ERROR;					    }

 if (comflags & ICOM_ABORT) {
  free(response);
  response = NULL;
  if ((*rtag == 'A') && isdigit(rtag[1]))
	return IMAP_ERROR;

  continue;		   }

 if ((p = strchr(response, ' ')) != NULL) {
	p++;
	if (!strcmp(tag, rtag) ||
		(!(comflags & ICOM_DONTLOG) && (lines < 15)))
		display_msg(MSG_LOG, "imap", "<- %.64s", p);
 	if ((p = strchr(p, ' ')) != NULL) {
		p++;
		strncpy(rmsg, p, 63);
		rmsg[63] = '\0';	  }
		  			}

 rnum = 0;
 while (imap_responces[rnum].command != ICOM_NOCOM) {
  if ((imap_responces[rnum].command == ICOM_MATCHALL) ||
	(imap_responces[rnum].command == command)) {
   if (token_comp(imap_responces[rnum].tag, rtag) &&
	token_comp(imap_responces[rnum].resp, rresp) &&
	token_comp(imap_responces[rnum].param, p)) 	{
     if (imap_responces[rnum].process == unk_process)
	comflags |= (ICOM_ABORT|ICOM_SILENT|ICOM_DONTLOG);
     if (pstatus == IMAP_OK)
	pstatus = imap_responces[rnum].process(command, rtag, rresp, p);
     else
	imap_responces[rnum].process(command, rtag, rresp, p);

     if (pstatus == IMAP_FATAL) {
	free(response);
	response = NULL;
	return pstatus;		}
     break;
							}
						   }
  rnum++;
					    }

  if (!strcmp(tag, rtag))
	break;

  if (response)
  	free(response);
  response = NULL;
 } while ((response = getline(NULL, -65535, imap_in)) != NULL);

 if (!response) {
	imap_close(0);
	return IMAP_ERROR;
		}
 else 	{
	free(response);
	response = NULL;
	}

 if (pstatus != IMAP_OK)
	return pstatus;

 if (!strcasecmp(rresp, "OK"))
	return IMAP_OK;

 if (!strcasecmp(rresp, "NO")) 	{
   if (!(comflags & ICOM_SILENT))
	display_msg(MSG_WARN, "IMAP", "%s", rmsg);
	return IMAP_NO;		}

 if (!strcasecmp(rresp, "BAD")) 	{
   if (!(comflags & ICOM_SILENT))
	display_msg(MSG_WARN, "IMAP error", "%s", rmsg);
	return IMAP_BAD;		}

 if (!(comflags & ICOM_SILENT))
   display_msg(MSG_WARN, "Invalid response from server", "%s", rmsg);

 return IMAP_ERROR;
}

int
imap_isconnected()
{
 return imapsock > 0 ? 1 : 0;
}

void
imap_timer_cb()
{
 if (!imap_isconnected()) {
	fl_set_timer(imap_timer, 0);
	return;		  }

 imap_command(ICOM_NOOP, NULL);
 fl_set_timer(imap_timer, IMAP_TIMER);
}

int
imap_connect()
{
 char buf[255];

 if (imapsock != -1)
	return 0;

 strcpy(buf, b_getcfg_str(conf_name, "imaphost", "127.0.0.1"));
 if ((imapsock = host_connect(buf, b_getcfg_str(conf_name, "imaport", "143"), NULL)) == -1) {
	display_msg(MSG_WARN, "IMAP connect", "Failed to connect to IMAP server");
	return -1;
	    }

 if ((imap_in = fdopen(imapsock, "r+")) == NULL) {
        display_msg(MSG_WARN, "IMAP connect", "fdopen failed");
        imap_close(0);
        return -1;
        }

 imap_out = imap_in;

 if (imap_command(ICOM_NOCOM, NULL) != IMAP_OK) {
        display_msg(MSG_WARN, "IMAP connect", "no greeting from IMAP server");
        imap_close(0);
        return -1;
        }

 return 0;
}

int
imap_login()
{
int res;
char imapuser[32];

 if (!imap_isconnected())
	return -1;

 if (!supress_errors)   {
  if ((strlen(b_getcfg_str(conf_name, "imapwd", "")) < 1) &&
	b_getcfg_int(conf_name,"imapstorpwd", 0) == 0)
        imap_account();
                        }

 strncpy(imapuser, b_getcfg_str(conf_name, "imapuser", user_n), 31);
 if ((res = imap_command(ICOM_LOGIN, "%s %s", imapuser,
	imap_string(b_getcfg_str(conf_name, "imapwd", "")))) != IMAP_OK) {
  if (res != IMAP_NO) {
	display_msg(MSG_WARN, "IMAP login", "login failed");
        imap_close(0);
	return -1;
		     							 }

  imap_account();
  strncpy(imapuser, b_getcfg_str(conf_name, "imapuser", user_n), 31);
  if ((res = imap_command(ICOM_LOGIN, "%s %s", imapuser,
	imap_string(b_getcfg_str(conf_name, "imapwd", "")))) != IMAP_OK) {
	display_msg(MSG_WARN, "IMAP login", "login failed");
        imap_close(0);
	return -1;					    		 }
								}

 return 0;
}

int
imap_init()
{
int i, res;
char llist[255], buf[255], *p, *p1;

 if (imapsock != -1) 	{
	display_msg(MSG_WARN, "IMAP", "Already connected");
	return -1;
			}

 imap_h = NULL;
 ifold = NULL;
 imsg = NULL;
 iinbox = NULL;
 itrash = NULL;
 istate = ISTATE_NOCON;
 if (search_res)
	free(search_res);
 search_res = NULL;
 discard_imap_folders();

 if (imap_connect() == -1)
	return -1;

 istate = ISTATE_NOAUTH;

 if (imap_command(ICOM_CAPABILITY, NULL) != IMAP_OK) {
        imap_close(0);
        return -1;
        }

 if (istate != ISTATE_AUTH) {
  if (imap_login() == -1)
	return -1;	    }

 istate = ISTATE_AUTH;

 strncpy(llist, b_getcfg_str(conf_name, "imapextra", "\"\""), 254);
 llist[254] = '\0';
 p = llist;
 do {
  while (*p == ' ')
	p++;
  if ((p1 = strchr(p, ' ')) != NULL) {
	*p1 = '\0';
	p1++;			     }

  if (strlen(p) < 1)
	p = "\"\"";

  strcpy(buf, p);
  imap_h = rem_tr_spacequotes(buf);
  if ((res = imap_command(ICOM_LIST, "%s \"*\"", p)) != IMAP_OK) 	{
	display_msg(MSG_WARN, "IMAP", "Failed to obtain folder list");
        imap_close(0);
	return -1;
									}
  imap_h = NULL;
  p = p1;
 } while (p);

 if ((iinbox = find_imap_folder(IMAP_INBOX)) == NULL) 	{
  if ((res = imap_command(ICOM_LIST, "\"\" %s", IMAP_INBOX)) != IMAP_OK) {
	display_msg(MSG_WARN, "IMAP", "Failed to find INBOX folder");
        imap_close(0);
	return -1;							 }
							}

 if ((iinbox = find_imap_folder(IMAP_INBOX)) == NULL) 	{
  if (imap_command(ICOM_CREATE, "%s", IMAP_INBOX) == IMAP_OK) 	{
   if ((res = imap_command(ICOM_LIST, "\"\" %s", IMAP_INBOX)) != IMAP_OK) {
	display_msg(MSG_WARN, "IMAP", "Failed to find INBOX folder");
        imap_close(0);
	return -1;							 }
								}
  else
	display_msg(MSG_WARN, "IMAP", "Failed to create INBOX folder");
							}

 if ((iinbox = find_imap_folder(IMAP_INBOX)) == NULL)
	display_msg(MSG_WARN, "IMAP", "Warning: No INBOX on server");

 if (b_getcfg_int(conf_name, "imapusetrash", 0)) {
  if ((itrash = find_imap_folder(IMAP_TRASH)) == NULL) 	{
   if ((res = imap_command(ICOM_LIST, "\"\" %s", IMAP_TRASH)) != IMAP_OK) {
	display_msg(MSG_WARN, "IMAP", "Failed to find TRASH folder");
        imap_close(0);
	return -1;							  }
							}

  if (!itrash && ((itrash = find_imap_folder(IMAP_TRASH)) == NULL)) {
	if ((itrash = create_imap_folder(NULL, IMAP_TRASH)) == NULL)
		display_msg(MSG_WARN, "IMAP", "Warning: Failed to create trash folder");
	else
		itrash->flags |= NOTRASH;
								    }
  else
	itrash->flags |= NOTRASH;
						 }

 folder_sort &= ~FLD_SORTED;

 if (icap & ICAP_STATUS) {
  for ( i = 0; i < folders_num; i++)    {
	if (!(mailbox[i]->flags & F_IMAP))
		continue;

	display_msg(MSG_STAT, NULL, "Checking %s", mailbox[i]->fold_path);
	if (imap_command(ICOM_STATUS, "%s (MESSAGES UNSEEN)", imap_string(mailbox[i]->fold_path)) != IMAP_OK) {
		display_msg(MSG_WARN, "IMAP", "STATUS failed");
		break;
		 }
					}
			 }
 else	{
  for ( i = 0; i < folders_num; i++)    {
	if (!(mailbox[i]->flags & F_IMAP))
		continue;

	ifold = mailbox[i];
	display_msg(MSG_STAT, NULL, "Checking %s", mailbox[i]->fold_path);
	if (imap_command(ICOM_EXAMINE, "%s", imap_string(mailbox[i]->fold_path)) != IMAP_OK) 		 {
		display_msg(MSG_WARN, "IMAP", "EXAMINE failed");
		break;
		 }
					}

  if (iinbox)
	iinbox->rescan(iinbox);
	}

 ifold = NULL;
 if (!fimap) {
 if ((fimap = create_mh_folder(NULL, FIMAP)) == NULL) 	{
	display_msg(MSG_WARN, "IMAP INIT", "Can not open folder %s", FIMAP);
	imap_close(0);
	return -1;					}

 fimap->empty(fimap);
 if (fimap->open(fimap, 0) == -1)	{
	display_msg(MSG_WARN, "IMAP INIT", "Can not open folder %s", FIMAP);
	imap_close(0);
	return -1;		}

	    }
 return 0;
}

void
imap_disconnect()
{
 if (imapsock == -1)
	return;

 host_disconnect(&imapsock);
 if (imap_in) 	{
	fclose(imap_in);
	imap_in = NULL;
	imap_out = NULL;
		}

 return;
}

int
imap_reconnect()
{
 imap_disconnect();

 if (response)
	free(response);
 response = NULL;
 plist = NULL;
 elem = NULL;
 if (search_res)
	free(search_res);
 search_res = NULL;

 if (istate == ISTATE_NOCON)
	return 0;

 if (imap_connect() != 0) {
	display_msg(MSG_WARN, "IMAP reconnect", "Failed to reconnect");
	return -1;	  }


 if (istate == ISTATE_NOAUTH)
	return 0;

 if (imap_login() != 0)   {
	display_msg(MSG_WARN, "IMAP reconnect", "Failed to log in");
	return -1;	  }

 if (ifold) {
  if (imap_command(ICOM_SELECT, "%s", imap_string(ifold->fold_path)) != IMAP_OK) {
	ifold = NULL;
	return -1;					    	    }
	    }

  return 0;
}

void
imap_close(bygui)
int bygui;
{
 fl_set_timer(imap_timer, 0);

 if (imapsock == -1)
	return;

 if (bygui) {
  if (ifold)
	ifold->close(ifold);

  if (istate != ISTATE_NOCON)
	imap_command(ICOM_LOGOUT, NULL);
	     }

 imap_disconnect();

 imap_h = NULL;
 icap = 0;
 istate = 0;
 ifold = NULL;
 imsg = NULL;
 iinbox = NULL;
 itrash = NULL;
 if (search_res)
	free(search_res);
 search_res = NULL;
 discard_imap_folders();

 if (!bygui) {
  current_folder = inbox;
  redraw_folders();
  update_title();
  fl_replace_menu_item(menu_misc, 6, "IMAP connect");
	     }

 return;
}

struct _mail_folder *
imap_folder_switch(folder)
struct _mail_folder *folder;
{
struct _mail_folder *ofold;

 if (folder == NULL) {
	if (ifold)
		return ifold;

	imap_command(ICOM_CLOSE, NULL);
	return NULL; }

 if (folder == ifold)
	return ifold;

 ofold = ifold;
 ifold = folder;
 if (imap_command(ICOM_SELECT, "%s", imap_string(folder->fold_path)) != IMAP_OK) {
	ifold = ofold;
	return NULL;						    }

 return ofold ? ofold : folder;
}

char *
get_imap_folder_name(folder)
struct _mail_folder *folder;
{
/*
char *p;

 if (folder->hdelim == '\0')
	return folder->fold_path;

 if ((p = strrchr(folder->fold_path, folder->hdelim)) == NULL)
	return folder->fold_path;

 p++;
 return p;
*/

 return folder->fold_path;
}

char *
get_imap_folder_path(folder)
struct _mail_folder *folder;
{
static char fpath[MAX_IMAP_FOLD_NAME_LEN];
char *p;

 if (folder->hdelim == '\0')
	return NULL;

 strcpy(fpath, folder->fold_path);

 if ((p = strrchr(fpath, folder->hdelim)) == NULL)
	return NULL;

 *p = '\0';

 return fpath;
}

struct _mail_folder *
get_imap_trash(msg)
struct _mail_msg *msg;
{
char buf[255], *p;
struct _mail_folder *trfold;

 if (!msg)
	return NULL;

 if ((msg->folder->flags & NOTRASH) || (msg->flags & DELPERM) ||
	!b_getcfg_int(conf_name, "imapusetrash", 0)) 
	return NULL;

 if ((p = get_imap_folder_domain(msg->folder)) != NULL)
	sprintf(buf, "%s%c%s", p, msg->folder->hdelim, IMAP_TRASH);
 else
	return itrash;

 if ((trfold = find_imap_folder(buf)) != NULL) 	{
	trfold->flags |= NOTRASH;
	if (trfold == msg->folder)
		return NULL;
	return trfold;				}

 if (imap_command(ICOM_LIST, "\"\" %s", buf) != IMAP_OK) {
        display_msg(MSG_WARN, "IMAP", "Failed to list %s folder", buf);
        return NULL;                                     }

 if ((trfold = find_imap_folder(buf)) == NULL) 	{
        if ((trfold = create_imap_folder(NULL, buf)) == NULL) 	{
            if (!display_msg(MSG_QUEST, "IMAP", "Failed to create %s folder\ncontinue without saving?", buf))
		return NULL;
	    msg->folder->flags |= NOTRASH;	   		}
	else
	    redraw_fld_win();
						}
 trfold->flags |= NOTRASH;

 if (trfold == msg->folder)
	return NULL;

 return trfold;
}

int
delete_imap_message_range(msg)
struct _mail_msg *msg;
{
char buf[255];
int mask;
int top, bottom, i, deleted;
struct _mail_folder *trfold, *ofold;
struct _mail_msg *msg1;

 if (!imap_isconnected())
	return -1;

 if (!msg || !msg->folder)
        return -1;
 
 if (msg->folder && msg->folder->flags & FRONLY) 	{
	display_msg(MSG_WARN, "IMAP", "READ-ONLY folder");
	msg->flags &= ~DELETED;
	return -1;					}

 if (msg->flags & LOCKED) {
	msg->flags &= ~DELETED;
	return -1;	  }

 mask = DELETED;
 if (msg->flags & DELPERM)
	mask |= DELPERM;

 expand_uid_range(msg->folder, msg, mask, 0, &bottom, &top);
 if (top == bottom)
	return delete_imap_message(msg);

 for (i = bottom; i <= top; i++)	{
  if ((msg1 = get_msg_by_uid(msg->folder, i)) != NULL)
	msg1->flags &= ~DELETED;	}

 if ((ofold = imap_folder_switch(msg->folder)) == NULL)
	return -1;

 imsg = NULL;

 if ((trfold = get_imap_trash(msg)) != NULL) 	{
    if (imap_command(ICOM_UIDCOPY, "%d:%d %s", bottom, top, imap_string(trfold->fold_path))
							  != IMAP_OK)  {
      display_msg(MSG_WARN, "IMAP", "Can not copy messages to %s", trfold->fold_path);
      imap_folder_switch(ofold);
      return -1;						       }
						}

 if (imap_command(ICOM_UIDSTORE, "%d:%d FLAGS.SILENT (\\Deleted)", bottom, top)
			!= IMAP_OK) {
	imap_folder_switch(ofold);
	return -1;		    }

 imap_folder_switch(ofold);

 deleted = 0;
 ofold = msg->folder;
 for (i = bottom; i <= top; i++)	{
  if ((msg1 = get_msg_by_uid(ofold, i)) == NULL)
	continue;

  msg1->flags &= ~DELETED;
  if (msg1->num > 0)   {
	sprintf(buf, "%s/%d", fimap->fold_path, msg1->num);
	unlink(buf);   }

  if ( ((folder_sort & 0x0f) == BY_MSGNUM) ||
        ( ( (folder_sort & 0x0f) == BY_UNREAD) &&
                (msg1->flags & UNREAD) ) )
   folder_sort &= ~FLD_SORTED;

  if (trfold) 	{
    trfold->num_msg++;

    if (msg->flags & UNREAD)
	trfold->unread_num++;
		}

  unlink_message(msg1);
  discard_message(msg1);
  deleted++;
					}

 return deleted;
}

int
delete_imap_message(msg)
struct _mail_msg *msg;
{
char buf[255];
struct _mail_folder *trfold, *ofold;

 if (!imap_isconnected())
	return -1;

 if (!msg || !msg->folder)
        return -1;
 
 msg->flags &= ~DELETED;

 if (msg->folder && msg->folder->flags & FRONLY) 	{
	display_msg(MSG_WARN, "IMAP", "READ-ONLY folder");
	return -1;					}

 if (msg->flags & LOCKED)
        return -1;
          
 imsg = msg;

 if ((ofold = imap_folder_switch(msg->folder)) == NULL)
	return -1;

 if ((trfold = get_imap_trash(msg)) != NULL)	{
    if (imap_command(ICOM_UIDCOPY, "%d %s", msg->uid, imap_string(trfold->fold_path))
							  != IMAP_OK)  {
      display_msg(MSG_WARN, "IMAP", "Can not copy message to %s", trfold->fold_path);
      imsg = NULL;
      imap_folder_switch(ofold);
      return -1;						       }

    trfold->num_msg++;

    if (msg->flags & UNREAD)
	trfold->unread_num++;
						}

 if (imap_command(ICOM_UIDSTORE, "%d FLAGS.SILENT (\\Deleted)", msg->uid)
			!= IMAP_OK) {
	imsg = NULL;
	imap_folder_switch(ofold);
	return -1;		    }

 imap_folder_switch(ofold);

 imsg = NULL;

 if (msg->num > 0)   {
	sprintf(buf, "%s/%d", fimap->fold_path, msg->num);
	unlink(buf); }

 if ( ((folder_sort & 0x0f) == BY_MSGNUM) ||
        ( ( (folder_sort & 0x0f) == BY_UNREAD) &&
                (msg->flags & UNREAD) ) )
   folder_sort &= ~FLD_SORTED;

 unlink_message(msg);
 discard_message(msg);

 return 0;
}

char *
get_imap_msg_file(msg)
struct _mail_msg *msg;
{
static char buf[255];
struct _mail_folder *ofold;

 if (!imap_isconnected())
	return NULL;

 if ((msg->num == -1) || (msg->flags & H_ONLY)) {
  if ((ofold = imap_folder_switch(msg->folder)) == NULL)
	return NULL;

  imsg = msg;
  if (imap_command(ICOM_UIDFETCH, "%d (RFC822)", msg->uid) != IMAP_OK) 	{
	imap_folder_switch(ofold);
	imsg = NULL;
	return NULL;							}
  imsg = NULL;

  imap_folder_switch(ofold);
						}

 sprintf(buf, "%s/%d", fimap->fold_path, msg->num);

 return buf;
}

char *
get_imap_minus_flags(msg)
struct _mail_msg *msg;
{
static char flags[64];
int num;

 num = 0;
 flags[0] = '\0';
 if ((msg->flags & UNREAD) && !(msg->header->flags & UNREAD)) 	{
	strcat(flags, "\\Seen");
	num++;							}

 if (!(msg->flags & ANSWERED) && (msg->header->flags & ANSWERED)) 	{
	strcat(flags, "\\Answered");
	num++;								}

 if (!(msg->flags & MARKED) && (msg->header->flags & MARKED)) 	{
	strcat(flags, "\\Flagged");
	num++;							}

 if (!(msg->flags & DELETED) && (msg->header->flags & DELETED)) 	{
	strcat(flags, "\\Flagged");
	num++;								}

 return num ? flags : NULL;
}

char *
get_imap_plus_flags(msg)
struct _mail_msg *msg;
{
static char flags[64];
int num;

 num = 0;
 flags[0] = '\0';
 if ((msg->flags & UNREAD) && (msg->header->flags & UNREAD)) 	{
	strcat(flags, "\\Seen");
	num++;							}

 if ((msg->flags & ANSWERED) && !(msg->header->flags & ANSWERED)) 	{
	strcat(flags, "\\Answered");
	num++;								}

 if ((msg->flags & MARKED) && !(msg->header->flags & MARKED)) 	{
	strcat(flags, "\\Flagged");
	num++;							}

 if ((msg->flags & DELETED) && !(msg->header->flags & DELETED)) 	{
	strcat(flags, "\\Flagged");
	num++;								}

 return num ? flags : NULL;
}

char *
get_imap_flags(msg)
struct _mail_msg *msg;
{
static char flags[64];
int num;

 num = 0;
 flags[0] = '\0';
 if (!(msg->flags & UNREAD)) 	{
	strcat(flags, "\\Seen");
	num++;			}

 if (msg->flags & ANSWERED)	{
	strcat(flags, num ? " \\Answered" : "\\Answered");
	num++;			}

 if (msg->flags & MARKED)	{
	strcat(flags, num ? " \\Flagged" : "\\Flagged");
	num++;			}

 if (msg->flags & DELETED)	{
	strcat(flags, num ? " \\Deleted" : "\\Deleted");
	num++;			}

 return num ? flags : NULL;
}

void
update_imap_message_range(msg) 
struct _mail_msg *msg;
{
struct _mail_folder *ofold;
struct _mail_msg *msg1;
char *flags;
int checkmask, plusmask, minusmask;
int bottom, top, i;

 if (!imap_isconnected())
	return;

/*
 if (msg->folder->flags & FRONLY) 	{
	msg->header->flags = msg->flags;
	return;				}
*/

 checkmask = (UNREAD|ANSWERED|MARKED|DELETED);
 if ((msg->flags & checkmask) == (msg->header->flags & checkmask))
	return;

 plusmask = (msg->flags & checkmask) & ~(msg->header->flags & checkmask);
 minusmask = ~(msg->flags & checkmask) & (msg->header->flags & checkmask);
 expand_uid_range(msg->folder, msg, plusmask, minusmask, &bottom, &top);

 if (bottom == top)	{
	update_imap_message(msg);
	return;		}

 if ((ofold = imap_folder_switch(msg->folder)) == NULL)
	return;

 imsg = NULL;

 if ((flags = get_imap_flags(msg)) != NULL) {
  if (imap_command(ICOM_UIDSTORE, "%d:%d FLAGS.SILENT (%s)",
	bottom, top, flags) == IMAP_OK)
	msg->header->flags = msg->flags;
					    }
 else	{
  flags = get_imap_minus_flags(msg);
  if (imap_command(ICOM_UIDSTORE, "%d:%d -FLAGS.SILENT (%s)",
	bottom, top, flags ? flags : "\\Seen") == IMAP_OK)
	msg->header->flags = msg->flags;
	}

 imap_folder_switch(ofold);

 ofold = msg->folder;
 for (i = bottom; i <= top; i++)	{
  if ((msg1 = get_msg_by_uid(ofold, i)) == NULL)
	continue;
  msg1->header->flags = msg1->flags;
					}

 return;
}

int
update_imap_message(msg)  
struct _mail_msg *msg;
{
struct _mail_folder *ofold;
char *flags;

 if (!imap_isconnected())
	return -1;

/*
 if (msg->folder->flags & FRONLY) 	{
	msg->header->flags = msg->flags;
	return -1;			}
*/

 if ( ((msg->flags & UNREAD) == (msg->header->flags & UNREAD)) &&
	((msg->flags & ANSWERED) == (msg->header->flags & ANSWERED)) &&
	((msg->flags & MARKED) == (msg->header->flags & MARKED)) &&
	((msg->flags & DELETED) == (msg->header->flags & DELETED)) )
	return 0;

 if ((ofold = imap_folder_switch(msg->folder)) == NULL)
	return -1;

 imsg = msg;
 if ((flags = get_imap_flags(msg)) != NULL) {
  if (imap_command(ICOM_UIDSTORE, "%d FLAGS.SILENT (%s)",
	msg->uid, flags) == IMAP_OK)
	msg->header->flags = msg->flags;
					    }
 else	{
  flags = get_imap_minus_flags(msg);
  if (imap_command(ICOM_UIDSTORE, "%d -FLAGS.SILENT (%s)",
	msg->uid, flags ? flags : "\\Seen") == IMAP_OK)
	msg->header->flags = msg->flags;
	}

 imsg = NULL;
 imap_folder_switch(ofold);

 return 0;
}

void
imap_message(msg)
struct _mail_msg *msg;
{
 msg->delete = delete_imap_message;
 msg->print = print_message;
 msg->print_body = print_message_body;
 msg->get_text = get_message_text;
 msg->free_text = free_message_text;
 msg->get_file = get_imap_msg_file;
 msg->update = update_imap_message; 
 msg->flags |= M_IMAP;
}

int
open_imap_folder(folder, flags)
struct _mail_folder *folder;
int flags;
{
struct _mail_msg *msg;
int i, k;

 if (!imap_isconnected())
	return -1;

 if (!folder)
	return -1;

 if (folder->flags & OPENED)
	folder->close(folder);

 ifold = folder;
 if (imap_command(ICOM_SELECT, "%s", imap_string(folder->fold_path)) != IMAP_OK) {
	display_msg(MSG_WARN, "IMAP", "Can not SELECT folder");
	ifold = NULL;
	return -1;						    }

 istate = ISTATE_SELECT;

 i = folder->num_msg;
 folder->num_msg = 0;
 folder->unread_num = 0;

 if (i) {
  if (flags & FOPEN_UNREAD) {
	if (imap_command(ICOM_SEARCH, "UNSEEN") != IMAP_OK) {
	display_msg(MSG_WARN, "IMAP", "Failed to find unseen messages");
	ifold = NULL;
	folder->close(folder);
	return -1;		       			    }

	if (search_res && (search_res[0] > 0))		    {
	 for (k = 1; k <= search_res[0]; k++)	{
	  if (imap_command(ICOM_FETCH, "%d (INTERNALDATE RFC822.SIZE ENVELOPE FLAGS BODY UID)", search_res[k]) != IMAP_OK) {
	   display_msg(MSG_WARN, "IMAP", "Failed to fetch %d", search_res[k]);
	   ifold = NULL;
	   folder->close(folder);
	   return -1;		   	   }
						}
	 free(search_res);
	 search_res = NULL;
							    }
			    }
  else		{
	if (imap_command(ICOM_FETCH, "1:%d (INTERNALDATE RFC822.SIZE ENVELOPE FLAGS BODY UID)", i) != IMAP_OK) {
	display_msg(MSG_WARN, "IMAP", "Failed to scan folder");
	ifold = NULL;
	folder->close(folder);
	return -1;		       }
		}
	}

 if (folder->num_msg >= MAX_MSG_IN_FOLDER)
	display_msg(MSG_WARN, "IMAP", "Too many messages in folder");

 folder->num_msg = 0;
 folder->unread_num = 0;
 msg = folder->messages;
 while (msg) 	{
  folder->num_msg++;
  if (msg->flags & UNREAD)
	folder->unread_num++;
  msg = msg->next;
		}

 if (flags & FOPEN_UNREAD)
	folder->num_msg = i;

 folder->flags |= OPENED;
 sort_folder(folder);

 return 0;
}

int
rescan_imap_folder(folder)
struct _mail_folder *folder;
{
struct _mail_folder *ofold;

 if (!imap_isconnected())
	return -1;

 if (icap & ICAP_STATUS) {
   if (ifold)
	imap_command(ICOM_CLOSE, NULL);

   if (imap_command(ICOM_STATUS, "%s (MESSAGES UNSEEN)", imap_string(folder->fold_path)) != IMAP_OK) {
   	if (ifold)
		imap_command(ICOM_SELECT, "%s", imap_string(ifold->fold_path));
	return -1;
	}
   if (ifold)
	imap_command(ICOM_SELECT, "%s", imap_string(ifold->fold_path));
   return 0;		 }

 if ((ofold = imap_folder_switch(folder)) == NULL)
	return -1;

 if (imap_command(ICOM_SEARCH, "UNSEEN") != IMAP_OK) {
	imap_folder_switch(ofold);
	return -1;				     }

 imap_folder_switch(ofold);

 if (search_res) {
  folder->unread_num = search_res[0];
  free(search_res);
  search_res = NULL;
		 }

 return 0;
}

void
close_imap_folder(folder)
struct _mail_folder *folder;  
{
struct _mail_msg *msg, *msg1, *msg_lock;
char buf[255];

 if (!imap_isconnected())
	return;

 if (!folder)
	return;

 if (!(folder->flags & FRONLY))
	folder->update(folder);

 msg = folder->messages;
 msg_lock = NULL;
 while(msg) {
	msg->update(msg);
        msg1 = msg->next;
        if (msg->flags & LOCKED) {
          msg->next = msg_lock;
          msg_lock = msg; }
        else {
	  if (msg->num > 0)  {
		sprintf(buf, "%s/%d", fimap->fold_path, msg->num);
		unlink(buf); }

          discard_message(msg);
	     }

        msg = msg1;
            }

 folder->messages = msg_lock;

 if ((folder->flags & OPENED) && (istate == ISTATE_SELECT)) {
  if (imap_command(ICOM_CLOSE, NULL) != IMAP_OK)
	display_msg(MSG_WARN, "IMAP", "Can not CLOSE folder");
			      				    }

 folder->flags &= ~OPENED;
 folder->flags &= ~SORTED;
 folder->flags &= ~SEARCH;

 ifold = NULL;

 return;
}

void
empty_imap_folder(folder)
struct _mail_folder *folder;  
{
char buf[255];
struct _mail_folder *ofold;
struct _mail_msg *msg, *msg_nxt;

 if (!imap_isconnected())
	return;

 msg = folder->messages;
 while (msg) {
  if (msg->flags & LOCKED) {
	display_msg(MSG_WARN, "IMAP", "Can not empty folder with opened messages");
	return;		   }
  msg = msg->next;
	     }	

 msg = folder->messages;
 while (msg) {
  msg_nxt = msg->next;
  if (msg->num > 0) {
	sprintf(buf, "%s/%d", fimap->fold_path, msg->num);
	unlink(buf);}
  discard_message(msg);
  msg = msg_nxt;
	     }

 folder->messages = NULL;
 folder->unread_num = 0;

 if ((ofold = imap_folder_switch(folder)) == NULL)
	return;

 if (ifold == ofold) {
  if (imap_command(ICOM_SELECT, "%s", imap_string(folder->fold_path)) != IMAP_OK)  {
	imap_folder_switch(ofold);
	return;							      }
		    }

 if (folder->flags & FRONLY) {
	display_msg(MSG_WARN, "IMAP", "Can not empty read only folder");
	return;		     }

 if (folder->num_msg == 0) {
	imap_folder_switch(ofold);
	return;		   }

 if (imap_command(ICOM_STORE, "1:%d +FLAGS.SILENT (\\Deleted)", folder->num_msg) !=
	IMAP_OK) {
 	imap_folder_switch(ofold);
	return;  }

 imap_command(ICOM_EXPUNGE, NULL);

 folder->num_msg = 0;

 imap_folder_switch(ofold);

 return;
}

int
delete_imap_folder(folder)
struct _mail_folder *folder;
{
struct _mail_msg *msg, *msg1;
int i;

 if (!imap_isconnected())
	return -1;

 if (folder->flags & FRONLY) 	{
	display_msg(MSG_WARN, "IMAP", "Can not delete read only folder");
	return -1;		}

 if ((folder == iinbox) || (folder == itrash) ||
	(folder->flags & NOTRASH)) 	{
	display_msg(MSG_WARN, "IMAP", "Can not delete this folder");
	return -1;
					}

 msg = folder->messages;
 while (msg) {
  if (msg->flags & LOCKED) {
	display_msg(MSG_WARN, "IMAP", "Close all messages in this folder and try again");
	return -1;
			   }
  msg = msg->next;
	     }

 if (folder == ifold) 	{
	ifold = NULL;
	if (imap_command(ICOM_CLOSE, NULL) != IMAP_OK)
		return -1;
			}

 if (imap_command(ICOM_DELETE, "%s", imap_string(folder->fold_path)) != IMAP_OK)
	return -1;

 for ( i = 0; i < folders_num; i++)    {
	if (mailbox[i] != folder)
		continue;

	if (folders_num)
		folders_num--;

	msg = folder->messages;
	while (msg) {
	  msg1 = msg->next;
	  discard_message(msg);
	  msg = msg1;
		    }

	free(mailbox[i]);
	mailbox[i] = NULL;
	for (; i < folders_num; i++)
		mailbox[i] = mailbox[i+1];

	break;
						}
 return 0;
}

int
rename_imap_folder(folder, name)
struct _mail_folder *folder;
char *name;
{
char *path;
char fname[MAX_IMAP_FOLD_NAME_LEN];

 if (!imap_isconnected())
	return -1;

 if (!folder || !name)
	return -1;

 if (folder->flags & SYSTEM) {
        display_msg(MSG_WARN, "rename", "Can not rename system folder");
        return -1;           }


 if (folder->flags & FRONLY) {
	display_msg(MSG_WARN, "IMAP", "Can not rename read only folder");
	return -1;	     }

 if ((strlen(name) < 1) || (strlen(name) > MAX_IMAP_FOLD_NAME_LEN)) {
	display_msg(MSG_WARN, "IMAP", "Invalid name length");
	return -1;						    }

 if ((folder->hdelim != '\0') && (strchr(name, folder->hdelim) != NULL)) {
	display_msg(MSG_WARN, "IMAP", "Illegal character in folder name");
	return -1;							 }

 path = get_imap_folder_path(folder);
 if ((folder->hdelim != '\0') && path)  {
	if (strlen(path) + strlen(name) >= MAX_IMAP_FOLD_NAME_LEN) {
		display_msg(MSG_WARN, "IMAP", "Name too long");
		return -1;					   }
	sprintf(fname, "%s%c%s", path, folder->hdelim, name);
					}
 else
	strcpy(fname, name);

 if (find_imap_folder(name)) {
	display_msg(MSG_WARN, "IMAP", "IMAP folder '%s' already exits", name);
	return -1;	     }

 if ((folder == iinbox) || (folder == itrash)) {
	display_msg(MSG_WARN, "IMAP", "Can not rename this folder");
	return -1;			       }


 if (imap_command(ICOM_RENAME, "%s %s", imap_string(folder->fold_path), fname) != IMAP_OK) {
	display_msg(MSG_WARN, "IMAP", "Rename failed");
	return -1;							      }

 strcpy(folder->fold_path, name);
 folder_sort &= ~FLD_SORTED;

 return 0;
}

void
update_imap_folder(folder)
struct _mail_folder *folder;
{
struct _mail_folder *ofold, *fld;
struct _mail_msg *msg, *msg_next;
int deleted;

 if (!imap_isconnected())
	return;

 if (!folder->messages)
	return;


 msg = folder->messages;
 msg_next = msg->next;
 deleted = 0;

 if ((ofold = imap_folder_switch(folder)) == NULL)
	return;

 while (msg) 	{
    msg_next = msg->next;

    if ((msg->flags & LOCKED) &&
	!(msg->flags & COPIED)) {
	msg->flags &= ~DELETED;
	msg->flags &= ~MOVED;
	msg->flags &= ~NOT_SENT;
	msg = msg->next;
	continue;	        }

    if (msg->flags & DELETED)   {
	msg->flags &= ~DELETED;
	msg->flags &= ~DELPERM;
	if (folder->flags & FRONLY)
		continue;
	display_msg(MSG_STAT, NULL, "Deleting %d", msg->uid);
	if (delete_imap_message_range(msg) >= 0)
		deleted++;
	msg = folder->messages;
	continue;
				}
    else
    if (msg->flags & MOVED)     {
	msg->flags &= ~MOVED;
	if (folder->flags & FRONLY)
		continue;
	fld = msg->folder;
	msg->folder = folder;
	display_msg(MSG_STAT, NULL, "Moving %d", msg->uid);
	update_imap_message_range(msg);
	if (fld->flags & F_IMAP) {
	 if (move_to_imap_folder_range(msg, fld) >= 0)
		deleted++;
				 }
	else	{
	 if (fld->move(msg, fld) == 0)
		deleted++;
		}
	msg = folder->messages;
	continue;
				}
    else
    if (msg->flags & COPIED)     {
	fld = msg->folder;
	msg->folder = folder;
	msg->flags &= ~COPIED;
	display_msg(MSG_STAT, NULL, "Copying %d", msg->uid);
	update_imap_message_range(msg);
	if (fld->flags & F_IMAP)
		copy_to_imap_folder_range(msg, fld);
	else
		fld->copy(msg, fld);
				 }
    else
    if (msg->flags & NOT_SENT)     {
		msg->flags &= ~NOT_SENT;
		display_msg(MSG_STAT, NULL, "Sending %d", msg->uid);
		msg->update(msg);
		send_message(msg);
				   }
    else
	update_imap_message_range(msg);

    msg = msg_next;
		}

 if (deleted)
	imap_command(ICOM_EXPUNGE, NULL);

 imap_folder_switch(ofold);

 return;
}

struct _mail_msg *
copy_to_imap_folder_range(msg, folder)
struct _mail_msg *msg;
struct _mail_folder *folder;
{
int i, top, bottom, copied;
struct _mail_folder *ofold;
struct _mail_msg *msg1;

 if (!msg || !folder || !(folder->flags & F_IMAP)) 
        return NULL;
 
 msg->flags &= ~COPIED;
 if (folder->flags & FRONLY) {
	display_msg(MSG_WARN, "IMAP", "Can not copy messages to read only folder");
	return NULL;	     }

 if (msg->flags & LOCKED)
	return NULL;

 if (!(msg->flags & M_IMAP))
	return copy_to_imap_folder(msg, folder);

 ofold = msg->folder;
 msg->folder = folder;
 msg->flags |= COPIED;
 expand_uid_range(ofold, msg, COPIED, 0, &bottom, &top);
 msg->flags &= ~COPIED;
 msg->folder = ofold;

 if (top == bottom)
	return copy_to_imap_folder(msg, folder);

 copied = 0;
 for (i = bottom; i <= top; i++)	{
  if ((msg1 = get_msg_by_uid(msg->folder, i)) != NULL) 	{
	msg1->folder = msg->folder;
	copied++;
	msg1->flags &= ~COPIED;
	msg1->update(msg1);				}
					}

 if (folder->num_msg + copied > MAX_MSG_IN_FOLDER) {
        display_msg(MSG_WARN, "copy","Folder %s is full", folder->fold_path);
        return NULL; 			  	  }

 if (folder->flags & OPENED)
	return copy_to_imap_folder(msg, folder);

 if ((ofold = imap_folder_switch(msg->folder)) == NULL)
	return NULL;

 if (imap_command(ICOM_UIDCOPY, "%d:%d %s", bottom, top, imap_string(folder->fold_path))
	!= IMAP_OK) {
  	imap_folder_switch(ofold);
	return NULL;  }

 imap_folder_switch(ofold);

 ofold = msg->folder;
 for (i = bottom; i <= top; i++)	{
  if ((msg1 = get_msg_by_uid(ofold, i)) != NULL) 	{
	folder->num_msg++;
	if (msg1->flags & UNREAD)
        	folder->unread_num++;
							}
					}

 folder->flags &= ~SORTED;

 return NULL;
}

struct _mail_msg *
copy_to_imap_folder(msg, folder)
struct _mail_msg *msg;
struct _mail_folder *folder;
{
struct _mail_folder *ofold;
struct _mail_msg *msg1;
int i;
char *flags;

 if (!msg || !folder || !(folder->flags & F_IMAP)) 
        return NULL;
 
 if (folder->flags & FRONLY) {
	display_msg(MSG_WARN, "IMAP", "Can not copy messages to read only folder");
	return NULL;	     }

 msg->flags &= ~COPIED;

 if (folder->num_msg > MAX_MSG_IN_FOLDER) {
        display_msg(MSG_WARN, "IMAP","Folder %s is full", folder->fold_path);
        return NULL; 			  }

 uid = -1;
 if (((folder->flags & OPENED) || (msg->flags & LOCKED)) 
	&& (icap & ICAP_STATUS))  {
  if (imap_command(ICOM_STATUS, "%s (UID-NEXT)", imap_string(folder->fold_path)) != IMAP_OK)
	uid = -1;
				  }

 msg->update(msg);
 if (msg->flags & M_IMAP) {
  if ((ofold = imap_folder_switch(msg->folder)) == NULL)
	return NULL;

  if (imap_command(ICOM_UIDCOPY, "%d %s", msg->uid, imap_string(folder->fold_path))
	!= IMAP_OK)   {
  	imap_folder_switch(ofold);
	return NULL;  }

  imap_folder_switch(ofold);
			  }
 else	{
  imsg = msg;

  flags = get_imap_flags(msg);
  if (icap & ICAP_IMAP4REV1) {
   if (imap_command(ICOM_APPEND, "%s (%s) \"%s\" {%d}",
	imap_string(folder->fold_path), flags ? flags : "",
	get_arpa_date(msg->header->rcv_time) , calc_msg_len(msg)) != IMAP_OK)  {
	display_msg(MSG_WARN, "IMAP", "Append failed");
	imsg = NULL;
	return NULL;							       }
			     }
   else	{
   if (imap_command(ICOM_APPEND, "%s {%d}", imap_string(folder->fold_path),
	calc_msg_len(msg)) != IMAP_OK)  {
	display_msg(MSG_WARN, "IMAP", "Append failed");
	imsg = NULL;
	return NULL;			}
	}

  imsg = NULL;
	}

 folder->num_msg++;
 if (msg->flags & UNREAD)
        folder->unread_num++;

 if ((folder->flags & OPENED) || (msg->flags & LOCKED)) {
  if (uid == -1) {
   if ((ofold = imap_folder_switch(folder)) == NULL)
	return NULL;

   imap_command(ICOM_UIDSEARCH, "ALL");
   if (search_res) {
    for (i = 1; i <= search_res[0]; i++) {
     if (!get_msg_by_uid(folder, search_res[i])) {
	uid = search_res[i];
	break;				 	 }
					 }
    free(search_res);
    search_res = NULL;
		  }
    imap_folder_switch(ofold);
	        }

  msg1 = copy_msg(msg);
  imap_message(msg1);
  msg1->uid = uid;
  msg1->flags |= H_ONLY;
  msg1->flags &= ~LOCKED;
  msg1->next = folder->messages;
  folder->messages = msg1;

  return msg1;
							}

 if ( ((folder_sort & 0x0f) == BY_MSGNUM) ||
        ( ( (folder_sort & 0x0f) == BY_UNREAD) &&
                (msg->flags & UNREAD) ) )
   folder_sort &= ~FLD_SORTED;

 return msg;
}

int
move_to_imap_folder_range(msg, folder)
struct _mail_msg *msg;
struct _mail_folder *folder;
{
int i, top, bottom, moved;
struct _mail_folder *ofold;
struct _mail_msg *msg1;

 if (!msg || !folder || !(folder->flags & F_IMAP)) 
        return -1;
 
 msg->flags &= ~MOVED;
 if (folder->flags & FRONLY) {
	display_msg(MSG_WARN, "IMAP", "Can not move messages to read only folder");
	return -1;	     }

 if (msg->folder && msg->folder->flags & FRONLY) {
	display_msg(MSG_WARN, "IMAP", "Can not move messages from read only folder");
	return -1;	     			 }

 if (msg->flags & LOCKED)
	return -1;

 if ((msg->folder) &&
	(msg->folder == folder))
                return 0;

 if (!(msg->flags & M_IMAP))
	return move_to_imap_folder(msg, folder);

 ofold = msg->folder;
 msg->folder = folder;
 msg->flags |= MOVED;
 expand_uid_range(ofold, msg, MOVED, 0, &bottom, &top);
 msg->flags &= ~MOVED;
 msg->folder = ofold;

 if (top == bottom)
	return move_to_imap_folder(msg, folder);

 moved = 0;
 for (i = bottom; i <= top; i++)	{
  if ((msg1 = get_msg_by_uid(msg->folder, i)) != NULL) 	{
	msg1->folder = msg->folder;
	moved++;
	msg1->flags &= ~MOVED;
	msg1->update(msg1);				}
					}

 if (folder->num_msg + moved > MAX_MSG_IN_FOLDER) {
        display_msg(MSG_WARN, "move","Folder %s is full", folder->fold_path);
        return -1; 			  	  }

 if (folder->flags & OPENED)
	return move_to_imap_folder(msg, folder);

 if ((ofold = imap_folder_switch(msg->folder)) == NULL)
	return -1;

 if (imap_command(ICOM_UIDCOPY, "%d:%d %s", bottom, top, imap_string(folder->fold_path))
	!= IMAP_OK) {
  	imap_folder_switch(ofold);
	return -1;  }

 if (imap_command(ICOM_UIDSTORE, "%d:%d FLAGS.SILENT (\\Deleted)", bottom, top)
	!= IMAP_OK) {
  	imap_folder_switch(ofold);
	return -1;  }

 imap_folder_switch(ofold);

 ofold = msg->folder;
 for (i = bottom; i <= top; i++)	{
  if ((msg1 = get_msg_by_uid(ofold, i)) != NULL) 	{
	folder->num_msg++;
	if (msg1->flags & UNREAD)
        	folder->unread_num++;
	if (ofold->num_msg)
		ofold->num_msg--;
	if ((msg1->flags & UNREAD) && ofold->unread_num)
		ofold->unread_num--;

	unlink_message(msg1);
	discard_message(msg1);
							}
					}

 folder->flags &= ~SORTED;

 return moved;
}

int
move_to_imap_folder(msg, folder)
struct _mail_msg *msg;
struct _mail_folder *folder;
{
struct _mail_folder *ofold;
struct _mail_msg *msg1;
int i;
char *flags;

 if (!msg || !folder || !(folder->flags & F_IMAP)) 
        return -1;
 
 if (folder->flags & FRONLY) {
	display_msg(MSG_WARN, "IMAP", "Can not move messages to read only folder");
	return -1;	     }

 if (msg->folder && msg->folder->flags & FRONLY) {
	display_msg(MSG_WARN, "IMAP", "Can not move messages from read only folder");
	return -1;	     			 }

 msg->flags &= ~MOVED;

 if (msg->flags & LOCKED)
        return -1;

 if ((msg->folder) &&
	(msg->folder == folder))
                return 0;

 if (folder->num_msg > MAX_MSG_IN_FOLDER) {
        display_msg(MSG_WARN, "move","Folder %s is full", folder->fold_path);
        return -1; 			  }

 uid = -1;
 if ((folder->flags & OPENED) && (icap & ICAP_STATUS))  {
  if (imap_command(ICOM_STATUS, "%s (UID-NEXT)", imap_string(folder->fold_path)) != IMAP_OK)
	uid = -1;
			 				}

 msg->update(msg);
 if (msg->flags & M_IMAP) {
  if ((ofold = imap_folder_switch(msg->folder)) == NULL)
	return -1;

  if (imap_command(ICOM_UIDCOPY, "%d %s", msg->uid, imap_string(folder->fold_path))
	!= IMAP_OK) {
  	imap_folder_switch(ofold);
	return -1;  }

  imap_folder_switch(ofold);
			  }
 else	{
  imsg = msg;

  flags = get_imap_flags(msg);
  if (icap & ICAP_IMAP4REV1) {
   if (imap_command(ICOM_APPEND, "%s (%s) \"%s\" {%d}",
	imap_string(folder->fold_path), flags ? flags : "",
	get_arpa_date(msg->header->rcv_time) , calc_msg_len(msg)) != IMAP_OK)  {
	display_msg(MSG_WARN, "IMAP", "Append failed");
	imsg = NULL;
	return -1;							       }
			     }
   else	{
   if (imap_command(ICOM_APPEND, "%s {%d}", imap_string(folder->fold_path),
	calc_msg_len(msg)) != IMAP_OK)  {
	display_msg(MSG_WARN, "IMAP", "Append failed");
	imsg = NULL;
	return -1;				}
	}

  imsg = NULL;
	}

 msg->flags &= ~M_TEMP;

 folder->num_msg++;
 if (msg->flags & UNREAD)
        folder->unread_num++;

 if (folder->flags & OPENED) {
  if (uid == -1) {
   if ((ofold = imap_folder_switch(folder)) == NULL)
	return -1;

   imap_command(ICOM_UIDSEARCH, "ALL");
   if (search_res) {
    for (i = 1; i <= search_res[0]; i++) {
     if (!get_msg_by_uid(folder, search_res[i])) {
	uid = search_res[i];
	break;				 	 }
					 }
    free(search_res);
    search_res = NULL;
		   }
   imap_folder_switch(ofold);
		 }

  msg1 = copy_msg(msg);
  imap_message(msg1);
  msg1->uid = uid;
  msg1->flags |= H_ONLY;
  msg1->next = folder->messages;
  folder->messages = msg1;
			     }

 msg->flags |= (DELETED|DELPERM);
 msg->delete(msg);

 folder->flags &= ~SORTED;

 if ( ((folder_sort & 0x0f) == BY_MSGNUM) ||
        ( ( (folder_sort & 0x0f) == BY_UNREAD) &&
                (msg->flags & UNREAD) ) )
   folder_sort &= ~FLD_SORTED;

 return 0;
}

char *
get_search_string(where)
char *where;
{
static char str[75];

 if (!strcasecmp(where, "Header"))
	return "TEXT";

 if (!strcasecmp(where, "Body"))
	return "BODY";

 if (!strcasecmp(where, "Message"))
	return "TEXT";

 if (!strcasecmp(where, "From"))
	return "FROM";

 if (!strcasecmp(where, "To"))
	return "TO";

 if (!strcasecmp(where, "Cc"))
	return "CC";

 if (!strcasecmp(where, "Bcc"))
	return "BCC";

 if (!strcasecmp(where, "Sender"))
	return "SENDER";

 if (!strcasecmp(where, "Subject"))
	return "SUBJECT";

 sprintf(str, "HEADER \"%s\"", where);

 return str;
}

int
imap_find_text(folder, text, where, flags)
struct _mail_folder *folder;
char *text;
char *where;
int flags;
{
struct _mail_folder *ofold;
struct _mail_msg *msg;
int found, i;

 if (!text || !folder || !(folder->flags & F_IMAP))
	return -1;

 if (strlen(where) > 63) {
	display_msg(MSG_WARN, "IMAP SEARCH", "Message part too long");
	return -1;	 }

 if (strlen(text) > 127) {
	display_msg(MSG_WARN, "IMAP SEARCH", "Expression too long");
	return -1;	 }

 if (!(folder->flags & OPENED) && (folder->flags & FRONLY)) {
	display_msg(MSG_WARN, "IMAP SEARCH", "Folder is read only\ncan not store search results");
	return -1;					    }

 if ((ofold = imap_folder_switch(folder)) == NULL)
	return -1;

 if (search_res)
	free(search_res);
 search_res = NULL;
	
 if (imap_command(ICOM_UIDSEARCH, "%s \"%s\"",
	get_search_string(where), text) != IMAP_OK) 	{
	display_msg(MSG_WARN, "IMAP SEARCH", "Search failed");
	imap_folder_switch(ofold);
	return -1;					}

 if (!search_res)
	return 0;

 if (search_res[0] == 0) {
	free(search_res);
	search_res = NULL;
	return 0;	 }

 found = search_res[0];

 if (folder->flags & OPENED) 	{
    for (i = 1; i <= found; i++) {
     if ((msg = get_msg_by_uid(folder, search_res[i])) != NULL)
	msg->flags |= MARKTMP;
     else
	imap_command(ICOM_UIDSTORE, "%d +FLAGS.SILENT (\\Flagged)",
		search_res[i]);
		}
				}
 else	{
    for (i = 1; i <= found; i++)
	imap_command(ICOM_UIDSTORE, "%d +FLAGS.SILENT (\\Flagged)",
		search_res[i]);
	}

 imap_folder_switch(ofold);
 
 free(search_res);
 search_res = NULL;

 return found;
}

void
imap_folder(folder)
struct _mail_folder *folder; 
{
 folder->name = get_imap_folder_name; 
 folder->open = open_imap_folder;
 folder->rescan = rescan_imap_folder;
 folder->close = close_imap_folder;
 folder->empty = empty_imap_folder;
 folder->delete = delete_imap_folder;
 folder->rename = rename_imap_folder;
 folder->update = update_imap_folder;
 folder->move = move_to_imap_folder;
 folder->copy = copy_to_imap_folder;
 folder->search = imap_find_text;
 folder->hdelim = '\0';
 folder->flags |= F_IMAP;
}

char *
get_imap_folder_domain_path(folder)
struct _mail_folder *folder;
{
char *p;

 if ((*folder->fold_path != '#') || (folder->hdelim == '\0'))
	return folder->fold_path;

 if ((p = strchr(folder->fold_path, folder->hdelim)) != NULL) {
	p++;
	return p;					      }

 return folder->fold_path;
}

char *
get_imap_folder_domain(folder)
struct _mail_folder *folder;
{
static char fdomain[16];
char *p;

 if (*folder->fold_path != '#')
	return NULL;

 strncpy(fdomain, folder->fold_path, 15);
 fdomain[15] = '\0';

 if ((folder->hdelim != '\0') &&
	((p = strchr(fdomain, folder->hdelim)) != NULL))
	*p = '\0';

 return fdomain;
}

struct _mail_folder *
find_imap_folder(name)
char *name;
{
int i;  
                
  if (!name)
        return NULL;

  if ((strlen(name) < 1) || (strlen(name) > MAX_FOLD_NAME_LEN))
        return NULL;  

  for ( i = 0; i < folders_num; i++)    {
	if (!(mailbox[i]->flags & F_IMAP))
		continue;
        if (!strcmp(mailbox[i]->fold_path, name))
                return mailbox[i];
                                        }

  return NULL;
}

struct _mail_folder *
find_imap_folder_by_name(name)
char *name;
{
int i;  
                
  if (!name)
        return NULL;

  if ((strlen(name) < 1) || (strlen(name) > MAX_FOLD_NAME_LEN))
        return NULL;  

  for ( i = 0; i < folders_num; i++)    {
	if (!(mailbox[i]->flags & F_IMAP))
		continue;
        if (!strcmp(mailbox[i]->name(mailbox[i]), name))
                return mailbox[i];
                                        }

  return NULL;
}
void
discard_imap_folders()
{
int i, k;
struct _mail_msg *msg, *msg1;
char buf[255];

  for ( i = 0; i < folders_num; i++)    {
	if (!(mailbox[i]->flags & F_IMAP))
		continue;
	if (folders_num)
		folders_num--;
	mailbox[i]->close(mailbox[i]);
	msg = mailbox[i]->messages;
	while (msg) {
	  msg1 = msg->next;
	  if (msg->flags & LOCKED) {
		msg->folder = fimap;
		msg->next = fimap->messages;
		fimap->messages = msg;
		local_message(msg);
				   }
	  else {
		if (msg->num > 0) {
		  sprintf(buf, "%s/%d", fimap->fold_path, msg->num);
		  unlink(buf);    }
		discard_message(msg);
		}

	  msg = msg1;
		    }

	free(mailbox[i]);
	mailbox[i] = NULL;
	for (k = i; k < folders_num; k++)
		mailbox[k] = mailbox[k+1];
	i--;
					}

  folder_sort &= ~FLD_SORTED;
  iinbox = NULL;
  itrash = NULL;

  return;
}

struct _mail_folder *
create_imap_folder(pfolder, name)
struct _mail_folder *pfolder;
char *name;
{
struct _mail_folder *folder, *ofold;

 if (!name)
	return NULL;

 if (!imap_isconnected()) 	{
	display_msg(MSG_WARN, "IMAP", "Not connected");
	return NULL;
				}

 if ((strlen(name) < 1) || (strlen(name) > MAX_IMAP_FOLD_NAME_LEN)) 	{
	display_msg(MSG_WARN, "IMAP", "Invalid folder name");
	return NULL;
									}

 if (find_imap_folder(name) != NULL) 	{
	display_msg(MSG_WARN, "IMAP", "Folder already exists");
	return NULL;
					}

 if (imap_command(ICOM_CREATE, "%s", name) != IMAP_OK) 	{
	display_msg(MSG_WARN, "IMAP", "Create failed");
	return NULL;
							}

 imap_h = name;
 if (imap_command(ICOM_LIST, "\"\" %s", name) != IMAP_OK) 	{
	display_msg(MSG_WARN, "IMAP", "Failed to list new folder");
	imap_h = NULL;
	return NULL;
								}
 imap_h = NULL;

 if ((folder = find_imap_folder(name)) == NULL) 	{
	display_msg(MSG_WARN, "IMAP", "Folder was not created");
	return NULL;
							}

 if ((ofold = imap_folder_switch(folder)) == NULL) {
	display_msg(MSG_WARN, "IMAP", "Folder can not be selected");
	return NULL;				   }

 if (folder->flags & FRONLY)
	display_msg(MSG_WARN, "IMAP", "New folder is read-only");

 imap_folder_switch(ofold);

 folder_sort &= ~FLD_SORTED;

 return folder;
}

void
imap_inbox()
{
 if (iinbox)
	return;

 if (!fimap) {
 if ((fimap = create_mh_folder(NULL, FIMAP)) == NULL) 	{
	display_msg(MSG_WARN, "IMAP INIT", "Can not open folder %s", FIMAP);
	imap_close(0);
	return;						}

 fimap->empty(fimap);
 if (fimap->open(fimap, 0) == -1)	{
	display_msg(MSG_WARN, "IMAP INIT", "Can not open folder %s", FIMAP);
	imap_close(0);
	return;			}

	    }

 if (imap_command(ICOM_LIST, "\"\" %s", IMAP_INBOX) != IMAP_OK)	{
	display_msg(MSG_WARN, "IMAP", "Failed to find INBOX folder");
	imap_disconnect(0);
	return;					 		}

 iinbox = find_imap_folder(IMAP_INBOX);

 return;
}

int
imap_inc()
{
int i, num, connected;
struct _mail_folder *ofold;
struct _mail_msg *msg;

 connected = 0;
 if (!imap_isconnected()) {
	if (imap_connect() != 0)
		return -1;
	if (imap_login() != 0)
		return -1;
	imap_inbox();
	connected = 1;    }

 if (!iinbox)	{
	display_msg(MSG_WARN, "IMAP", "No INBOX on server");
	if (connected) 			{
		discard_imap_folders();
		imap_command(ICOM_LOGOUT, NULL);
		imap_disconnect();	}
	return -1;
		}

 if (ifold == iinbox)	{
	if (imap_command(ICOM_SELECT, "%s", iinbox->fold_path) != IMAP_OK) {
	 display_msg(MSG_WARN, "IMAP", "Folder can not be selected");
	 if (connected) 		{
		discard_imap_folders();
		imap_command(ICOM_LOGOUT, NULL);
		imap_disconnect();	}
	 return -1;				   			   }
	ofold = iinbox;
			}
 else
 if ((ofold = imap_folder_switch(iinbox)) == NULL) {
	display_msg(MSG_WARN, "IMAP", "Folder can not be selected");
	if (connected) 			{
		discard_imap_folders();
		imap_command(ICOM_LOGOUT, NULL);
		imap_disconnect();	}
	return -1;				   }

 if (imap_command(ICOM_UIDSEARCH, "NOT SEEN") != IMAP_OK) {
	display_msg(MSG_WARN, "IMAP", "Search failed");
	imap_folder_switch(ofold);
	if (connected) 			{
		discard_imap_folders();
		imap_command(ICOM_LOGOUT, NULL);
		imap_disconnect();	}
	return -1;					  }

 if (!search_res || (search_res[0] == 0)) {
	if (search_res)
		free(search_res);
	search_res = NULL;
	imap_folder_switch(ofold);
	if (connected) 			{
		discard_imap_folders();
		imap_command(ICOM_LOGOUT, NULL);
		imap_disconnect();	}
	return 0;			  }

 num = search_res[0];
 for (i = 1; i <= num; i++) {
  if (imap_command(ICOM_UIDFETCH, "%d (INTERNALDATE RFC822.SIZE ENVELOPE FLAGS BODY UID)", search_res[i]) != IMAP_OK) 	{
	display_msg(MSG_WARN, "IMAP", "Failed to fetch message");
	if (search_res)
		free(search_res);
	search_res = NULL;
	imap_folder_switch(ofold);
	if (connected) 			{
		discard_imap_folders();
		imap_command(ICOM_LOGOUT, NULL);
		imap_disconnect();	}
	return -1;
					}
  imap_command(ICOM_UIDSTORE, "%d +FLAGS.SILENT (\\Seen)", search_res[i]);

  if ((msg = get_msg_by_uid(iinbox, search_res[i])) == NULL) 	{
	display_msg(MSG_WARN, "IMAP", "Failed to fetch message");
	if (search_res)
		free(search_res);
	search_res = NULL;
	imap_folder_switch(ofold);
	if (connected) 			{
		discard_imap_folders();
		imap_command(ICOM_LOGOUT, NULL);
		imap_disconnect();	}
	return -1;
								}

#ifdef FACES
  update_faces(msg);
#endif

  set_flags_by_status(msg);
  convert_fields(msg);
  msg->flags |= (CHANGED | RECENT);

  if (apply_rule(msg, 0) == -1) {
   if (search_res)
	free(search_res);
   search_res = NULL;
   imap_folder_switch(ofold);
   if (connected)	   {
	discard_imap_folders();
	imap_command(ICOM_LOGOUT, NULL);
	imap_disconnect(); }
   return -1;			} 

			    }

 imap_command(ICOM_EXPUNGE, NULL);

 if (search_res)
	free(search_res);
 search_res = NULL;

 imap_folder_switch(ofold);
 if (connected) 		{
	discard_imap_folders();
	imap_command(ICOM_LOGOUT, NULL);
	imap_disconnect();	}

 return num;
}

int
check_imap_folder(folder)
struct _mail_folder *folder;
{
struct _mail_folder *ofold;

 if (!folder || !(folder->flags & F_IMAP))
	return -1;

 if ((ofold = imap_folder_switch(folder)) == NULL) {
	display_msg(MSG_WARN, "IMAP", "Folder can not be selected");
	return -1;				   }

 if (imap_command(ICOM_UIDFETCH, "%d:*", get_max_uid(folder)) != IMAP_OK) {
	display_msg(MSG_WARN, "IMAP", "Can not check folder");
 	imap_folder_switch(ofold);
	return -1;							  }

 imap_folder_switch(ofold);

 return 0;
}

char *
skip_plist(ptr)
char *ptr;
{
char *p, *p1, c;
static int nlevel = 0;

 nlevel++;

 if (ptr == NULL)    {
	nlevel--;
	return NULL; }

 if (nlevel > 10) {
	display_msg(MSG_WARN, "IMAP-PARSE", "Nesting level too high");
	nlevel--;
	return NULL;
		  }

 p = ptr;

 while (*p != '\0') 	{
  p1 = p + strlen(p);
  if ((p = strpbrk(p, "\"'()")) == NULL) {
	nlevel--;
	return p1;			 }

  c = *p;
  switch (c) {
   case '"':
   case '\'':
	p++;
	if ((p = strchr(p, c)) == NULL) {
		display_msg(MSG_WARN, "IMAP-PARSE", "Unterminated string");
		nlevel--;
		return p1;		}
	p++;
   break;

   case '(':
	p++;
	p = skip_plist(p);
	if ((nlevel > 1) && (*p == ')') && *p)
		p++;
   break;

   case ')':
	nlevel--;
	return p;
   break;
	     }
			}

  nlevel--;
  return p;
}

char *
plist_getnext(list, next)
char *list, **next;
{
char *p1, c, *ln;
int len, res;
long llen;

 if (list)
	plist = list;
 else
  if (!plist || !*plist) {
	*next = NULL;
	plist = NULL;
	return NULL;
		 }

 if (elem)
	free(elem);
 elem = NULL;

 while (*plist == ' ')
	plist++;

 while (*plist != '\0') {
  c = *plist;
  switch (c) 	{
   case '"':
   case '\'':
    plist++;
    if ((p1 = strchr(plist, c)) == NULL) 	{
	display_msg(MSG_WARN, "IMAP - PARSE", "Unterminated string");
	return NULL;			}

    len = p1 - plist;
    if ((elem = (char *)malloc(len + 1)) == NULL) {
	display_msg(MSG_FATAL, "IMAP - PARSE", "Malloc failed");
	return NULL;			      	  }

    strncpy(elem, plist, len);
    elem[len] = '\0';
    p1++;
    plist = p1;
    while (*plist == ' ')
	plist++;
    *next = plist;

    return elem;
   break;

   case '(':
    if ((p1 = skip_plist(plist)) == NULL) 	{
	display_msg(MSG_WARN, "IMAP - PARSE", "Unterminated list");
	return NULL;				}

    if (*p1)
    	p1++;
    else {
     p1--;
     if (*p1 == '}') {
	p1--;
	while ((p1 != plist) && isdigit(*p1))
		p1--;
	if ((llen = is_literal(p1)) != -1) {
	 if(llen >= MAX_IMAP_STRING)   	{
		display_msg(MSG_WARN, "IMAP","unexpected literal data too big");
		return NULL;		}
	 len = plist - response;
	 response = (char *)realloc(response, strlen(response) + llen + 2 - strlen(p1));
	 if (response == NULL) 	{
		display_msg(MSG_FATAL, "IMAP", "Realloc failed");
		return NULL;	}
	 p1 = strrchr(response, '{');
	 if ((res = getdata(p1, llen, imap_in, NULL)) < 0) {
          if (res == -1) 	{
          	display_msg(MSG_WARN, "IMAP", "Can not receive string");
                imap_close(0);	}
	  else
	  if (res == -2)
		imap_reconnect();
          return NULL;                                     }
	 strcat(response, " ");

         if ((ln = getline(NULL, -65535, imap_in)) == NULL) 	{
	  display_msg(MSG_WARN, "IMAP", "Uncomplete responce");
	  imap_close(0);
       	  return NULL;		   		 	        }

         if (!*ln)     {
	  imap_reconnect();
	  return NULL; }

	 response = (char *)realloc(response,strlen(response) + strlen(ln) + 1);
	 if (response == NULL) 	{
		display_msg(MSG_FATAL, "IMAP", "Realloc failed");
		return NULL;	}

	 strcat(response, ln);
	 free(ln);
	 plist = response + len;

	 continue;
					   }
		     }
	 }

    len = p1 - plist;
    if ((elem = (char *)malloc(len + 1)) == NULL) {
	display_msg(MSG_FATAL, "IMAP - PARSE", "Malloc failed");
	return NULL;			      	  }

    strncpy(elem, plist, len);
    elem[len] = '\0';
    plist = p1;
    while (*plist == ' ')
	plist++;
    *next = plist;

    return elem;
   break;

   case ')':
	plist++;
	while (*plist == ' ')
		plist++;
	*next = plist;
	return NULL;
   break;

   default:
	if ((p1 = strpbrk(plist, ") ")) == NULL) {
		elem = strdup(plist);
		plist = NULL;
		*next = NULL;
		return elem;			 }

        len = p1 - plist;
        if ((elem = (char *)malloc(len + 1)) == NULL) {
	 display_msg(MSG_FATAL, "IMAP - PARSE", "Malloc failed");
	 return NULL;			      	      }

	strncpy(elem, plist, len);
	elem[len] = '\0';
	plist = p1;
	while (*plist == ' ')
		plist++;
	*next = plist;

	return elem;
   break;
		}
  plist++;
		    }

 plist = NULL;
 *next = NULL;
 return NULL;
}

int
start_plist()
{
 if (!plist)
	return -1;

 while (*plist == ' ')
	plist++;

 if (*plist == ')') {
	plist++;
	return -1;  }

 if (!strncasecmp(plist, "NIL", 3)) {
	plist += 3;
	return -1;		    }

 if (*plist != '(') {
	display_msg(MSG_WARN, "IMAP", "Missing parenthized list");
	return -1;  }
 else
	plist++;

 return 0;
}

void
end_plist()
{
 if (!plist)
	return;

 while (*plist && (*plist != ')'))
	plist++;

 if (*plist != ')')
	display_msg(MSG_WARN, "IMAP", "Unterminated parenthized list");
 else
	plist++;
}

char *
plist_getnext_string(plist, next)
char *plist, **next;
{
char *p, *str;

 if ((p = plist_getnext(plist, next)) == NULL)
	return NULL;

 if ((str = get_imap_string(p, NULL)) == NULL)  {
	free(elem);
	elem = NULL;
	return NULL;				}

 free(elem);
 elem = NULL;

 return str;
}

void
expand_uid_range(folder, msg, mask1, mask2, uid_bottom, uid_top)
struct _mail_folder *folder;
struct _mail_msg *msg;
int mask1, mask2;
int *uid_bottom, *uid_top;
{
struct _mail_msg *msg1;
int bottom, top, check;

 if ((mask1 & MOVED) || (mask1 & COPIED))
	check = 1;
 else
 if ((mask1 & DELETED) || (mask1 & DELPERM))
	check = 2;
 else
	check = 3;

 bottom = msg->uid;
 top = msg->uid;

 msg1 = msg;
 while (((msg1 = get_smaller_uid(folder, msg1->uid)) != NULL) &&
	!(msg1->flags & LOCKED)) {
	if ((check == 1) && (msg->folder != msg1->folder))
		break;
	else
	if ((check == 2) && !(msg1->flags & mask1))
		break;
	else
	if (check == 3)	{
	 if (((msg1->flags & mask1) != mask1) ||
		((msg1->header->flags & mask1) != 0))
		break;
	 if (((msg1->flags & mask2) != 0) ||
		((msg1->header->flags & mask2) != mask2))
		break;
			}	
	bottom = msg1->uid;	 }

 msg1 = msg;
 while (((msg1 = get_larger_uid(folder, msg1->uid)) != NULL) &&
	!(msg1->flags & LOCKED)) {
	if ((check == 1) && (msg->folder != msg1->folder))
		break;
	else
	if ((check == 2) && !(msg1->flags & mask1))
		break;
	else
	if (check == 3)	{
	 if (((msg1->flags & mask1) != mask1) ||
		((msg1->header->flags & mask1) != 0))
		break;
	 if (((msg1->flags & mask2) != 0) ||
		((msg1->header->flags & mask2) != mask2))
		break;
			}
	top = msg1->uid;	 }

 *uid_bottom = bottom;
 *uid_top = top;

 return;
}
