/*
 * Copyright(c) 1995,1996 by Gennady B. Sorokopud (gena@NetVision.net.il)
 *
 * This software can be freely redistributed and modified for 
 * non-commercial purposes as long as above copyright
 * message and this permission notice appear in all
 * copies of distributed source code and included as separate file
 * in binary distribution.
 *
 * Any commercial use of this software requires author's permission.
 *
 * This software is provided "as is" without expressed or implied
 * warranty of any kind.
 * Under no circumstances is the author responsible for the proper
 * functioning of this software, nor does the author assume any
 * responsibility for damages incurred with its use.
 *
 */

/* $Id: font.c,v 1.21 1996/12/15 14:22:28 gena Exp $
 */

#include <fmail.h>
#include <umail.h>
#include <fmail.h>
#include <choose_folder.h>

#define	MAX_ELEMENTS	64

static FL_FORM *form;

static FL_OBJECT
        *textobj,
        *styleobj,
        *sizeobj,
        *placeobj,
	*bg_obj,
	*fg_obj;

extern FL_OBJECT *b_folders, *b_messages;
static int ready = 0;
static int nfonts = 0;
static FD_Charset *charset_obj;
static int cchanged = 0;

typedef struct _size_el {
        char el_name[16];
	char el_descr[16];
        FL_OBJECT **el_obj;
	u_int type;
#define NO_TYPE 0x00
#define BROWSER 0x01
#define MENU 	0x02
#define FONT 	0x03
	u_int def_size, def_style, def_bg, def_fg;
        } size_el;

struct _size_el elements[MAX_ELEMENTS] = {
{ "Folders",	"Folder browser",  &b_folders, BROWSER, FL_NORMAL_SIZE, FL_FIXED_STYLE, FL_COL1, FL_YELLOW },
{ "Messages",	"Message browser", &b_messages,BROWSER, FL_NORMAL_SIZE, FL_FIXED_STYLE, FL_COL1, FL_YELLOW },
{ "ChooseFolder", "Folder chooser",NULL,       NO_TYPE, FL_NORMAL_SIZE, FL_FIXED_STYLE, FL_COL1, FL_YELLOW },
{ "AddrBook",	"Address book",    NULL,       NO_TYPE, FL_NORMAL_SIZE, FL_FIXED_STYLE, FL_COL1, FL_YELLOW },
{ "Signature",	"Signature editor",NULL,       NO_TYPE, FL_NORMAL_SIZE, FL_FIXED_STYLE, FL_COL1, FL_BLACK },
{ "View",	"Message viewer",  NULL,       NO_TYPE, FL_NORMAL_SIZE, FL_FIXED_STYLE, FL_COL1, FL_BLACK },
{ "Edit",	"Message editor",  NULL,       NO_TYPE, FL_NORMAL_SIZE, FL_FIXED_STYLE, FL_COL1, FL_BLACK },
{ "_",		"_", 		   NULL,       NO_TYPE, FL_NORMAL_SIZE, FL_FIXED_STYLE, FL_COL1, FL_YELLOW }
                        };

void display_styles();
void display_charsets();

void
restore_fonts()
{
int i;
char fname[16];

  for (i = 16; i < 32; i++) {
	sprintf(fname, "Font%d", i);
	if (cfg_exist(fname)) 
		fl_set_font_name(i, b_getcfg_str(conf_name, fname, ""));
	else
		break;
			    }

  return;
}

void
restore_charsets()
{
int k = 0;
char cname[32], aname[16], adescr[16], *str;

 while (supp_charsets[k].charset_code != CHAR_UNKNOWN) 	{
	if (is_charset_alias(supp_charsets[k].charset_name)) {
		k++;
		continue;
		}

	sprintf(cname, "%sFStyle", supp_charsets[k].charset_name);
	set_charset_style(supp_charsets[k].charset_name, b_getcfg_int(conf_name, cname, supp_charsets[k].font_style));

	sprintf(cname, "%sFSize", supp_charsets[k].charset_name);
	set_charset_size(supp_charsets[k].charset_name, b_getcfg_int(conf_name, cname, supp_charsets[k].font_size));

	sprintf(cname, "%sAlias", supp_charsets[k].charset_name);
	if (cfg_exist(cname)) 	{
	 str = b_getcfg_str(conf_name, cname, "");
	 while (1) {
	  aname[0] = '\0';
	  adescr[0] = '\0';
	  if (sscanf(str, "%12s %12s", aname, adescr) != 2)
		break;
	  add_charset(aname, adescr, supp_charsets[k].charset_code);
	  if ((str = strchr(str, ' ')) == NULL)
		break;
	  while (*str == ' ')
		str++;
	  if ((str = strchr(str, ' ')) == NULL)
		break;
	  while (*str == ' ')
		str++;
	  if (*str ==  '\0')
		break;
		   }
				}

  k++;
							}

 def_encoding = b_getcfg_int(conf_name, "encoding", DEFAULT_ENCODING);
 def_charset = b_getcfg_int(conf_name, "charset", DEFAULT_CHARSET);

 return;
}

void font_add(FL_OBJECT *obj, long arg)
{
char *p;
int i;
char *str, fname[16];

  if (nfonts >= 32) {
	display_msg(MSG_WARN, "FONTS", "Can not use more then 32 different fonts");
	return;
			}

  str = (char *)fl_show_input("Enter X-Windows font name", "");

  if (!str)
	return;

  if (strlen(str) < 2)
	return;
  
  i = 0;
  p = str;
  while ((p = strchr(p, '-')) != NULL) {
   p++;
   i++;
   if (i == 8) 	{
	if (*p == '*')
		*p = '?';

	break;
		}
					 }

  if (i < 8) {
	display_msg(MSG_WARN, "Invalid font", "%s", str);
	return;
	}

  sprintf(fname, "Font%d", nfonts);
  b_putcfg_str(conf_name, fname, str);

  fl_set_font_name(nfonts, str);
  display_styles();
  return;
}

void font_del(FL_OBJECT *obj, long arg)
{
char fname[32];
int i;

  if (nfonts <= 16) {
	display_msg(MSG_WARN, "FONT", "Can not delete predefined font");
	return;
		}

 i = 0;
 while (strlen(elements[i].el_name) > 2) {
	sprintf(fname, "%sFStyle", elements[i].el_name);
	
	if (b_getcfg_int(conf_name, fname, elements[i].def_style) == (nfonts - 1)) {
		display_msg(MSG_WARN, "FONT", "The font is in use. Can not delete");
		return;
						}
        i++;
					}

  nfonts--;

  sprintf(fname, "Font%d", nfonts);
  b_delcfg_str(conf_name, fname);

  fl_set_font_name(nfonts, "");
  display_styles();
  return;
}

void
display_size(el_num)
int el_num;
{
char buf[64];
int size, style, bgcol, fgcol;
int i;

 size  = elements[el_num].def_size;
 style = elements[el_num].def_style;
 bgcol = elements[el_num].def_bg;
 fgcol = elements[el_num].def_fg;

 switch (elements[el_num].type) {
	case NO_TYPE:
		fl_activate_object(bg_obj);
		fl_activate_object(fg_obj);
	break;

	case BROWSER:
		fl_activate_object(bg_obj);
		fl_activate_object(fg_obj);
		if (elements[el_num].el_obj) {
		bgcol = (*elements[el_num].el_obj)->col1;
		fgcol = (*elements[el_num].el_obj)->col2;
						}
	break;

	case MENU:
		fl_deactivate_object(bg_obj);
		fl_deactivate_object(fg_obj);
	break;

	default:
	break;
				}

 sprintf(buf, "%sFStyle", elements[el_num].el_name);
 style = b_getcfg_int(conf_name,buf,style);
 sprintf(buf, "%sFSize", elements[el_num].el_name);
 size = b_getcfg_int(conf_name,buf,size);

 if (elements[el_num].type == FONT) {
 strcpy(buf, "ViewBgCol");
 bgcol = b_getcfg_int(conf_name,buf,bgcol);
 strcpy(buf, "ViewFgCol");
 fgcol = b_getcfg_int(conf_name,buf,fgcol);
					}
 else {
 sprintf(buf, "%sBgCol", elements[el_num].el_name);
 bgcol = b_getcfg_int(conf_name,buf,bgcol);
 sprintf(buf, "%sFgCol", elements[el_num].el_name);
 fgcol = b_getcfg_int(conf_name,buf,fgcol);
	}

 fl_select_browser_line(styleobj, style + 1);

 switch(size) {
	case 8: i = 1; break;
	case 10: i = 2; break;
	case 11: i = 3; break;
	case 12: i = 4; break;
	case 13: i = 5; break;
	case 14: i = 6; break;
	case 18: i = 7; break;
	case 24: i = 8; break;
	case 30: i = 9; break;
	default: i = 1; break;
	      }

 fl_select_browser_line(sizeobj, i);

 fl_set_object_lstyle(textobj, style);
 fl_set_object_lsize(textobj,size);
 fl_set_object_color(textobj, bgcol, textobj->col2);
 fl_set_object_lcol(textobj, fgcol);
 fl_redraw_object(textobj);

 if (!elements[el_num].el_obj)
	return;

 switch (elements[el_num].type) {
	case NO_TYPE:
	break;

	case BROWSER:
		fl_set_object_color(*elements[el_num].el_obj, bgcol, fgcol);
		fl_set_browser_fontstyle(*elements[el_num].el_obj, style);
		fl_set_browser_fontsize(*elements[el_num].el_obj, size);
 		fl_redraw_object(*elements[el_num].el_obj);
	break;

	default:
	break;
			}
}

void color_cb(FL_OBJECT *obj, long arg)
{
char buf[64];
FL_COLOR col = fl_show_colormap( arg ? textobj->col1 : textobj->lcol); 
int num = fl_get_browser(placeobj) - 1;
if (arg)
	fl_set_object_color(textobj, col, textobj->col2);
else
	fl_set_object_lcol(textobj, col);

 if (elements[num].type == FONT) {
 	display_size(num);
	return; }

 sprintf(buf, "%sBgCol", elements[num].el_name);
 b_putcfg_int(conf_name, buf, textobj->col1);
 sprintf(buf, "%sFgCol", elements[num].el_name);
 b_putcfg_int(conf_name, buf, textobj->lcol);

 display_size(num);

}

void place_cb(FL_OBJECT *obj, long arg)
{
display_size(fl_get_browser(placeobj) - 1);
}

void style_cb(FL_OBJECT *obj, long arg)
{
char buf[64];
int num = fl_get_browser(placeobj) - 1;
int style = fl_get_browser(obj) - 1;
 sprintf(buf, "%sFStyle", elements[num].el_name);
 b_putcfg_int(conf_name, buf, style);

 if (elements[num].type == MENU)
	fl_show_message("Warning!", "This change will take effect", "after you restart XFMail");

 if (elements[num].type == FONT)
	set_charset_style(elements[num].el_name, style);

 display_size(num);
}

void size_cb(FL_OBJECT *obj, long arg)
{
char buf[64];
int num = fl_get_browser(placeobj) - 1;
int size = FL_DEFAULT_SIZE;

  switch (fl_get_browser(obj))
  {
    case 1: size = 8; break;
    case 2: size = 10; break;
    case 3: size = 11; break;
    case 4: size = 12; break;
    case 5: size = 13; break;
    case 6: size = 14; break;
    case 7: size = 18; break;
    case 8: size = 24; break;
    case 9: size = 30; break;
  }

 if (elements[num].type == FONT)
	set_charset_size(elements[num].el_name, size);

 sprintf(buf, "%sFSize", elements[num].el_name);
 b_putcfg_int(conf_name, buf, size);

 if (elements[num].type == MENU)
	fl_show_message("Warning!", "This change will take effect", "after you restart XFMail");

 display_size(num);
}

void create_form_form(void)
{
  FL_OBJECT *obj;

  form = fl_bgn_form(FL_NO_BOX,550.0,300.0);
  fl_add_box(FL_FLAT_BOX,0,0,550.0,300.0,"");
  textobj = obj = fl_add_text(FL_NORMAL_TEXT,10.0,10.0,530.0,100.0,"XFMail\nCopyright (c) 1995,1996 by\n Gennady B. Sorokopud\n\374\314\305\313\324\322\317\316\316\301\321 \320\317\336\324\301\n\351\360\345\370\350\367\354\340 \370\340\345\343");
    fl_set_object_boxtype(obj,FL_FRAME_BOX);
    fl_set_object_lalign(obj,FL_ALIGN_CENTER|FL_ALIGN_INSIDE);
    fl_set_object_color(obj,FL_INDIANRED,FL_YELLOW);
    fl_set_object_dblbuffer(obj, 1);
  styleobj = obj = fl_add_browser(FL_HOLD_BROWSER,10,120,180,130,"");
    fl_set_object_color(obj,FL_TOP_BCOL,FL_YELLOW);
    fl_set_object_callback(obj,style_cb,0);
  sizeobj = obj = fl_add_browser(FL_HOLD_BROWSER,200,120,170,130,"");
    fl_set_object_color(obj,FL_TOP_BCOL,FL_YELLOW);
    fl_set_object_callback(obj,size_cb,0);
  placeobj = obj = fl_add_browser(FL_HOLD_BROWSER,380,120,160,130,"");
    fl_set_object_color(obj,FL_TOP_BCOL,FL_YELLOW);
    fl_set_object_callback(obj,place_cb,0);

  bg_obj = fl_add_button(FL_NORMAL_BUTTON,330,260,70,30,"Bg Col");
  fl_set_object_callback(bg_obj,color_cb,1);
  fl_set_object_lsize(bg_obj,FL_NORMAL_SIZE);

  fg_obj = fl_add_button(FL_NORMAL_BUTTON,410,260,70,30,"Fg Col");
  fl_set_object_callback(fg_obj,color_cb,0);
  fl_set_object_lsize(fg_obj,FL_NORMAL_SIZE);

  obj = fl_add_button(FL_RETURN_BUTTON,10,260,70,30,"Ok");
  fl_set_object_lsize(obj,FL_NORMAL_SIZE);
  fl_set_object_lstyle(obj,FL_BOLD_STYLE);

  obj = fl_add_button(FL_NORMAL_BUTTON,90,260,70,30,"Help");
  fl_set_button_shortcut(obj,"#H",1);
  fl_set_object_callback(obj,cb_help_button,15);
  fl_set_object_lsize(obj,FL_NORMAL_SIZE);

  obj = fl_add_button(FL_NORMAL_BUTTON,170,260,70,30,"Add");
  fl_set_object_callback(obj,font_add,0);
  fl_set_object_lsize(obj,FL_NORMAL_SIZE);

  obj = fl_add_button(FL_NORMAL_BUTTON,250,260,70,30,"Delete");
  fl_set_object_callback(obj,font_del,0);
  fl_set_object_lsize(obj,FL_NORMAL_SIZE);

  fl_end_form();
}

void create_the_forms(void)
{
  create_form_form();
}

void addit(const char *str)
{
 nfonts++;
 fl_add_browser_line(styleobj,str);
}

void
display_styles()
{
int i = fl_get_browser(styleobj);

  if (i == 0)
	i = 1;

  fl_freeze_form(form);
  fl_clear_browser(styleobj);
  nfonts = 0;
  fl_enumerate_fonts(addit, 1);
  fl_select_browser_line(styleobj, i);
  fl_unfreeze_form(form);
}

void
set_el_size()
{
struct _size_el *sz = elements;
char chdescr[32];
int i = 0, k;
int sep;

 if (ready)
        return;

 create_the_forms();
 ready = 1;
 while (sz->el_name[0] != '_') {
        fl_add_browser_line(placeobj, sz->el_descr);
        i++;
        sz = &elements[i]; }
 k = 0;
 sep = i;
 while (supp_charsets[k].charset_code != CHAR_UNKNOWN) 	{
   if (i >= MAX_ELEMENTS) {
	display_msg(MSG_WARN, "FONTS", "Too many charsets, can't display");
	break;
			}

   if (is_charset_alias(supp_charsets[k].charset_name)) {
	k++;
	continue;
	}

   strcpy(elements[i].el_name, supp_charsets[k].charset_name);
   strcpy(elements[i].el_descr, supp_charsets[k].charset_descr);
   elements[i].el_obj = NULL;
   elements[i].type = FONT;
   elements[i].def_style = supp_charsets[k].font_style;
   elements[i].def_size = supp_charsets[k].font_size;
   elements[i].def_bg = FL_COL1;
   elements[i].def_fg = FL_BLACK;
   sprintf(chdescr, "%s (%s)", elements[i].el_name, elements[i].el_descr);
   fl_add_browser_line(placeobj, chdescr);
   i++;
   k++;
							}
 
  display_styles();

  fl_addto_browser(sizeobj,"8 (small)");
  fl_addto_browser(sizeobj,"10(normal1)");
  fl_addto_browser(sizeobj,"11(scaled)");
  fl_addto_browser(sizeobj,"12(normal2)");
  fl_addto_browser(sizeobj,"13(scaled)");
  fl_addto_browser(sizeobj,"14(normal2)");
  fl_addto_browser(sizeobj,"18(large)");
  fl_addto_browser(sizeobj,"24(Huge)");
  fl_addto_browser(sizeobj,"30(scaled)");

  fl_select_browser_line(sizeobj,2);
  fl_select_browser_line(placeobj,1);
  fl_select_browser_line(styleobj,1);
  display_size(0);

  fl_show_form(form,FL_PLACE_CENTER,FL_TRANSIENT,"Choose Style");

  fl_do_only_forms();

 elements[sep].el_name[0] = '_';

 fl_hide_form(form);
 fl_free_form(form);
 ready = 0;
 return;

}

void CHSET_Select_Call(FL_OBJECT *obj, long param)
{
}

void CHSET_AAlias_Call(FL_OBJECT *obj, long param)
{
char *p, aname[13], adescr[13], buf[127];
int i = fl_get_browser(charset_obj->CHSET_Select);

 if (i < 1)
	return;
 i--;

 if (is_charset_alias(supp_charsets[i].charset_name)) 	{
	display_msg(MSG_WARN, "add charset alias", "Can not make alias to alias");
	return;						}

 sprintf(buf, "New alias name of charset %s:", supp_charsets[i].charset_name);
 p=(char *)fl_show_input(buf, "");
 if (!*p)
	return;

 if (strlen(p) > 12) {
	display_msg(MSG_WARN, "add charset alias", "charset name can not be longer then 12 characters");
	return;	     }

 strncpy(aname, p, 12);
 aname[13] = '\0';
 p = aname;
 while (*p != '\0') {
  if (!isgraph(*p)) {
	display_msg(MSG_WARN, "add charset alias", "invalid character in charset name");
	return;	    }
  p++;		    }

 p=(char *)fl_show_input("Enter charset description:", "");
 if (!*p)
	return;

 if (strlen(p) > 12) {
	display_msg(MSG_WARN, "add charset alias", "charset description can not be longer then 12 characters");
	return;	     }

 strncpy(adescr, p, 12);
 adescr[13] = '\0';
 p = adescr;
 while (*p != '\0') {
  if (!isgraph(*p) || (*p == ' ')) {
	display_msg(MSG_WARN, "add charset alias", "invalid character in charset description");
	return;	    		   }
  p++;		    }

 if (add_charset(aname, adescr, supp_charsets[i].charset_code) != 0)
	return;

 display_charsets();
 cchanged = 1;

 return;
}

void CHSET_DAlias_Call(FL_OBJECT *obj, long param)
{
int i = fl_get_browser(charset_obj->CHSET_Select);

 if (i < 1)
	return;
 i--;

 if (!is_charset_alias(supp_charsets[i].charset_name)) 	{
	display_msg(MSG_WARN, "delete charset alias", "Only charset aliases can be deleted");
	return;						}

 if (del_charset(supp_charsets[i].charset_name) != 0)
	return;

 display_charsets();
 cchanged = 1;

 return;
}

void
display_charsets()
{
char buf[255];
int i, k;

 i = 0;
 fl_freeze_form(charset_obj->Charset);
 fl_clear_browser(charset_obj->CHSET_Select);
 while (supp_charsets[i].charset_code != CHAR_UNKNOWN) 	{
  k = is_charset_alias(supp_charsets[i].charset_name);
  sprintf(buf, "%-11.11s %-11.11s %-11.11s %-4d %-4d",
   supp_charsets[i].charset_name,
   supp_charsets[i].charset_descr ?  supp_charsets[i].charset_descr : "",
   k > 0 ? supp_charsets[k - 1].charset_name : "",
   supp_charsets[i].font_style, supp_charsets[i].font_size);
  fl_add_browser_line(charset_obj->CHSET_Select, buf);
  i++;
							}
 fl_select_browser_line(charset_obj->CHSET_Select, 1);
 fl_unfreeze_form(charset_obj->Charset);

 return;
}

void
charset_conf()
{
char buf[255], aname[32];
int i, k;

 if (ready)
        return;
 
 ready = 1;
 cchanged = 0;
 
 charset_obj = create_form_Charset();
 fl_set_browser_fontstyle(charset_obj->CHSET_Select, FL_FIXED_STYLE);
 fl_set_browser_fontsize(charset_obj->CHSET_Select, FL_SMALL_SIZE);
 display_charsets();
 
 fl_show_form(charset_obj->Charset, FL_PLACE_CENTER, FL_TRANSIENT, "Character sets");
 fl_do_only_forms();

 fl_hide_form(charset_obj->Charset);
 fl_free_form(charset_obj->Charset);
 fl_free(charset_obj);
 charset_obj = NULL;

 if (cchanged) {
  i = 0;
  while (supp_charsets[i].charset_code != CHAR_UNKNOWN) 	{
   if (is_charset_alias(supp_charsets[i].charset_name)) {
	i++;
	continue;					}

   k = i + 1;
   buf[0] = '\0';
   while (supp_charsets[k].charset_code != CHAR_UNKNOWN) {
    if (supp_charsets[k].flags & CHARSET_FIXED) {
	k++;
	continue;				}
    if (supp_charsets[k].charset_code == supp_charsets[i].charset_code) {
     if (*buf)
	strcat(buf, " ");
     strcat(buf, supp_charsets[k].charset_name);
     strcat(buf, " ");
     strcat(buf, supp_charsets[k].charset_descr);
									}
    k++;
							 }
   sprintf(aname, "%sAlias", supp_charsets[i].charset_name);
   if (*buf)
	b_putcfg_str(conf_name, aname, buf);
   else	{
    if (cfg_exist(aname))
	b_delcfg_str(conf_name, aname);
	}
   i++;
								}
	       }
 
 ready = 0;
 return;
}
