/*
 * Copyright(c) 1995,1996 by Gennady B. Sorokopud (gena@NetVision.net.il)
 *
 * This software can be freely redistributed and modified for 
 * non-commercial purposes as long as above copyright
 * message and this permission notice appear in all
 * copies of distributed source code and included as separate file
 * in binary distribution.
 *
 * Any commercial use of this software requires author's permission.
 *
 * This software is provided "as is" without expressed or implied
 * warranty of any kind.
 * Under no circumstances is the author responsible for the proper
 * functioning of this software, nor does the author assume any
 * responsibility for damages incurred with its use.
 *
 * XFMail pgp interface
 * Thomas Pundt <pundtt@math.uni-muenster.de>
 * Gennady B. Sorokopud <gena@NetVision.net.il>
 *
 */

/* $Id: pgp.c,v 1.5 1996/12/15 14:22:32 gena Exp $
 */

#include <fmail.h>
#include <umail.h>
#include <dialogs.h>
#include <unistd.h>
#include <stdio.h>
#include <errno.h>

#define	FORGET_PHRASE	7200	/* forget passphrase after 2 hours */
static char *phrase = NULL;	/* scrambled passphrase 	   */
static FD_Passphrase *pass = NULL; /* passphrase input form */
static int pgpext;		/* PGP exit status */
static FD_PGP_Conf *pgp_obj;
static int ready = 0;

void PGP_Mime_Call(FL_OBJECT *obj, long param)
{
}

void PGP_Prog_Call(FL_OBJECT *obj, long param)
{
}

void PGP_User_Call(FL_OBJECT *obj, long param)
{
}

void
cb_input_passphrase(FL_OBJECT *obj,long param)
{
}

void
PGP_Timer_Call(FL_OBJECT *obj,long param)
{
 if (phrase) {
	free(phrase);
	phrase = NULL;
	     }
}

void
pgp_cleanup(struct _proc_info *pinfo)
{
 pgpext = pinfo->status;
}

void
scramble(string)
char *string;
{
unsigned char *p;
 
   if (!string)
        return;
 
   p = (unsigned char *)string;
   while (*p != '\0')   {
    if (*p != 0xaa)
        *p = *p ^ 0xaa;
        p++;
                        }
 
   return;
}

char* input_passphrase()
{
char *rphrase;

 if (phrase)
	goto retp;

 if (!pass)
  pass=create_form_Passphrase();

 fl_show_form(pass->Passphrase,FL_PLACE_MOUSE,FL_TRANSIENT,"Input Passphrase");
 fl_do_only_forms();
 phrase = strdup((char *)fl_get_input(pass->Input_Passphrase));
 scramble(phrase);
 fl_set_timer(pgp_timer, FORGET_PHRASE);
 fl_hide_form(pass->Passphrase);
 fl_free_form(pass->Passphrase);
 fl_free(pass);
 pass = NULL;

retp:
 rphrase = strdup(phrase);
 scramble(rphrase);

 return rphrase;
}      

void
pgp_init(struct _proc_info *pinfo)
{
  putenv("PGPPASSFD=0"); /* see pgp/doc/appnote.doc */
}

/*
 *      inbuf: file with text to encode/decode
 *       mode: ENCODE|SIGN or DECODE
 *       recp: if mode&(ENCODE|SIGN), the list of recipient(s)
 * passphrase: if mode==DECODE, or mode==SIGN, the pass phrase
 * see also "pgp/doc/appnote.doc"
 */

int
pgp_action(char *inbuf, int mode, char *passphrase, char *recp)
{
  char buf[1024];
  char args[1024];
  char nfile[64];
  char *pgpprg, *pgpuser, *p, *q, *r;
  int n=0, move_result=1;
  struct _proc_info pinfo;
  FILE *file;

  init_pinfo(&pinfo);
  pinfo.handle = pgp_cleanup;
  pinfo.init = pgp_init;
  pinfo.fd_in[0]=pinfo.fd_out[0]=pinfo.fd_err[0]=0;

  pgpuser=strdup(b_getcfg_str(conf_name, "pgpuser", user_n));
  pgpprg=b_getcfg_str(conf_name, "pgp", "/usr/local/bin/pgp");

  if(mode&(ENCODE|SIGN))
    sprintf(args,"%s +batchmode +language=en -fat %s%s %s%s",pgpprg,
	(mode&ENCODE) ? "-e " : "", recp,
	(mode&SIGN)? "-su " : "", (mode&SIGN) ? pgpuser : "");
  free(pgpuser);
    
  if(mode&(DECODE|VERIFY))
    sprintf(args,"%s +verbose=1 +batchmode +language=en -f",pgpprg);

  if (mode&EXTKEY)
    sprintf(args,"%s +verbose=1 +batchmode +language=en -f -ka",pgpprg);

  if (mode&ADDKEY)
    sprintf(args,"%s +verbose=1 +batchmode +language=en -f -kxa %s",pgpprg, recp);

  if(exec_child(args,&pinfo)==-1) {
    pgp_cleanup(&pinfo);
    if (pinfo.fd_in[1] > 0)
	close(pinfo.fd_in[1]);
    if (pinfo.fd_out[0] > 0)
	close(pinfo.fd_out[0]);
    if (pinfo.fd_err[0] > 0)
	close(pinfo.fd_err[0]);

    return -1;
  }

  /* init pgp exit status */
  pgpext = -100;

  /* first line of stdin must contain passphrase */
  if(passphrase && (mode&(DECODE|SIGN)))
    write(pinfo.fd_in[1],passphrase,strlen(passphrase));
  write(pinfo.fd_in[1],"\n",1);

  if ((file=fopen(inbuf,"r")) != NULL) 	{
   while((n=fread(buf,1,1023,file))>0)
    write(pinfo.fd_in[1],buf,n);
   close(pinfo.fd_in[1]);
   fclose(file);			}

  strcpy(nfile, get_temp_file("pgp"));
  if ((file=fopen(nfile,"w")) != NULL) 	{
   while((n=read(pinfo.fd_out[0],buf,1023))>0)
    fwrite(buf,1,n,file);
   close(pinfo.fd_out[0]);
   fclose(file);			}

  while((n=read(pinfo.fd_err[0],buf,1023))>0) {
    buf[n] = '\0';
    p = strstr(buf, "Current");
    if (p)
	p = strchr(p, '\n');
    if (p)
	p++;

    if (p) {
     while ((q = strchr(p, '\n')) != NULL) {
	*q = '\0';
	if (strlen(p))
	 display_msg(MSG_LOG, "PGP", p);
	*q = '\n';
	p = q + 1;
					   }
     if (strlen(p))
      display_msg(MSG_LOG, "PGP", p);
	   }

    if((p=strstr(buf,"Good signature")) ||
       (p=strstr(buf,"WARNING:"))) {
	if (strlen(p) > 255)
		p[255] = '\0';
	if ((q = strchr(p, '\n')) != NULL) 	{
		*q = '\0';
		q++;
      		display_msg(MSG_MSG,p,q);	}
	else
		display_msg(MSG_MSG,"PGP messsage (see Log for more details)",p);
    }
    if((p=strstr(buf,"ERROR:")) ||
       (p=strstr(buf,"Error:")) ||
       (p=strstr(buf,"You do not have the secret key")) ||
       (p=strstr(buf,"Key matching expected")) ||
       (p=strstr(buf,"Encryption error")) ||
       (p=strchr(buf,'\007')) ) {
        move_result=0;
	if ((q = strchr(p, '\n')) != NULL) 	{
		*q = '\0';
		if ((r = strchr(q, '\n')) != NULL)
			*r = '\0';
		while (*q < 32)
			q++;
		while (*p < 32)
			p++;
		if (strlen(q) > 80)
			q[80] = '\0';
		if (strlen(p) > 80)
			p[80] = '\0';
      		display_msg(MSG_WARN,p,q);	}
	else	{
		while (*p < 32)
			p++;
		if (strlen(p) > 80)
			p[80] = '\0';
		display_msg(MSG_WARN,"PGP Error (See Log for more details)",p);
		}
	/* forget the pass phrase */
      if(mode&(DECODE|SIGN))
	PGP_Timer_Call(NULL, 0);
    }
  }
  
  close(pinfo.fd_err[0]);

  if (pgpext == -100) 	{	/* pgp is still running */
	display_msg(MSG_WARN, "PGP", "Stuck PGP process");
	kill(pinfo.pid, SIGKILL);	/* nuke it */
	move_result = 0;
			}
  else /* pgp exit status was not 0 and not 1 (??) */
   if ((pgpext != 0) && (pgpext != 1)) 	{
	display_msg(MSG_WARN, "PGP", "PGP terminated with status %d", pgpext);
	move_result = 0;
			}

  if(move_result) {
    if (do_move(nfile,inbuf) != 0)
	return -1;
    return 0;
  } else
    unlink(nfile);

  return -1;
}

void
pgp_conf()
{
 if (ready)
        return;

 ready = 1;

 pgp_obj = create_form_PGP_Conf();

 fl_set_input(pgp_obj->PGP_Prog, b_getcfg_str(conf_name, "pgp", "/usr/local/bin/pgp"));
 fl_set_input(pgp_obj->PGP_User, b_getcfg_str(conf_name, "pgpuser", user_n));
 fl_set_button(pgp_obj->PGP_Mime, b_getcfg_int(conf_name, "pgpmime", 0));
 fl_set_button(pgp_obj->PGP_Sign, b_getcfg_int(conf_name, "pgpsign", 0));

 fl_show_form(pgp_obj->PGP_Conf, FL_PLACE_CENTER, FL_TRANSIENT, "PGP");
 fl_do_only_forms();

 b_putcfg_str(conf_name, "pgp", (char *)fl_get_input(pgp_obj->PGP_Prog));
 b_putcfg_str(conf_name, "pgpuser", (char *)fl_get_input(pgp_obj->PGP_User));
 b_putcfg_int(conf_name, "pgpmime", fl_get_button(pgp_obj->PGP_Mime));
 b_putcfg_int(conf_name, "pgpsign", fl_get_button(pgp_obj->PGP_Sign));

 fl_hide_form(pgp_obj->PGP_Conf);
 fl_free_form(pgp_obj->PGP_Conf);
 fl_free(pgp_obj);
 pgp_obj = NULL;

 ready = 0;
 return;
}

