/*
 * Copyright(c) 1995,1996 by Gennady B. Sorokopud (gena@NetVision.net.il)
 *
 * This software can be freely redistributed and modified for 
 * non-commercial purposes as long as above copyright
 * message and this permission notice appear in all
 * copies of distributed source code and included as separate file
 * in binary distribution.
 *
 * Any commercial use of this software requires author's permission.
 *
 * This software is provided "as is" without expressed or implied
 * warranty of any kind.
 * Under no circumstances is the author responsible for the proper
 * functioning of this software, nor does the author assume any
 * responsibility for damages incurred with its use.
 *
 */

/* $Id: rules.c,v 1.35 1996/12/19 14:39:33 gena Exp $
 */

#include <fmail.h>
#include <umail.h>
#include <fmail.h>
#include <choose_folder.h>

static int ready = 0;
static FD_Rules_Form *rule_obj;
int changed, r_modified;

void display_rule(int num);
int compose_rule(int num);
int rule_by_name(char *rname);
void Rule_Modify_Call(FL_OBJECT *obj, long param);

#define MAX_RULES	64

static struct _xf_rule rules[MAX_RULES];

char *
strcasestr(s, find, scase)
        register const char *s, *find;
	int scase;
{
        register size_t len; 
 
	if (!scase)
		return strstr(s, find);

	len = strlen(find);
	while (strncasecmp(s, find, len) != 0) 	{
	 s++;
	 if (*s == 0)
		return (NULL);
						}

        return ((char *)s);
} 

int
rule_by_name(rname)
char *rname;
{
int i;

 for (i = 0; i < MAX_RULES; i++) {
  if (!strcasecmp(rname, rules[i].name))
	return i;
		}

 return -1;
}

int
load_rules()
{
FILE *rfd;
char buf[255];
char rfname[255];
int i;

#ifdef HS_REGEX
char errbuf[2048];
int regerr;
int regflags;
#endif

 sprintf(rfname, "%s/.xfmrules", homedir);

 for (i = 0; i < MAX_RULES; i++)
	rules[i].action = 0;

 if ((rfd = fopen(rfname, "a+")) == NULL) {
	display_msg(MSG_WARN, "Can not read rules database", "%s", rfname);
	return -1;	}

 fseek(rfd, 0L, SEEK_SET);
 i = 0;
 while (fgets(buf, 254, rfd)) {
	if (buf[0] != '@')
		continue;

	strip_newline(buf);

	sscanf(buf + 1, "%s %d %d %s %s", rules[i].name, &rules[i].action, &rules[i].flags, rules[i].data, rules[i].fmatch);

	if (fgets(buf, 254, rfd)) {
		if (strlen(buf))
		  strip_newline(buf);
		strncpy(rules[i].tmatch, buf, 64);
				  }
	else
		rules[i].tmatch[0] = '\0';

#ifdef HS_REGEX
	regfree(&rules[i].rx);
	regflags = REG_NOSUB | REG_EXTENDED;
	if (rules[i].flags & CASE_I)
		regflags |= REG_ICASE;

	if ((regerr = regcomp(&rules[i].rx, rules[i].tmatch, regflags))) {
		regerror(regerr, &rules[i].rx, errbuf, sizeof(errbuf));
		display_msg(MSG_WARN, "Invalid regular expression", "%s", errbuf);
		regfree(&rules[i].rx);
		continue;
		     							 }
#endif

	if ((rules[i].action < 1) || (rules[i].action > RULE_MAXACTION)) {
		display_msg(MSG_WARN, "rules", "Invalid action code %d", rules[i].action);
		continue; }

	i++;
	if (i >= MAX_RULES) {
		display_msg(MSG_WARN, "rules", "Rules file too long");
		fclose(rfd);
		return 0; }
				}

 fclose(rfd);

 return 0;
}

int
save_rules()
{
FILE *rfd;
char rfname[255];
int i;

 sprintf(rfname, "%s/.xfmrules", homedir);

 if (!(rfd = fopen(rfname, "w"))) {
	display_msg(MSG_WARN, "Can not save rules database", "%s", rfname);
	return -1;	}

 for (i = 0;i < MAX_RULES; i++) {
	if (rules[i].action == 0)
		continue;

	fprintf(rfd, "@%s %d %d %s %s\n", rules[i].name, rules[i].action, rules[i].flags, rules[i].data, rules[i].fmatch);

	fprintf(rfd, "%s\n", rules[i].tmatch);
				}

 changed = 0;
 r_modified = 0;

 fclose(rfd);
 return 0;
}

u_int
match_addr(addr, rule)
mail_addr *addr;
struct _xf_rule *rule;
{
  if (!addr || !rule)
	return 0;

  while (addr)  {
#ifdef	HS_REGEX
   if (regexec(&rule->rx, addr->addr, 0, 0, 0) == 0)
	return 1;

   if (addr->name && (regexec(&rule->rx, addr->name, 0, 0, 0) == 0))
	return 1;

   if (addr->comment && (regexec(&rule->rx, addr->comment, 0, 0, 0) == 0))
	return 1;
#else
   if (strcasestr(addr->addr, rules->tmatch,
	rules->flags & CASE_I))
	return 1;

   if (addr->name &&
	strcasestr(addr->name, rules->tmatch, rules->flags & CASE_I))
	return 1;

   if (addr->comment &&
	strcasestr(addr->comment, rules->tmatch, rules->flags & CASE_I))
	return 1;
#endif

   addr = addr->next_addr;
		}

  return 0;
}

void
init_rule(rule)
struct _xf_rule *rule;
{
 if (!rule)
	return;

 *rule->name = '\0';
 *rule->fmatch = '\0';
 *rule->tmatch = '\0';
 *rule->data = '\0';

  rule->action = 0;
  rule->flags = 0;

  return;
}

int
match_rule(msg, rule)
struct _mail_msg *msg;
struct _xf_rule *rule;
{
struct _mail_addr *addr;
int fr;
char *p;

 if (!msg || !rule)
	return 0;

 if ((addr = get_addr_by_name(msg, rule->fmatch)) != NULL)
	return match_addr(addr, rule);

 if (!strcasecmp(rule->fmatch, "Recipients"))  {
   if (match_addr(msg->header->To, rule))
	return 1;
   if (match_addr(msg->header->Cc, rule))
	return 1;
   if (match_addr(msg->header->Bcc, rule))
	return 1;

   return 0;
					       }

 if (!(p = get_field_content(msg, rule->fmatch, &fr)))
	return 0;

#ifdef HS_REGEX
 if (regexec(&rule->rx, p, 0, 0, 0) == 0) {
	free_field_content(msg, p, fr);
	return 1;			  }
#else
 if (strcasestr(p, rule->tmatch, rule->flags & CASE_I)) {
	free_field_content(msg, p, fr);
	return 1;			 		}
#endif

 free_field_content(msg, p, fr);

 return 0;
}

struct _xf_rule *
match_msg(msg, type)
struct _mail_msg *msg;
int type;
{
int i;

 if (!msg)
	return NULL;

 for (i = 0;i < MAX_RULES; i++) {
	if (rules[i].action == 0)
		continue;

	if ((type > 0) && (rules[i].action != type))
		continue;

	if (match_rule(msg, &rules[i]))
		return &rules[i];
				}

 return NULL;
}

int
apply_rule(msg, delay)
struct _mail_msg *msg;
int delay;
{
struct _mail_folder *folder;
struct _mail_msg *msg1;
struct _xf_rule *rule;
int action;
char *data;


	if ((rule = match_msg(msg, -1)) == NULL) {
		action = R_MOVE;
		data = INBOX;		     }
	else	{
		action = rule->action;
		data = rule->data;
		}

	if (msg->flags & RECENT) {
	 if (rule) {
	  if (rule->flags & R_SAVE_ADDR)
		add_msg_addr(msg);
		  }
	 else	{
	  if (b_getcfg_int(conf_name, "saveaddr", 0))
		add_msg_addr(msg);
		}
				 }

        switch(action) {
                case R_MOVE:
                 folder = get_folder_by_name(data);
                 if (!folder)
                  folder = inbox;

		 if (delay && msg->folder) {
			msg->folder = folder;
			msg->flags |= MOVED;
			break;
			}

                 if (folder->move(msg, folder) == -1) {
                        discard_message(msg);
                        return -1; }
                break;

                case R_DISCARD:
			msg->flags |= (DELETED|DELPERM);
			if (!delay)
                        	msg->delete(msg);
                break;

                case R_FORWARD:
                        msg1 = get_fwd_msg(msg);
                        if (!msg1) {
                                discard_message(msg);
                                return -1; }

			if (!strcmp(data, "$sender"))
			   msg1->header->To = copy_address(msg->header->Sender);
			else
			if (!strcmp(data, "$from"))
			   msg1->header->To = copy_address(msg->header->From);
			else
			   msg1->header->To = get_address(data, 0);

                        if (send_message(msg1) != 0) {
                                discard_message(msg);
				msg1->flags |= (DELETED|DELPERM);
                                msg1->delete(msg1);
                                return -1; }

			msg->flags |= (DELETED|DELPERM);
			if (!delay)
                        	msg->delete(msg);
                break;

		case R_VACATION:
                        msg1 = get_vac_msg(msg, data);
                        if (!msg1) {
                                discard_message(msg);
                                return -1; }

                        if (send_message(msg1) != 0) {
                                discard_message(msg);
				msg1->flags |= (DELETED|DELPERM);
                                msg1->delete(msg1);
                                return -1; }

			if (rule->flags & R_VAC_KEEP) 	{
				if ((rule = match_msg(msg, R_MOVE)) == NULL)
				 folder = inbox;
				else	{
				 if ((folder = get_folder_by_name(rule->data)) == NULL)
					folder = inbox;
					}

		 		if (delay && msg->folder) {
					msg->folder = folder;
					msg->flags |= MOVED;
							  }
				else	{
				if (folder->move(msg, folder) == -1) {
					discard_message(msg);
					return -1;		     }
					}
							}
			else	{
				msg->flags |= (DELETED|DELPERM);
				if (!delay)
                        		msg->delete(msg);
				}
		break;

                default:
		 if (delay && msg->folder) {
			msg->folder = inbox;
			msg->flags |= MOVED;
			break;
				}

                 if (inbox->move(msg, inbox) == -1) {
                        discard_message(msg);
                        return -1; }
                break;
                                }

	return 0;
}

void Rule_Name_Call(FL_OBJECT *obj, long param)
{
}

void Rule_Case_Call(FL_OBJECT *obj, long param)
{
r_modified = 1;
}

void Rule_Up_Call(FL_OBJECT *obj, long param)
{
struct _xf_rule rule;
int r1, r2;
int i;
int k = fl_get_browser(rule_obj->Rule_List);

 if (k < 2)
        return;

r1 = rule_by_name((char *)fl_get_browser_line(rule_obj->Rule_List, k));
r2 = rule_by_name((char *)fl_get_browser_line(rule_obj->Rule_List, k - 1));

if ((r1 == -1) || (r2 == -1))
        return;

memcpy(&rule, &rules[r1], sizeof(struct _xf_rule));
memcpy(&rules[r1], &rules[r2], sizeof(struct _xf_rule));
memcpy(&rules[r2], &rule, sizeof(struct _xf_rule));

fl_freeze_form(rule_obj->Rules_Form);
fl_clear_browser(rule_obj->Rule_List);
for (i = 0; i < MAX_RULES; i++) {
	if (rules[i].action == 0)
		continue;

	fl_add_browser_line(rule_obj->Rule_List, rules[i].name);
				}
fl_select_browser_line(rule_obj->Rule_List, k - 1);
fl_unfreeze_form(rule_obj->Rules_Form);
changed = 1;
return;
}

void Rule_Down_Call(FL_OBJECT *obj, long param)
{
struct _xf_rule rule;
int r1, r2;
int i;
int k = fl_get_browser(rule_obj->Rule_List);

 if (k == fl_get_browser_maxline(rule_obj->Rule_List))
        return;

r1 = rule_by_name((char *)fl_get_browser_line(rule_obj->Rule_List, k));
r2 = rule_by_name((char *)fl_get_browser_line(rule_obj->Rule_List, k + 1));

if ((r1 == -1) || (r2 == -1))
        return;

memcpy(&rule, &rules[r1], sizeof(struct _xf_rule));
memcpy(&rules[r1], &rules[r2], sizeof(struct _xf_rule));
memcpy(&rules[r2], &rule, sizeof(struct _xf_rule));

fl_freeze_form(rule_obj->Rules_Form);
fl_clear_browser(rule_obj->Rule_List);
for (i = 0; i < MAX_RULES; i++) {
	if (rules[i].action == 0)
		continue;

	fl_add_browser_line(rule_obj->Rule_List, rules[i].name);
				}
fl_select_browser_line(rule_obj->Rule_List, k + 1);
fl_unfreeze_form(rule_obj->Rules_Form);
changed = 1;
return;
}

void Rule_Common_Field_Call(FL_OBJECT *obj, long param)
{
char fname[16], *p;

 strncpy(fname, fl_get_choice_text(obj), 15);
 if ((p = strrchr(fname, '%')) != NULL)
	*p = '\0';

 fl_set_input(rule_obj->Rule_Field, fname);
 r_modified = 1;
}

void Rule_Dumb_Call(FL_OBJECT *obj, long param)
{
 r_modified = 1;
}

void Rule_List_Call(FL_OBJECT *obj, long param)
{
 if (r_modified && display_msg(MSG_QUEST,"rule was modified", "store changes?"))
	Rule_Modify_Call(NULL, param);

 r_modified = 0;
 display_rule(rule_by_name((char *)fl_get_browser_line(obj,fl_get_browser(obj))));
}

void Rule_Add_Call(FL_OBJECT *obj, long param)
{
int i, k, l;

 if (rule_by_name((char *)fl_get_input(rule_obj->Rule_Name)) != -1) {
	display_msg(MSG_WARN, "add", "Rule with name %s already exists", fl_get_input(rule_obj->Rule_Name));
	return; }

 r_modified = 0;

 for (i = 0; i < MAX_RULES; i++) {
	if (rules[i].action == 0) {
		if (compose_rule(i) != -1) {

		fl_freeze_form(rule_obj->Rules_Form);
		fl_clear_browser(rule_obj->Rule_List);

		l = 0;
		for (k = 0; k < MAX_RULES; k++) {
		  if (rules[k].action == 0)
			continue;

		  fl_add_browser_line(rule_obj->Rule_List, rules[k].name);
		  l++;
		  if (k == i) {
			fl_select_browser_line(rule_obj->Rule_List, l);
			fl_set_browser_topline(rule_obj->Rule_List, l); 
			}
				}

		fl_unfreeze_form(rule_obj->Rules_Form);
 		changed = 1;
					   }
		else {
		 display_msg(MSG_WARN, "add rule", "Invalid rule");
		 rules[i].name[0] = '\0';
		     }
		return;
				}
				}

 display_msg(MSG_WARN, "add rule", "Can not add more then %d rules", MAX_RULES);
}

void Rule_Delete_Call(FL_OBJECT *obj, long param)
{
int i;
int k = fl_get_browser(rule_obj->Rule_List);

 if (k == 0)
	return;

i = rule_by_name((char *)fl_get_browser_line(rule_obj->Rule_List, k));

if (i == -1)
	return;

 r_modified = 0;
 fl_delete_browser_line(rule_obj->Rule_List, k);
 rules[i].action = 0;
 rules[i].name[0] = '\0';

#ifdef HS_REGEX
 regfree(&rules[i].rx);
#endif

 changed = 1;

 display_rule(0);
 fl_select_browser_line(rule_obj->Rule_List, 1);
}

void Rule_Modify_Call(FL_OBJECT *obj, long param)
{
int i;

 i = rule_by_name((char *)fl_get_input(rule_obj->Rule_Name));

 if (i == -1)
	return;

 if (compose_rule(i) != -1) {
	fl_replace_browser_line(rule_obj->Rule_List, i + 1, rules[i].name);
	changed = 1; 	    }
 else {
	display_msg(MSG_WARN, "modify", "Invalid rule");
	display_rule(i);
      }

 r_modified = 0;
}

void Rule_Save_Call(FL_OBJECT *obj, long param)
{
 changed = 0;
 Rule_Modify_Call(NULL, param);
 save_rules();
}

void Rule_Folder_Call(FL_OBJECT *obj, long param)
{
 struct _mail_folder *mf;
 fl_freeze_form(rule_obj->Rules_Form);
 if (!(mf = choose_folder())) {
 	fl_unfreeze_form(rule_obj->Rules_Form);
	return; }

 fl_unfreeze_form(rule_obj->Rules_Form);
 fl_set_input(rule_obj->Rule_Fold_Move, get_folder_full_name(mf));
 r_modified = 1;
}

void Rule_File_Call(FL_OBJECT *obj, long param)
{
char *fname;

 fl_freeze_form(rule_obj->Rules_Form);
 fl_set_fselector_title("Choose vacation file");
 if (!(fname = (char *)fl_show_file_selector("Vacation","","","vacation.txt"))) {
 	fl_unfreeze_form(rule_obj->Rules_Form);
	return; }

 fl_unfreeze_form(rule_obj->Rules_Form);
 fl_set_input(rule_obj->Rule_Vac, fname);
 r_modified = 1;
}

void Rule_Addr_Call(FL_OBJECT *obj, long param)
{
 struct _mail_addr *addr;
 fl_freeze_form(rule_obj->Rules_Form);
 if (!(addr = addr_book(NULL, NULL, 0))) {
 	fl_unfreeze_form(rule_obj->Rules_Form);
	return; }

 fl_unfreeze_form(rule_obj->Rules_Form);
 fl_set_input(rule_obj->Rule_Fwd, addr->addr);
 r_modified = 1;
}

void Rule_Load_Call(FL_OBJECT *obj, long param)
{
 if (changed) {
	if (display_msg(MSG_QUEST, "Do you want to save changes?", "")) {
	 changed = 0;
         save_rules(); }
	}

 load_rules();
 r_modified = 0;
}

void Rule_Move_To_Call(FL_OBJECT *obj, long param)
{
fl_activate_object(rule_obj->Rule_Fold_Move);
fl_activate_object(rule_obj->Rule_Folder);
fl_deactivate_object(rule_obj->Rule_Fwd);
fl_deactivate_object(rule_obj->Rule_Addr);
fl_deactivate_object(rule_obj->Rule_Vac);
fl_deactivate_object(rule_obj->Rule_Vac_Keep);
fl_deactivate_object(rule_obj->Rule_File);
r_modified = 1;
}

void Rule_Fwd_To_Call(FL_OBJECT *obj, long param)
{
fl_deactivate_object(rule_obj->Rule_Fold_Move);
fl_deactivate_object(rule_obj->Rule_Folder);
fl_activate_object(rule_obj->Rule_Fwd);
fl_activate_object(rule_obj->Rule_Addr);
fl_deactivate_object(rule_obj->Rule_Vac);
fl_deactivate_object(rule_obj->Rule_Vac_Keep);
fl_deactivate_object(rule_obj->Rule_File);
r_modified = 1;
}

void Rule_Discard_Call(FL_OBJECT *obj, long param)
{
fl_deactivate_object(rule_obj->Rule_Fold_Move);
fl_deactivate_object(rule_obj->Rule_Folder);
fl_deactivate_object(rule_obj->Rule_Fwd);
fl_deactivate_object(rule_obj->Rule_Addr);
fl_deactivate_object(rule_obj->Rule_Vac);
fl_deactivate_object(rule_obj->Rule_Vac_Keep);
fl_deactivate_object(rule_obj->Rule_File);
r_modified = 1;
}

void Rule_Vacation_Call(FL_OBJECT *obj, long param)
{
fl_activate_object(rule_obj->Rule_File);
fl_deactivate_object(rule_obj->Rule_Vac);
fl_activate_object(rule_obj->Rule_Vac_Keep);
fl_deactivate_object(rule_obj->Rule_Fold_Move);
fl_deactivate_object(rule_obj->Rule_Folder);
fl_deactivate_object(rule_obj->Rule_Fwd);
fl_deactivate_object(rule_obj->Rule_Addr);
r_modified = 1;
}

int
compose_rule(num)
int num;
{
int i;
char *p;

#ifdef HS_REGEX
char errbuf[2048];
int regerr;
int regflags;
#endif

 if (num >= MAX_RULES)
	return -1;

 p = (char *)fl_get_input(rule_obj->Rule_Name);
 if (!p || (strlen(p) < 2) || (strlen(p) > 15)) {
	display_msg(MSG_WARN, "rule", "rule name must have 2-15 characters");
	return -1;
					        }

 for (i = 0; i < strlen(p); i++) {
	if (!isalnum(p[i])) {
		display_msg(MSG_WARN, "rule", "rule name can contain only characters or digits");
		return -1;
			    }
				 }

 strcpy(rules[num].name, p);

#ifdef HS_REGEX
        regfree(&rules[num].rx);

	regflags = REG_NOSUB | REG_EXTENDED;
	if (rules[i].flags & CASE_I)
		regflags |= REG_ICASE;

        if ((regerr = regcomp(&rules[num].rx, fl_get_input(rule_obj->Rule_Match), regflags))) {
                regerror(regerr, &rules[num].rx, errbuf, sizeof errbuf);
                display_msg(MSG_WARN, "Invalid regular expression", "%s", errbuf);
                regfree(&rules[num].rx);
                return -1;
                     }
#endif

 rules[num].flags = 0;

 strncpy(rules[num].tmatch, fl_get_input(rule_obj->Rule_Match), 64);
 strncpy(rules[num].fmatch, fl_get_input(rule_obj->Rule_Field), MAX_FIELD_NAME_LEN );

 if (!fl_get_button(rule_obj->Rule_Case))
	rules[num].flags |= CASE_I;

 if (fl_get_button(rule_obj->Rule_Save_Addr))
	rules[num].flags |= R_SAVE_ADDR;

 if (fl_get_button(rule_obj->Rule_Move_To)) {
	rules[num].action = R_MOVE;
	strncpy(rules[num].data, fl_get_input(rule_obj->Rule_Fold_Move), 64);
					}
 else 
 if (fl_get_button(rule_obj->Rule_Discard)) {
	rules[num].action = R_DISCARD;
	strcpy(rules[num].data, TRASH); }
 else
 if (fl_get_button(rule_obj->Rule_Fwd_To)) {
	rules[num].action = R_FORWARD;
	strncpy(rules[num].data, fl_get_input(rule_obj->Rule_Fwd), 64);
					}
 else
 if (fl_get_button(rule_obj->Rule_Vacation)) {
	rules[num].action = R_VACATION;
	strncpy(rules[num].data, fl_get_input(rule_obj->Rule_Vac), 64);
	if (fl_get_button(rule_obj->Rule_Vac_Keep))
		rules[num].flags |= R_VAC_KEEP;
					}
 else   {
	rules[num].action = R_MOVE;
	strcpy(rules[num].data, INBOX);
	}

 return 0;
}

void
display_rule(num)
int num;
{
 if (num >= MAX_RULES)
	return;

 if (rules[num].action == 0)
	return;

 fl_set_input(rule_obj->Rule_Name, rules[num].name);
 fl_set_input(rule_obj->Rule_Match, rules[num].tmatch);
 fl_set_input(rule_obj->Rule_Field, rules[num].fmatch);

 fl_set_input(rule_obj->Rule_Fold_Move, "");
 fl_set_input(rule_obj->Rule_Fwd, "");

 fl_set_button(rule_obj->Rule_Move_To, 0);
 fl_set_button(rule_obj->Rule_Discard, 0);
 fl_set_button(rule_obj->Rule_Fwd_To, 0);
 fl_set_button(rule_obj->Rule_Vacation, 0);

 if (rules[num].flags & CASE_I)
 	fl_set_button(rule_obj->Rule_Case, 0);
 else
 	fl_set_button(rule_obj->Rule_Case, 1);

 if (rules[num].flags & R_SAVE_ADDR)
 	fl_set_button(rule_obj->Rule_Save_Addr, 1);
 else
 	fl_set_button(rule_obj->Rule_Save_Addr, 0);

 if (rules[num].flags & R_VAC_KEEP)
 	fl_set_button(rule_obj->Rule_Vac_Keep, 1);
 else
 	fl_set_button(rule_obj->Rule_Vac_Keep, 0);

 switch(rules[num].action) {
	case R_MOVE:
		fl_set_button(rule_obj->Rule_Move_To, 1);
		fl_set_input(rule_obj->Rule_Fold_Move, rules[num].data);

		fl_activate_object(rule_obj->Rule_Fold_Move);
		fl_activate_object(rule_obj->Rule_Folder);
		fl_deactivate_object(rule_obj->Rule_Addr);
		fl_deactivate_object(rule_obj->Rule_Fwd);
		fl_deactivate_object(rule_obj->Rule_Vac);
		fl_deactivate_object(rule_obj->Rule_Vac_Keep);
		fl_deactivate_object(rule_obj->Rule_File);
	break;

	case R_DISCARD:
		fl_set_button(rule_obj->Rule_Discard, 1);

		fl_deactivate_object(rule_obj->Rule_Fold_Move);
		fl_deactivate_object(rule_obj->Rule_Folder);
		fl_deactivate_object(rule_obj->Rule_Fwd);
		fl_deactivate_object(rule_obj->Rule_Addr);
		fl_deactivate_object(rule_obj->Rule_Vac);
		fl_deactivate_object(rule_obj->Rule_Vac_Keep);
		fl_deactivate_object(rule_obj->Rule_File);
	break;

	case R_FORWARD:
		fl_set_button(rule_obj->Rule_Fwd_To, 1);
		fl_set_input(rule_obj->Rule_Fwd, rules[num].data);

		fl_deactivate_object(rule_obj->Rule_Fold_Move);
		fl_deactivate_object(rule_obj->Rule_Folder);
		fl_activate_object(rule_obj->Rule_Fwd);
		fl_activate_object(rule_obj->Rule_Addr);
		fl_deactivate_object(rule_obj->Rule_Vac);
		fl_deactivate_object(rule_obj->Rule_Vac_Keep);
		fl_deactivate_object(rule_obj->Rule_File);
	break;

	case R_VACATION:
		fl_set_button(rule_obj->Rule_Vacation, 1);
		fl_set_input(rule_obj->Rule_Vac, rules[num].data);

		fl_activate_object(rule_obj->Rule_File);
		fl_deactivate_object(rule_obj->Rule_Vac);
		fl_activate_object(rule_obj->Rule_Vac_Keep);

		fl_deactivate_object(rule_obj->Rule_Fold_Move);
		fl_deactivate_object(rule_obj->Rule_Folder);
		fl_deactivate_object(rule_obj->Rule_Fwd);
		fl_deactivate_object(rule_obj->Rule_Addr);
	break;

	default:
		rules[num].action = R_MOVE;
		fl_set_button(rule_obj->Rule_Move_To, 1);
		fl_set_input(rule_obj->Rule_Fold_Move, rules[num].data);
	break;
			}
 
}

void
rules_conf()
{
int i;

 if (ready) {
	XRaiseWindow(fl_display, rule_obj->Rules_Form->window);
        return; }

 ready = 1;
 changed = 0;
 r_modified = 0;

 rule_obj = create_form_Rules_Form();

 for (i = 0; i < MAX_RULES; i++) {
	if (rules[i].action == 0)
		continue;

	fl_add_browser_line(rule_obj->Rule_List, rules[i].name);
				}

 fl_clear_choice(rule_obj->Rule_Common_Field);

 fl_addto_choice(rule_obj->Rule_Common_Field, "Message");
 fl_addto_choice(rule_obj->Rule_Common_Field, "Header");
 fl_addto_choice(rule_obj->Rule_Common_Field, "Recipients");
 fl_addto_choice(rule_obj->Rule_Common_Field, "Body%l");

 fl_addto_choice(rule_obj->Rule_Common_Field, "Subject");
 fl_addto_choice(rule_obj->Rule_Common_Field, "From");
 fl_addto_choice(rule_obj->Rule_Common_Field, "To");
 fl_addto_choice(rule_obj->Rule_Common_Field, "Cc");
 fl_addto_choice(rule_obj->Rule_Common_Field, "Bcc");
 fl_addto_choice(rule_obj->Rule_Common_Field, "Sender");

 fl_set_choice(rule_obj->Rule_Common_Field, 1);

 fl_set_button(rule_obj->Rule_Move_To, 1);
 display_rule(0);
 fl_select_browser_line(rule_obj->Rule_List, 1);
 
 fl_set_input_return(rule_obj->Rule_Match, FL_RETURN_CHANGED);
 fl_set_input_return(rule_obj->Rule_Field, FL_RETURN_CHANGED);
 fl_set_input_return(rule_obj->Rule_Fold_Move, FL_RETURN_CHANGED);
 fl_set_input_return(rule_obj->Rule_Fwd, FL_RETURN_CHANGED);
 fl_set_input_return(rule_obj->Rule_Vac, FL_RETURN_CHANGED);

 fl_show_form(rule_obj->Rules_Form, FL_PLACE_CENTER, FL_TRANSIENT, "Rules");
 fl_do_only_forms();

 if (changed) {
	if (display_msg(MSG_QUEST, "Do you want to save changes?", "")) {
	 changed = 0;
         save_rules(); }
	}

 fl_hide_form(rule_obj->Rules_Form);
 fl_free_form(rule_obj->Rules_Form);
 fl_free(rule_obj);
 rule_obj = NULL;

 ready = 0;
 return;         

}

