/*
 * Copyright(c) 1995,1996 by Gennady B. Sorokopud (gena@NetVision.net.il)
 *
 * This software can be freely redistributed and modified for 
 * non-commercial purposes as long as above copyright
 * message and this permission notice appear in all
 * copies of distributed source code and included as separate file
 * in binary distribution.
 *
 * Any commercial use of this software requires author's permission.
 *
 * This software is provided "as is" without expressed or implied
 * warranty of any kind.
 * Under no circumstances is the author responsible for the proper
 * functioning of this software, nor does the author assume any
 * responsibility for damages incurred with its use.
 *
 */

/* $Id: umail.h,v 1.86 1996/12/22 13:56:56 gena Exp $
 */

#ifndef		UMAIL_H
#define		UMAIL_H

#ifdef __linux__
#define __USE_BSD_SIGNAL
#endif

#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <signal.h>
#if __STDC__ || defined(USE_STDARG)
#include <stdarg.h>
#else
#include <varargs.h>
#endif
#include <ctype.h>
#include <sys/time.h>
#include <sys/types.h>
#include <pwd.h>
#include <forms.h>

#include "version.h"

#define SLEN		256
#define PLEN		MAXPATHLEN

#define TIME_DBLCLICK	500000
#define TIME_DBLCLICK_MIN	150000

#ifndef TRUE
#define	FALSE		(0)
#define	TRUE		(!FALSE)
#endif

#define SHIFT 0x01
#define CTRL 0x02

extern char conf_name[255];
extern char sender_name[127];
extern char homedir[255];
extern char user_n[32];
extern FL_FORM *main_form;

/* busy indicator */
extern u_int busy;

extern u_int supress_errors;

/* debug */
extern u_int debug;

#define STAT_LEN        0x00
#define DATE_LEN        0x01
#define SIZE_LEN        0x02
#define FROM_LEN        0x03
#define SUBJ_LEN        0x04
extern int show_msg[5];

#define FOLD_LEN        0x00
#define FNUM_LEN        0x01
#define FUNR_LEN        0x02
extern int show_fld[3];

/* read ~/.xfdefaults file */
void read_xfdefaults();

/* save everything to configuration file */
/* returns -1 on error 0 on success */
extern int b_savecfg();
/* load contents of configuration file */
/* returns -1 on error 0 on success */
extern int b_loadcfg(char *cfg_file_name);
/* set special flags for specific variable */
extern void b_setcfg_flags(char *line, int flags);
#define DONT_STORE      0x01	/* parameter should not be stored in cfg file */
#define CF_UNSECURE     0x02 	/* parameter may not be sent in bug report */
#define	CF_NOTCHANGED	0x04	/* configuration will be considered unchanged */
				/* even if parameter has been changed */
/* store configuration in file. If nopwd > 0 skip all unsecure parameters */
/* returns -1 on error and 0 on success */
int save_cfg_file(char *file, int nopwd);

/* add/replace line in configuration file */
/* returns -1 on error 0 on success */
extern int b_putcfg_str(char *cfg_file_name, char *line, char *contents);
/* delete line from configuration file */
/* returns -1 on error 0 on success */
extern int b_delcfg_str(char *cfg_file_name, char *line);
/* get line from config file */
/* returns NULL on error */
extern char *b_getcfg_str(char *cfg_file_name, char *line, char *def);
/* add/replace number in configuration file */
/* returns -1 on error 0 on success */
extern int b_putcfg_int(char *cfg_file_name, char *line, int contents);
/* get number from config file */
/* returns def on error */
extern int b_getcfg_int(char *cfg_file_name, char *line, int def);
/* check if varname exists in the configuration file */
/* returns 1 if yes and 0 if no */
extern int cfg_exist(char *varname);
/* checks version of configuration file */
/* returns 1 if don't match */
int cfg_check_version();

struct _mail_msg *next_msg(int ur, struct _mail_msg *cmsg);
struct _mail_msg *prev_msg(int ur, struct _mail_msg *cmsg);
struct _mail_msg *delete_msg(struct _mail_msg *cmsg);
struct _mail_msg *move_message(struct _mail_msg *cmsg,struct _mail_folder *fld);

/* check if timer operation is permitted. returns 1 if not and 0 if yes */
int is_iconized();

/* retrieve new mail */
void get_new_mail(int local_only);

/* return folder line suitable for placing in browser */
char *get_fld_line(struct _mail_folder *folder);

/*
 * Theese are definitions of functions which
 * return non-integer type and defined and used in
 * different files
 */
extern char *b_rmlsp();
extern FL_FORM *view_form(struct _mail_msg *msg, int view);
/* check if double click occured. id1 and id2 are unique for every different */
/* control. return 1 if double click and 0 if not */
extern int double_click(int id1, int id2);
extern int display_msg (int, const char *, const char *, ...);
#define MSG_FATAL	0x00	/* fatal error , program will terminate */
#define MSG_QUEST	0x01	/* ask yes/no question */
#define MSG_WARN	0x02	/* warning , program will continue */
#define MSG_STAT	0x04	/* display messages in status bar */
#define MSG_MSG 	0x05	/* display messages */
#define	MSG_LOG		0x06	/* message to be displayed in log window */

extern void fl_edit_error(char *fmt, ...);
/* set cursor of for for to cursor. If cursor = 0 restores default cursor */
/* should be called with cursor = -1 at startup for initalization */
void set_cursor(FL_FORM *form, int cursor);

/* current folder */
extern struct _mail_folder *current_folder;

int view_msg(struct _mail_msg *msg, int openwin);
int edit_msg(struct _mail_msg *msg);
void set_edit_cursor(struct _mail_msg *msg, int cx, int cy);
void msg_form_shutdown();
void msg_form_init();

void move_messages(struct _mail_folder *fld);
void copy_messages(struct _mail_folder *fld);

extern FL_OBJECT       *inc_timer, *send_timer, *trash_timer, *ttip_timer, *imap_timer, *pgp_timer;
extern FL_OBJECT       *menu_misc;
#define	TTIP_DELAY	1	/* delay before displaying a tooltip */
extern	void cb_ttip_timer(FL_OBJECT *obj, long param);

/* edit message icon pixmap */
extern Pixmap icon_MailEdit, icon_MailEdit_sh;

/* address book icon pixmap */
extern Pixmap icon_ABook, icon_ABook_sh;

void sig_timers();

/* faces (picons) staff */
char *xbm_to_face();
char *get_msg_face(struct _mail_msg *msg);
char *find_face(struct _mail_addr *addr, int *ftype);
void update_faces(struct _mail_msg *msg);
#define	FACE_XPM	0x01
#define	FACE_XBM	0x02
#define	FACEFIELD	"X-Face"

/* this functions redraws different parts of the main window */
extern void redraw_folders();
extern void redraw_fld(struct _mail_folder *fld);
extern void redraw_fld_line(struct _mail_folder *fld);
extern void redraw_msg(struct _mail_msg *msg);
extern void redraw_fld_win();
extern void restore_fld_win();
extern void redraw_msg_win();

extern FL_OBJECT *fl_find_last(FL_FORM* form, int find, FL_Coord mx, FL_Coord my);
extern void show_button_descr(FL_OBJECT *obj);

void add_msg_addr(struct _mail_msg *msg);

void pop_account();
void imap_account();

int abortpressed();

char *getmyhostname();
void update_title();

extern int ttips;	/* show tooltips */
/* xforms even handler */
int post_handler(FL_OBJECT *ob, int event,FL_Coord mx, FL_Coord my, int key, void *xev);

/* address book definitions */

#define	MAX_BOOKS		10
#define MAX_ADDR_IN_BOOK	1023
#define	MAX_ADDR_IN_ALIAS	100
#define	MAX_ABOOK_NAMELEN	16

/* address book entry */
typedef struct _abook_entry {
struct _mail_addr *addr;	/* addresses */
struct _abook_entry *next;	/* next entry */
char	*description;		/* short description */
u_int	type;			/* type of entry */
#define	ADDR_SINGLE	0x00	/* single address */
#define	ADDR_ALIAS	0x01	/* alias entry */
} abook_entry;

typedef struct _abook {
struct _abook_entry *entries;	/* address book entries */
char name[16];			/* address book name */
u_int	flags;			/* special flags */
#define	ABOOK_LOCAL	0x00	/* local address book */
#define	ABOOK_REMOTE	0x01	/* remote address book */
} abook;

#ifdef	USE_SOUND
extern int play_sound(char *soundfile, int volume);
#define AUDIO_DEVICE	"/dev/audio"
#endif

struct _proc_info;

/* function to be called from SIGCHLD handler */
typedef void (*child_handle)(struct _proc_info *info);

typedef struct _proc_info {
int pid;			/* process ID */
int status;			/* process exit status */
u_int wait;			/* how to wait for process */
#define	WAIT_ASYNC	0x00	/* don't wait, call handle at exit */
#define	WAIT_BG		0x01	/* wait , but make GUI work at the same time */
#define	WAIT_STOP	0x02	/* wait and stop program execution */
child_handle handle;		/* handler function */
child_handle init;		/* initialization function */
void *u_data;			/* any user data */
long ul_data;			/* any user data */
int ifd;			/* input file descriptor for child process */
int fd_in[2],			/* communication pipes for stdin, stdout   */
    fd_out[2],			/* and stderr */
    fd_err[2];
} proc_info;

#define	MAX_CHILDREN	0x0f	/* Can not run more then 15 child processes at the same time */

/* initialize pinfo */
void init_pinfo(struct _proc_info *pinfo);
/* duplicate pinf structure */
struct _proc_info *dup_pid_info(struct _proc_info *pinfo);
/* execute child process */
extern int exec_child(char *str, struct _proc_info *pinfo);
void my_check_forms(struct _proc_info *pinfo);
/* check forms and file descriptor. if type = 0 check for read , 1 - for  */
/* write and 2 - for both read and write */
/* returns 0 if fd has data available and -1 on error */
int my_check_io_forms(int fd, int type);
void handler(); /* SIGCHLD handler */

void init_procs();
void close_procs();

char *get_sign_file(struct _mail_msg *msg);
char *pup_choose_sign(FL_FORM *form);
extern	FL_FORM	*signform;

void cb_help_button(FL_OBJECT *obj, long param);
void help(char *chapter);

void my_deactivate();
void my_activate();

/* special xforms defaults */
typedef struct _xfdef {
int popfsize, popfstyle, popcursor;
int popbgcol, popfgcol;
} xfdef;

extern struct _xfdef xfdefs;
extern int xofft, yofft;

/* PGP functions and definitions */
extern int pgp_action(char* inbuf, int mode, char* passphrase, char* recp);
extern char* input_passphrase();
#define ENCODE 1	/* encode with PGP */
#define DECODE 2	/* decode with PGP */
#define SIGN   4	/* sign with PGP signature */
#define VERIFY 8	/* Verify PGP signature */
#define	EXTKEY 16	/* Extract public key */
#define	ADDKEY 32	/* Add public key to message */

/* file viewer */
int file_view(char *file, struct _proc_info *pinfo);

char *strcasestr(const char *s, const char *find, int scase);

#endif /* UMAIL_H */
