/*
  Xinvest Copyright 1995-97 Mark Buser,
  All Rights Reserved.

  Permission is hereby granted to copy
  and freely distribute copies of this
  program for non-commercial purposes
  without fee, provided that this notice
  appears in all copies.
  
  All redistributions must be in their
  entirety as originally distributed.
  Feel free to modify Xinvest, but
  modified versions may not be distributed
  without prior consent of the author.
  
  This software is provided 'as-is'
  without any express or implied warranty.
  In no event will the author be held
  liable for any damages resulting from
  the use of this software.

  $Revision: 2.5 $ $Date: 1997/05/23 22:15:57 $
*/

/* Change the month the calendar is displaying */

#include <stdio.h>

#include <Xm/Form.h>
#include <Xm/Label.h>
#include <Xm/PushB.h>
#include <Xm/RowColumn.h>
#include <Xm/Text.h>

#include "account.h"
#include "calendar.h"
#include "nav.h"

/*-------------------------------------------------------------
** Forward Declarations
*/
static void   day_highlight ( Widget, int, XmAnyCallbackStruct * );
static Widget get_widget_day ( int date );
static int    day_number ( int year, int month, int day );
static int    days_in_month( int year, int month );

/*-------------------------------------------------------------
** Extern Declarations
*/
extern Month months[];
extern Widget Gaincalrow;
extern Widget Gainyeartext;

/* Remember the current settings */
static int month = -1;
static int date = -1;
static int year = -1;

void displayCal ( int m, int d, int y)
{
  extern Widget Gainmonthoptmenu;
  Widget cascade, menu;
  WidgetList buttons;

  char   year_str[5];
  XmString month_str;

  month = m;
  date = d;
  year = y;

  /* Displays month, date */
  set_month ( NULL, month-1, NULL);

  /* Display year */
  sprintf ( year_str, "%4.2d", year );
  XmTextSetString( Gainyeartext, year_str );

  /* Set option menu history to this month */
  XtVaGetValues ( Gainmonthoptmenu, XmNsubMenuId, &menu, NULL );
  XtVaGetValues ( menu, XmNchildren, &buttons, NULL );
  XtVaSetValues ( Gainmonthoptmenu,
                  XmNmenuHistory, buttons[ month-1],
                  NULL);

  /* Set option menu to display this month */
  month_str = XmStringCreateLocalized (
                months[ month-1].name );

  cascade = XmOptionButtonGadget (Gainmonthoptmenu);
  XtVaSetValues ( cascade,
                  XmNlabelString, month_str,
                  NULL );
  XmStringFree ( month_str );
}

  
/* 
** Callbacks 
*/

/* ARGSUSED */
void year_updown ( Widget menu_item, int updown, XmAnyCallbackStruct *cbs )
{
  int account = activeAccount();
  NAV *navp;
  char   year_str[5];

  getAccount ( account, ACCOUNT_NAV, &navp );
  getNav ( navp, NAV_YEAR, &year);
  getNav ( navp, NAV_MONTH, &month);

  if (updown == 1) /* up */
    year++;
  else             /* down */
    year--;

  setNav ( navp, NAV_YEAR, &year);

  sprintf ( year_str, "%4.2d", year );
  XmTextSetString( Gainyeartext, year_str );

  set_month ( NULL, month-1, NULL);
}

/* ARGSUSED */
static void day_highlight ( Widget day_widget, int next_day, 
                            XmAnyCallbackStruct *cbs )
{
  static Widget current_day = NULL;

  int account = activeAccount();
  NAV *navp;

  /* next_day is only non-zero if we are dealing with the same month.
  ** In this case its safe to set current_day.  Otherwise current_day
  ** may have been destroyed when we weren't looking.
  */ 

  if ( current_day != day_widget ) { /* different date, change highlight */
    if ( next_day && current_day != NULL)
      XtVaSetValues ( current_day, XmNshadowThickness, 0, NULL ); 
    current_day = day_widget;
    XtVaSetValues ( current_day, XmNshadowThickness, 2, NULL ); 

    /* Set new date */
    XtVaGetValues ( current_day, 
                    XmNuserData, &date, 
                    NULL );

    if (account) {
      getAccount ( account, ACCOUNT_NAV, &navp );
      setNav ( navp, NAV_DAY, &date);
      accountMatchDate (account, navToLDate (navp)); 
    }
  }
}

/* ARGSUSED */
void set_month ( Widget menu_item, int item_no, XmAnyCallbackStruct *cbs )
{
  int account = activeAccount();
  NAV *navp;

  static int last_year = -1;
  static int last_month = -1;
  static int last_date = 0;

  char text[BUFSIZ];
  char *dayname[] = { "Su", "Mo", "Tu", "We", "Th", "Fr", "Sa" };

  int i, j, m, tot, day;
  Widget day_widget;

  /* uncreated widgets to zero */
  if ( last_month == -1 && last_year == -1) {
    for (i = 0; i < 11; i++)
      months[i].form = (Widget) NULL; 
  }

  if ( item_no == last_month && year == last_year ) { /* same month and year */
    day_widget = get_widget_day ( date );
    day_highlight ( day_widget, (int)date, NULL );
    return;                                 
  }

  if ( last_month >= 0 && months[last_month].form )  {
    XtUnmanageChild ( months[last_month].form ); /* unmanage previous month */
    XtDestroyWidget ( months[last_month].form );
  }


  /* Get new month */
  month = item_no +1;                   
  last_month = item_no;
  last_date = date;
  last_year = year;

  if (account) {
    getAccount ( account, ACCOUNT_NAV, &navp);
    setNav ( navp, NAV_MONTH, &month );
  }
 
  /* Create the month form and dates pushbuttons */
  months[last_month].form = XtVaCreateWidget ( "Gainmonthform", 
                       xmRowColumnWidgetClass, Gaincalrow,
                       XmNentryAlignment,      XmALIGNMENT_CENTER,
                       XmNorientation,         XmHORIZONTAL,
                       XmNnumColumns,          7,
                       XmNpacking,             XmPACK_COLUMN,
                       NULL );

  for ( day = 0; day < XtNumber( dayname) ; day++ ) 
     XtVaCreateManagedWidget ( dayname[day], xmLabelWidgetClass, 
                               months[last_month].form, NULL );

  /* calculate dates of the month */
  m = day_number ( year, month, 1);
  tot = days_in_month ( year, month );

  /* guard against date greater than allowed for this month */
  if ( tot < date ) {
    date = 1;
    last_date = date;
    if (account)
      setNav ( navp, NAV_DAY, &date);
  }

  for ( day = i = 0; i < 6; i++ ) {
    for ( j = 0; j < 7; j++, m += ( j > m && --tot > 0 ) ) {
        char *name;
        if ( j != m || tot < 1 )
           name = "  ";
        else {
           sprintf ( text, "%2d", ++day );
           name = text;
        }
  
        /* Set margin to two more than default so that when the shadow
        ** is incremented from 0 to 2 there is room to keep the button 
        ** the same size.  recomputeSize being false stops the button
        ** from shrinking/growing on shadowThickness change.
        */
        months[last_month].dates[i][j] = XtVaCreateManagedWidget ( name, 
                 xmPushButtonWidgetClass, months[last_month].form,
                 XmNuserData,             day,
                 XmNsensitive,            (j % 7 == m && tot > 0),
                 XmNshadowThickness,      0,
                 XmNrecomputeSize,        False,
                 XmNmarginHeight,         4,
                 XmNmarginWidth,          4,
                 NULL );
        XtAddCallback ( months[last_month].dates[i][j],
                        XmNactivateCallback, 
                        (XtCallbackProc) day_highlight, 
                        (XtPointer) day );
         
    }
    m = 0;
  }

  XtManageChild ( months[last_month].form);

  XtVaSetValues ( months[last_month].dates[5][6],
                  XmNshadowThickness, 2,
                  XmNmappedWhenManaged, False,
                  NULL );

  day_widget = get_widget_day ( date );
  day_highlight ( day_widget, (int) NULL, NULL );
}

static int mtbl[] = { 0,31,59,90,120,151,181,212,243,273,304,334,365 };

static int days_in_month( year, month )
int year, month;
{
   int days;

   days = mtbl[month] - mtbl[month - 1];
   if ( month == 2 && year % 4 == 0 && (year % 100 != 0 || year % 400 == 0) )
     days++;
  
   return days;
}

static int day_number (year, month, day )
int year, month, day;
{
  long days_ctr;

  year -= 1900;
  days_ctr = ((long)year * 365L) + ((year + 3) / 4);
  days_ctr += mtbl[month -1] + day + 6;

  year += 1900;
  if ( month > 2 && year % 4 == 0 && (year % 100 != 0 || year % 400 == 0) )
    days_ctr++;

  return (int) (days_ctr % 7L);
}

static Widget get_widget_day ( int d )
{
  int i, j, day;
  Widget fallback;

  for (i=0; i < 6; i++)
    for (j=0; j < 7; j++) {
        XtVaGetValues ( months[month-1].dates[i][j], 
                        XmNuserData, &day,
                        NULL);
        if ( day == d )
           return (months[month-1].dates[i][j]);
        if ( day == 1 )
           fallback = months[month-1].dates[i][j];
    }
  /* fallback to day one if something bad happens */
  return ( fallback );
}
