/*
  Xinvest is copyright 1995 by Mark Buser.

  Permission is hereby granted to copy
  and freely distribute copies of this
  program for non-commercial purposes
  without fee, provided that this notice
  appears in all copies.
  
  All redistributions must be in their
  entirety as originally distributed.
  Feel free to modify Xinvest, but
  modified versions may not be distributed
  without prior consent of the author.
  
  This software is provided 'as-is'
  without any express or implied warranty.
  In no event will the author be held
  liable for any damages resulting from
  the use of this software.

  $Revision: 2.0 $ $Date: 1996/09/14 13:19:37 $
*/
#include <stdio.h>
#include <Xm/Xm.h>
#include "color.h"

/* This must match order of enum Colors in color.h */
static char *colors[] = { 
  NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* user app colors */
  "Red", "Green", "Grey60", "White", "Black"      /* local named colors */
};
static Pixel color[XtNumber(colors)];


static Colormap _cmap;
Colormap GetColormap()
{
  return _cmap;
}

Colormap NewColormap( Display *dpy, Colormap cmap)
{
   Colormap new = XCopyColormapAndFree ( dpy, cmap );
   if ( new == (Colormap)BadAlloc ||
        new == (Colormap)BadColor) {
      return (Colormap)NULL;
   }
   _cmap = new;
   return new;
}



Pixel GetColor ( int which )
{
  return ( color[which] );
} 

/* 
** Allocate all colors in color table.  If it fails, create a 
** new colormap and return it.  Return NULL if any failures.
*/
Colormap InitColor(Widget top, Colormap cmap, Pixel *appColor) 
{

  XColor actual, exact;
  Display *dpy = XtDisplay (top);
  int i;
  int how_many = 0;

  /* Just copy app level color resources */
  if (appColor)
    for (i = USER0; i <= USER7; i++)
      color[i] = appColor[i];

  /* The remainder are hardcoded color names */
  do {
    if ( !XAllocNamedColor( dpy, cmap, colors[i], &actual, &exact) ) {

      if (how_many == 1)
        return (Colormap) NULL;

      fprintf (stderr, "Warning: could not allocate color \"%s\".\n", colors[i]);
      fprintf (stderr, "Warning: trying to allocate private colormap.\n");
      if ( (cmap = NewColormap( dpy, cmap )) == (Colormap)NULL)
        return (Colormap) NULL;
      how_many++;

    } else {
      /* Successful, add to color table */
      color[i] = actual.pixel;
      i++;
      how_many = 0;
    }
  } while (i < XtNumber(colors));

  _cmap = cmap;
  return cmap;
}
