/*
 * XIRC
 * DCC Support
 */

#include "xirc.h"

void init_dcc(void)
{
    fl_set_object_label(dcc_form->DCCFile1, "No file waiting or being sent.");
    fl_set_object_label(dcc_form->DCCFile2, "No file waiting or being sent.");
    fl_set_object_label(dcc_form->DCCFile3, "No file waiting or being sent.");
}

void add_dcc(int type, char *nic, char *site, char *fname, int port, int size)
{
    if (type == DCC_GET) {
	int i;

	for(i = 0; i < 3; i++) {
	    if (files[i].sock == -1) {
		struct protoent *p;
		struct sockaddr_in sin;
		struct hostent *hostent;
		char err[80], host[80], line[160];
		int flag = 1;

		bzero((char *) &sin, sizeof(sin));
		sin.sin_family = AF_INET;
		sin.sin_port = htons(port);

		bzero(host, 80);
		sscanf(site, "%*[^@]@%[^\r\n]\r\n", host);
		hostent = gethostbyname(host);

		bzero(err, 80);
		if (hostent == NULL) {
		    sprintf(err, "*** Error looking up hostname %s", host);
		    add_line(main_form->MainBrowser, err, 4);
		    return;
		}

		bcopy(hostent->h_addr, (char *) &sin.sin_addr, hostent->h_length);

		p = getprotobyname("tcp");
		if ((files[i].sock = socket(AF_INET, SOCK_STREAM, p->p_proto)) < 0) {
		    sprintf(err, "*** Could not create socket: %s", strerror(errno));
		    add_line(main_form->MainBrowser, err, 4);
		    return;
		}

		if (connect(files[i].sock, (struct sockaddr *) &sin, sizeof(sin)) < 0) {
		    sprintf(err, "*** Could not connect: %s", strerror(errno));
		    add_line(main_form->MainBrowser, err, 4);
		    return;
		}

		sprintf(err, "*** DCC Connection to %s established.", nic);
		add_line(main_form->MainBrowser, err, 1);

		bzero(line, 160);
		sprintf(line, "%-19.19s %-8d %-7d %-10d %-13.13s %-8.8s",
			fname, size, 0, 0, "RUNNING", "0%");

		if (i == 0)
		    fl_set_object_label(dcc_form->DCCFile1, line);
		else if (i == 1)
		    fl_set_object_label(dcc_form->DCCFile2, line);
		else if (i == 2)
		    fl_set_object_label(dcc_form->DCCFile3, line);

		strcpy(files[i].filename, fname);
		files[i].progress = 0;
		files[i].type = DCC_GET;
		files[i].status = 1;
		files[i].size = size;

		ioctl(files[i].sock, FIONBIO, &flag);

		FD_SET(files[i].sock, &readset);
		FD_SET(files[i].sock, &activeset);

		return;
	    }
	}
    }
}

int exists(char *filename)
{
    int fd;

    if ((fd = open(filename, O_RDONLY)) == -1)
	return 0;
    else {
	close(fd);
	return 1;
    }
}

void handle_dcc(int pos)
{
#ifdef	NO_MMAP
    if ((files[pos].sock) && (files[pos].type == DCC_GET) &&
	(files[pos].status == 1)) {
	char data[2048];
	int read_data;

	if (files[pos].fd == 0) {
	    if (exists(files[pos].filename)) {
		if (fl_show_question("", "A file by this name already exists.  Do you want to choose another?", "")) {
		    bzero(files[pos].filename, 600);
		    sprintf(files[pos].filename, "%s", fl_show_file_selector("File to save as", "", "*", ""));
		}
	    }

	    if (files[pos].filename[0] != 0) {
		files[pos].fd = open(files[pos].filename, O_WRONLY | O_CREAT, 0777);
		add_line(main_form->MainBrowser, "*** DCC GET Link Established", 1);
		files[pos].start_time = time(NULL);
 		fl_show_form(dcc_form->DCCForm, FL_PLACE_MOUSE,
			FL_FULLBORDER, "DCC Status Window");
	    } else {
		add_line(main_form->MainBrowser, "*** DCC GET cancelled.", 1);

		bzero(files[pos].filename, 600);
		files[pos].sock = -1;
		files[pos].fd = 0;

		if (pos == 0)
		    fl_set_object_label(dcc_form->DCCFile1, "No file waiting or being sent.");
		else if (pos == 1)
		    fl_set_object_label(dcc_form->DCCFile2, "No file waiting or being sent.");
		else if (pos == 2)
		    fl_set_object_label(dcc_form->DCCFile3, "No file waiting or being sent.");
	    }
	}

	read_data = read(files[pos].sock, (char *) &data, 2048);

	if (read_data <= 0) {
	    char err[80];

	    if (errno != 11) {
		bzero(err, 80);
		if (errno != 0) {
		    sprintf(err, "*** Connection to file %s lost: %s",
			files[pos].filename, strerror(errno));
		    add_line(main_form->MainBrowser, err, 4);
		} else {
		    time_t time_now = time(NULL);
		    int time_diff = time_now - files[pos].start_time;
		    float ks;

		    ks = ((float) files[pos].size / ((float) time_diff * 1000));
		    sprintf(err, "*** File %s completed.  %d seconds at %3.2fKb/Sec",
			files[pos].filename, time_diff, ks);
		    add_line(main_form->MainBrowser, err, 1);
		}

		close(files[pos].fd);

		FD_CLR(files[pos].sock, &readset);
		FD_CLR(files[pos].sock, &activeset);

		(void) shutdown(files[pos].sock, 2);
		close(files[pos].sock);

		bzero(files[pos].filename, 40);
		files[pos].sock = -1;
		files[pos].fd = 0;

		if (pos == 0)
		    fl_set_object_label(dcc_form->DCCFile1, "No file waiting or being sent.");
		else if (pos == 1)
		    fl_set_object_label(dcc_form->DCCFile2, "No file waiting or being sent.");
		else if (pos == 2)
		    fl_set_object_label(dcc_form->DCCFile3, "No file waiting or being sent.");
	    }
	} else {
	    if (write(files[pos].fd, data, read_data) < 0) {
		char err[80];

		bzero(err, 80);
		if (errno != 0) {
		    sprintf(err, "*** Connection to file %s lost: %s",
			files[pos].filename, strerror(errno));
		    add_line(main_form->MainBrowser, err, 4);
		} else {
		    time_t time_now = time(NULL);
		    int time_diff = time_now - files[pos].start_time;
		    float ks;

		    ks = ((float) files[pos].size / ((float) time_diff * 1000));
		    sprintf(err, "*** File %s completed.  %d seconds at %3.2fKb/Sec",
			files[pos].filename, time_diff, ks);
		    add_line(main_form->MainBrowser, err, 1);
		}

		close(files[pos].fd);

		FD_CLR(files[pos].sock, &readset);
		FD_CLR(files[pos].sock, &activeset);

		(void) shutdown(files[pos].sock, 2);
		close(files[pos].sock);

		bzero(files[pos].filename, 600);
		files[pos].sock = -1;
		files[pos].fd = 0;

		if (pos == 0)
		    fl_set_object_label(dcc_form->DCCFile1, "No file waiting or being sent.");
		else if (pos == 1)
		    fl_set_object_label(dcc_form->DCCFile2, "No file waiting or being sent.");
		else if (pos == 2)
		    fl_set_object_label(dcc_form->DCCFile3, "No file waiting or being sent.");
	    } else {
		float percentage;
		int total;
		char line[160], pcnt[10];

		files[pos].progress += read_data;

		total = htonl(files[pos].progress);
		(void) write(files[pos].sock, (char *) &total, sizeof(int));

		bzero(line, 160);
		bzero(pcnt, 10);

		percentage = (((float) files[pos].progress) / (float) files[pos].size) * (float) 100;
		sprintf(line, "%-19.19s %-8d %-7d %-10d %-13.13s %-d%%",
			files[pos].filename, files[pos].size, 0,
			files[pos].progress, "RUNNING", (int) percentage);

		if (pos == 0)
		    fl_set_object_label(dcc_form->DCCFile1, line);
		else if (pos == 1)
		    fl_set_object_label(dcc_form->DCCFile2, line);
		else if (pos == 2)
		    fl_set_object_label(dcc_form->DCCFile3, line);
	    }
	}
    }
#endif
}

void DCCCloseCallback(FL_OBJECT *obj, long dat)
{
    if (dsmode == 1) {
	fl_set_object_label(dcc_form->OptionType, " ");
	dsmode = 0;
    } else {
	fl_set_object_label(dcc_form->OptionType, "Close which file?");
	dsmode = 1;
    }
}

void DCCPauseCallback(FL_OBJECT *obj, long dat)
{
    if (dsmode == 2) {
	dsmode = 0;
	fl_set_object_label(dcc_form->OptionType, " ");
    } else {
	dsmode = 2;
	fl_set_object_label(dcc_form->OptionType, "Pause which file?");
    }
}

void close_dcc(int pos)
{
    char resp[160];

    bzero(resp, 160);
    close(files[pos].fd);

    FD_CLR(files[pos].sock, &readset);
    FD_CLR(files[pos].sock, &activeset);

    (void) shutdown(files[pos].sock, 2);
    close(files[pos].sock);

    sprintf(resp, "*** User DCC abort on file %s.", files[pos].filename);
    add_line(main_form->MainBrowser, resp, 1);

    bzero(files[pos].filename, 40);
    files[pos].sock = -1;
    files[pos].fd = 0;

    if (pos == 0)
	fl_set_object_label(dcc_form->DCCFile1, "No file waiting or being sent.");
    else if (pos == 1)
	fl_set_object_label(dcc_form->DCCFile2, "No file waiting or being sent.");
    else if (pos == 2)
	fl_set_object_label(dcc_form->DCCFile3, "No file waiting or being sent.");
}

void suspend_dcc(int pos)
{
    char resp[160], line[80];
    float percentage;

    bzero(resp, 160);

    if (files[0].status == 1)
	files[0].status = 2;
    else
	files[0].status = 1;

    percentage = (((float) files[pos].progress) / (float) files[pos].size) * (float) 100;
    sprintf(line, "%-19.19s %-8d %-7d %-10d %-13.13s %-d%%",
		files[pos].filename, files[pos].size, 0,
		files[pos].progress,
		(files[pos].status == 1) ? "RUNNING" : "PAUSED",
		(int) percentage);

    sprintf(resp, "*** File %s %s", files[pos].filename,
	(files[pos].status == 1) ? "resuming." : "paused.");

    add_line(main_form->MainBrowser, resp, 1);

    if (pos == 0)
	fl_set_object_label(dcc_form->DCCFile1, line);
    else if (pos == 1)
	fl_set_object_label(dcc_form->DCCFile2, line);
    else if (pos == 2)
	fl_set_object_label(dcc_form->DCCFile3, line);
}

void DCCFile1Callback(FL_OBJECT *obj, long dat)
{
    if (dsmode == 1) {
	if (files[0].filename[0] != 0) {
	    close_dcc(0);
	}
    } else if (dsmode == 2) {
	if (files[0].filename[0] != 0) {
	    suspend_dcc(0);
	}
    }

    fl_set_object_label(dcc_form->OptionType, " ");
    dsmode = 0;
}

void DCCFile2Callback(FL_OBJECT *obj, long dat)
{
    if (dsmode == 1) {
	if (files[1].filename[0] != 0) {
	    close_dcc(1);
	}
    } else if (dsmode == 2) {
	if (files[1].filename[0] != 0) {
	    suspend_dcc(1);
	}
    }

    fl_set_object_label(dcc_form->OptionType, " ");
    dsmode = 0;
}

void DCCFile3Callback(FL_OBJECT *obj, long dat)
{
    if (dsmode == 1) {
	if (files[2].filename[0] != 0) {
	    close_dcc(2);
	}
    } else if (dsmode == 2) {
	if (files[2].filename[0] != 0) {
	    suspend_dcc(2);
	}
    }

    fl_set_object_label(dcc_form->OptionType, " ");
    dsmode = 0;
}

void DCCDismissCallback(FL_OBJECT *obj, long dat)
{
    fl_hide_form(dcc_form->DCCForm);
}
