/*
 * XIRC
 * IRC Names Parsing routines
 */

#include "xirc.h"

void irc_names_nick(char *);
void irc_names_privmsg(char *);
void irc_names_ping(char *);
void irc_names_join(char *);
void irc_names_part(char *);
void irc_names_quit(char *);
void irc_names_mode(char *);
void irc_names_error(char *);
void irc_names_kick(char *);

NAMECODETABLE irc_names_parse[] = {
    { "NICK", irc_names_nick },
    { "PRIVMSG", irc_names_privmsg },
    { "PING", irc_names_ping },
    { "JOIN", irc_names_join },
    { "PART", irc_names_part },
    { "QUIT", irc_names_quit },
    { "MODE", irc_names_mode },
    { "ERROR", irc_names_error },
    { "KICK", irc_names_kick },
    { NULL, NULL }
};

void irc_names_nick(char *str)
{
    char nic[160], site[160], newnick[160], ret[600];

    bzero(nic, 160);
    bzero(site, 160);
    bzero(newnick, 160);
    bzero(ret, 600);

    fl_freeze_form(main_form->MainForm);
    sscanf(str, ":%[^!]!%[^ ] %*[^ ] :%[^\r\n]\r\n", nic, site, newnick);
    if ((!strncmp(nick, nic, strlen(nic))) &&
	strlen(nic) == strlen(nick)) {
	if (newnick[0] != 0) {
	    char title[80];

	    bzero(title, 80);
	    sprintf(ret, "*** You are now known as %s", newnick);
	    add_line(main_form->MainBrowser, ret, 1);
	    names_refresh();
	    bzero(nick, 9);
	    strcpy(nick, newnick);
	    change_who(nic, newnick);
	}
    } else {
	if ((nic[0] != 0) && (newnick[0] != 0)) {
	    sprintf(ret, "*** %s is now known as %s", nic, newnick);
	    add_line(main_form->MainBrowser, ret, 1);
	    names_refresh();
	    change_who(nic, newnick);
	}
    }
    fl_unfreeze_form(main_form->MainForm);
}

void do_ctcp(char *nic, char *site, char *msg)
{
    char ret[600], rep[600];
    int is_action = 0;

    bzero(ret, 600);
    bzero(rep, 600);

    strcpy(msg, msg + 1);
    if (!strncmp(msg, "ACTION", 6)) {
	strcpy(msg, msg + 7);
	sprintf(ret, "* %s %s", nic, msg);
	fl_freeze_form(main_form->MainForm);
	fl_addto_browser(main_form->MainBrowser, ret);
	fl_unfreeze_form(main_form->MainForm);
	is_action = 1;
    } else if (!strncmp(msg, "CLIENTINFO", 10)) {
	sprintf(ret, "NOTICE %s :Commands are: FINGER, HELP, PING, and VERSION\r\n", nic);
	sprintf(rep, "*** CTCP CLIENTINFO request from %s", nic);
	add_line(main_form->MainBrowser, rep, 4);
    } else if (!strncmp(msg, "FINGER", 6)) {
	sprintf(ret, "NOTICE %s :XIRC version 0.0.1 (XForms) by Ken Hollis <khollis@bitgate.com>\r\n", nic);
	sprintf(rep, "*** CTCP FINGER request from %s", nic);
	add_line(main_form->MainBrowser, rep, 4);
    } else if (!strncmp(msg, "HELP", 4)) {
	sprintf(ret, "NOTICE %s :Help not available.\r\n", nic);
	sprintf(rep, "*** CTCP HELP request from %s", nic);
	add_line(main_form->MainBrowser, rep, 4);
    } else if (!strncmp(msg, "VERSION", 7)) {
	sprintf(ret, "NOTICE %s :XIRC 0.0.1 by Ken Hollis, written using the XForms Library.  Copyright (C) 1996, Bitgate Software.\r\n", nic);
	sprintf(rep, "*** CTCP VERSION request from %s", nic);
	add_line(main_form->MainBrowser, rep, 4);
    } else if (!strncmp(msg, "PING", 4)) {
	long tim = time(NULL);

	sprintf(ret, "NOTICE %s :%cPING %ld%c\r\n", nic, 0x01, tim, 0x01);
	sprintf(rep, "*** CTCP PING request from %s", nic);
	add_line(main_form->MainBrowser, rep, 4);
    } else if (!strncmp(msg, "DCC", 3)) {
	char type[80], arg[80], addr[80], port[80], size[80];
	int size_int, sock_int;

	bzero(type, 80);
	bzero(arg, 80);
	bzero(addr, 80);
	bzero(port, 80);
	bzero(size, 80);

	sscanf(msg, "DCC %[^ ] %[^ ] %[^ ] %[^ ] %[^\r\n]\r\n",
		type, arg, addr, port, size);

	if (!strcasecmp(type, "SEND")) {
	    sprintf(rep, "*** DCC %s request from %s(%s) (%s size %s)", type,
		nic, site, arg, size);

	    size_int = atoi(size);
	    sock_int = atoi(port);

	    add_line(main_form->MainBrowser, rep, 4);

	    add_dcc(DCC_GET, nic, site, arg, sock_int, size_int);
	}
    }

    if (!is_action)
	(void) write(irc, ret, strlen(ret));
}

void irc_names_privmsg(char *dat)
{
    char nick[160], site[160], bleah[160], msg[600], ret[820];

    bzero(nick, 160);
    bzero(site, 160);
    bzero(bleah, 160);
    bzero(msg, 600);
    bzero(ret, 820);

    fl_freeze_form(main_form->MainForm);
    sscanf(dat, ":%[^!]!%[^ ] %*[^ ] %[^ ] :%[^\r\n]\r\n", nick, site, bleah, msg);
    if (bleah[0] == '#') {
	if ((msg[0] == 0x01) && (msg[strlen(msg) - 1] == 0x01))
	    do_ctcp(nick, site, msg);
	else {
	    sprintf(ret, "<%s> %s", nick, msg);
	    fl_addto_browser(main_form->MainBrowser, ret);
	}
    } else {
	if ((msg[0] == 0x01) && (msg[strlen(msg) - 1] == 0x01))
	    do_ctcp(nick, site, msg);
	else {
	    sprintf(ret, "[%s:%s] %s", nick, site, msg);
	    add_line(main_form->MainBrowser, ret, 2);
	}
    }
    fl_unfreeze_form(main_form->MainForm);
}

void irc_names_ping(char *dat)
{
    char ret[600], site[160];

    bzero(ret, 600);
    bzero(site, 160);

    sscanf(dat, "%*[^ ] :%[^\r\n]\r\n", site);
    sprintf(ret, "PONG %s\r\n", site);

    (void) write(irc, ret, strlen(ret));
}

void irc_names_join(char *dat)
{
    char nic[160], site[160], ret[600];

    bzero(nic, 160);
    bzero(site, 160);
    bzero(ret, 600);

    fl_freeze_form(main_form->MainForm);
    sscanf(dat, ":%[^!]!%[^ ] %*[^\r\n]\r\n", nic, site);
    if ((!strncmp(nick, nic, strlen(nick))) &&
	(strlen(nick) == strlen(nic))) {
	char title[80];

	bzero(title, 80);
	sprintf(ret, "*** You have joined channel %s", channel);
	add_line(main_form->MainBrowser, ret, 1);

	clear_who();
	bzero(title, 80);
	sprintf(title, "WHO %s\r\n", channel);
	(void) write(irc, title, strlen(title));

	fl_show_form(ul_form->ULISTForm, FL_PLACE_MOUSE, FL_FULLBORDER,
		"Names Listing");
    } else {
	sprintf(ret, "*** %s (%s) has just joined channel %s", nic, site, channel);
	add_line(main_form->MainBrowser, ret, 1);
	add_names(nic);
	add_who(nic, site);
    }

    fl_unfreeze_form(main_form->MainForm);
}

void irc_names_part(char *dat)
{
    char nic[160], site[160], ret[600];

    bzero(nic, 160);
    bzero(site, 160);
    bzero(ret, 600);

    fl_freeze_form(main_form->MainForm);
    sscanf(dat, ":%[^!]!%[^ ] %*[^\r\n]\r\n", nic, site);

    if ((!strncmp(nic, nick, strlen(nick))) &&
	strlen(nic) == strlen(nick)) {
	char title[80];

	bzero(title, 80);
	sprintf(ret, "*** You have just left channel %s", channel);
	add_line(main_form->MainBrowser, ret, 1);
	bzero(channel, 80);

	fl_clear_browser(ul_form->NickBrowser);
	sprintf(title, "%s : No channel", nick);
	XStoreName(fl_get_display(), main_form->MainForm->window, title);

	clear_who();
    } else {
	sprintf(ret, "*** %s has just left channel %s", nic, channel);
	add_line(main_form->MainBrowser, ret, 1);
	remove_names(nic);
	remove_who(nic);
    }

    fl_unfreeze_form(main_form->MainForm);
}

int is_split(char *str)
{
    int spaces = 0, i;
    char site1[160], site2[160];

    bzero(site1, 160);
    bzero(site2, 160);

    for(i = 0; i < strlen(str); i++)
	if (str[i] == ' ')
	    spaces++;

    if (spaces == 1) {
	int s1 = 0, s2 = 0;

	sscanf(str, "%[^ ] %[^\r\n]\r\n", site1, site2);

	if (strstr(site1, "com") ||
	    strstr(site1, "us") ||
	    strstr(site1, "irc.") ||
	    strstr(site1, "edu") ||
	    strstr(site1, "gov"))
	    s1 = 1;

	if (strstr(site2, "com") ||
	    strstr(site2, "us") ||
	    strstr(site2, "irc.") ||
	    strstr(site2, "edu") ||
	    strstr(site2, "gov"))
	    s2 = 1;

	if ((s1 && s2) ||
	    (s1) || (s2))
	    return 1;
    }

    return 0;
}

void irc_names_quit(char *dat)
{
    char nic[160], site[160], msg[600], ret[900];

    bzero(nic, 160);
    bzero(site, 160);
    bzero(msg, 600);
    bzero(ret, 900);

    sscanf(dat, ":%[^!]!%[^ ] %*[^ ] :%[^\r\n]\r\n", nic, site, msg);

    if (is_split(msg)) {
	char server[80], split_from[80];

	bzero(server, 80);
	bzero(split_from, 80);

	sscanf(msg, "%[^ ] %[^\r\n]\r\n", server, split_from);

	sprintf(ret, "*** %s on server %s split from %s", nic, server, split_from);
    } else
	sprintf(ret, "*** Signoff: %s (%s)", nic, msg);

    fl_freeze_form(main_form->MainForm);
    add_line(main_form->MainBrowser, ret, 3);
    fl_unfreeze_form(main_form->MainForm);

    remove_names(nic);
    remove_who(nic);
}

void irc_names_mode(char *dat)
{
    char nic[160], site[160], mode[160], modeparms[400], ret[800];

    bzero(nic, 160);
    bzero(site, 160);
    bzero(mode, 160);
    bzero(modeparms, 400);
    bzero(ret, 800);

    if ((dat[0] == ':') && strstr(dat, "!")) {
	sscanf(dat, ":%[^!]!%[^ ] %*[^ ] %*[^ ] %[^ ] %[^\r\n]\r\n",
		nic, site, mode, modeparms);

	modeparms[strlen(modeparms) - 1] = 0;

	sprintf(ret, "*** Mode change \"%s %s\" on channel %s by %s",
		mode, modeparms, channel, nic);
	fl_freeze_form(main_form->MainForm);
	add_line(main_form->MainBrowser, ret, 1);
	fl_unfreeze_form(main_form->MainForm);

	names_refresh();

	/* If we have a channel locking on +nst, do it here! */
	if (irc_data.config1 & CONFIG_LOCKNST) {
	    bzero(ret, 800);
	    sprintf(ret, "MODE %s +nst\r\n", channel);
	    (void) write(irc, ret, strlen(ret));
	}
    }
}

void irc_names_error(char *dat)
{
    char msg[600], ret[660];

    bzero(msg, 600);
    bzero(ret, 660);

    sscanf(dat, "%*[^ ] :%[^\r\n]\r\n", msg);
    sprintf(ret, "*** %s", msg);

    fl_freeze_form(main_form->MainForm);
    add_line(main_form->MainBrowser, ret, 1);
    fl_unfreeze_form(main_form->MainForm);

    clear_who();
}

void irc_names_kick(char *dat)
{
    char nic[160], site[160], chan[160], newnick[160], msg[600], ret[1200];

    bzero(nic, 160);
    bzero(site, 160);
    bzero(chan, 160);
    bzero(newnick, 160);
    bzero(msg, 600);
    bzero(ret, 1200);

    fl_freeze_form(main_form->MainForm);

    sscanf(dat, ":%[^!]!%[^ ] %*[^ ] %[^ ] %[^ ] :%[^\r\n]\r\n",
	nic, site, chan, newnick, msg);
    if ((!strncmp(nick, newnick, strlen(newnick))) &&
	strlen(newnick) == strlen(nick)) {
	char title[80], rep[160];

	bzero(title, 80);
	sprintf(ret, "*** You have been kicked off channel %s by %s (%s)",
		chan, nic, msg);
	add_line(main_form->MainBrowser, ret, 1);
	fl_hide_form(ul_form->ULISTForm);

	/* If we can auto-join on a kick, then do it here! */
	if (irc_data.config1 & CONFIG_AUTOJOINONKICK) {
	    bzero(rep, 160);
	    sleep(1);
	    sprintf(rep, "JOIN %s\r\n", channel);
	    (void) write(irc, rep, strlen(rep));
	}

	clear_who();
    } else {
	sprintf(ret, "*** %s was kicked off channel %s by %s (%s)",
		newnick, channel, nic, msg);
	add_line(main_form->MainBrowser, ret, 1);
	remove_names(nic);
	remove_who(newnick);
    }

    fl_unfreeze_form(main_form->MainForm);
}
