/*
 * XIRC
 * Number replies
 */

#include "xirc.h"

void irc_nick(int, char *);
void irc_info(int, char *);
void irc_motd(int, char *);
void irc_names(int, char *);
void irc_whois(int, char *);
void irc_list(int, char *);
void irc_who(int, char *);
void irc_topic(int, char *);

NUMCODETABLE irc_nums_parse[] = {
    { ERR_NICKNAMEINUSE,	irc_nick },

    { RPL_MOTDSTART,		irc_motd },
    { RPL_MOTD,			irc_motd },
    { RPL_ENDOFMOTD,		irc_motd },

    { RPL_IRCINFO1,		irc_info },
    { RPL_IRCINFO2,		irc_info },
    { RPL_IRCINFO3,		irc_info },
    { RPL_LUSERCLIENT,		irc_info },
    { RPL_LUSEROP,		irc_info },
    { RPL_LUSERUNKNOWN,		irc_info },
    { RPL_LUSERCHANNELS,	irc_info },
    { RPL_LUSERME,		irc_info },

    { RPL_NAMREPLY,		irc_names },

    { RPL_WHOREPLY,		irc_who },

    { RPL_WHOISUSER,		irc_whois },
    { RPL_WHOISSERVER,		irc_whois },
    { RPL_WHOISOPERATOR,	irc_whois },
    { RPL_WHOISIDLE,		irc_whois },
    { RPL_WHOISCHANNELS,	irc_whois },

    { RPL_LIST,			irc_list },
    { RPL_LISTEND,		irc_list },

    { RPL_TOPIC,		irc_topic },
    { RPL_TOPICSETBY,		irc_topic },

    { 0, NULL }
};

void irc_nick(int opt, char *dat)
{
    char nic[80], ret[80], title[80];

    bzero(nic, 80);
    bzero(ret, 80);

    sscanf(dat, "%*[^ ] %[^ ] %*[^\r\n]\r\n", nic);

    switch(opt) {
	case 433:
	    bzero(title, 80);
	    sprintf(ret, "*** Nickname %s already in use.", nic);
	    add_line(main_form->MainBrowser, ret, 1);
	    bzero(nick, 9);
	    break;
    }
}

void irc_info(int opt, char *dat)
{
    char out[600];

    bzero(out, 600);

    switch(opt) {
	case RPL_IRCINFO1:
	case RPL_IRCINFO2:
	case RPL_IRCINFO3:
	case RPL_LUSERCLIENT:
	case RPL_LUSERME: {
	    char st[600];

	    bzero(st, 600);
	    sscanf(dat, "%*[^ ] :%[^\r\n]\r\n", st);
	    sprintf(out, "*** %s", st);
	    add_line(main_form->MainBrowser, out, 1);
	} break;

	case RPL_LUSERUNKNOWN: {
	    char st[600];

	    bzero(st, 600);
	    sscanf(dat, "%*[^ ] %[^ ] %*[^\r\n]\r\n", st);
	    sprintf(out, "*** %s unknown connection(s)", st);
	    add_line(main_form->MainBrowser, out, 1);
	} break;

	case RPL_LUSERCHANNELS: {
	    char st[600];

	    bzero(st, 600);
	    sscanf(dat, "%*[^ ] %[^ ] %*[^\r\n]\r\n", st);
	    sprintf(out, "*** %s channel(s) formed", st);
	    add_line(main_form->MainBrowser, out, 1);
	} break;
    }
}

void irc_motd(int opt, char *dat)
{
    char str[400], rep[480];

    bzero(str, 400);
    bzero(rep, 480);

    sscanf(dat, "%*[^:]:-%[^\r\n]\r\n", str);
    sprintf(rep, "*** %s", str);

    if (str[0] != 0) {
	fl_freeze_form(main_form->MainForm);
	add_line(main_form->MainBrowser, rep, 1);
	fl_unfreeze_form(main_form->MainForm);
    }
}

void irc_names(int bleah, char *dat)
{
    char ret[1200], sret[1260];

    bzero(ret, 1200);
    bzero(sret, 1260);

    sscanf(dat, "%*[^:]:%[^\r\n]\r\n", ret);
    add_names_list(ret);
}

void irc_whois(int opt, char *dat)
{
    char nic[80], user[160], host[160], realname[160], ret[600];

    bzero(nic, 80);
    bzero(user, 160);
    bzero(host, 160);
    bzero(realname, 160);
    bzero(ret, 600);

    switch(opt) {
	case 311: sscanf(dat, "%*[^ ] %[^ ] %[^ ] %[^ ] * :%[^\r\n]\r\n",
			nic, user, host, realname);
		  sprintf(ret, "*** %s is %s@%s (%s)", nic, user, host, realname);
		  break;

	case 312: sscanf(dat, "%*[^ ] %[^ ] %[^ ] :%[^\r\n]\r\n", nic,
			user, host);
		  sprintf(ret, "*** %s is on server %s (%s)", nic, user, host);
		  break;

	case 313: sscanf(dat, "%*[^ ] %[^ ] %*[^\r\n]\r\n", nic);
		  sprintf(ret, "*** %s is an IRC operator", nic);
		  break;

	case 317: sscanf(dat, "%*[^ ] %[^ ] %[^ ] %*[^\r\n]\r\n", nic, user);
		  sprintf(ret, "*** %s has been idle %s seconds", nic, user);
		  break;

	case 319: sscanf(dat, "%*[^ ] %[^ ] :%[^\r\n]\r\n", nic, user);
		  sprintf(ret, "*** %s is on channel(s) %s", nic, user);
		  break;
    }

    fl_addto_browser(text_form->TextBrowser, ret);
}

void irc_topic(int opt, char *dat)
{
    char newtopic[600], ret[660];

    bzero(newtopic, 600);
    bzero(ret, 660);

    switch(opt) {
	case RPL_TOPIC:
	    sscanf(dat, "%*[^ ] %*[^ ] :%[^\r\n]\r\n", newtopic);
	    sprintf(ret, "*** Topic for channel %s: %s", channel, newtopic);
	    break;
	case RPL_TOPICSETBY:
	    sscanf(dat, "%*[^ ] %*[^ ] %[^ ] %*[^\r\n]\r\n", newtopic);
	    sprintf(ret, "*** Topic for channel %s set by %s", channel, newtopic);
	    break;
    }

    add_line(main_form->MainBrowser, ret, 1);
}

void irc_list(int opt, char *dat)
{
    char num[80], chan[600], topic[480], rep[1200];

    bzero(num,80);
    bzero(chan, 600);
    bzero(topic, 480);
    bzero(rep, 1200);

    if (opt == 322) {
	sscanf(dat, "%*[^ ] %[^ ] %[^ ] :%[^\r\n]\r\n", chan, num, topic);
	sprintf(rep, "@f*** %-12.12s %-3.3s %s", chan, num, topic);
	fl_addto_browser(text_form->TextBrowser, rep);
    } else {
	int i;
	char reply[600];

	bzero(reply, 600);

	i = fl_get_browser_maxline(text_form->TextBrowser) - 1;

	sprintf(reply, "@f*** Total listing of %d channels.", i);
	fl_set_browser_fontsize(text_form->TextBrowser, 12);

	fl_addto_browser(text_form->TextBrowser, "@f***");
	fl_addto_browser(text_form->TextBrowser, reply);
	add_line(main_form->MainBrowser, "*** Userlist generation complete.", 1);

	fl_show_form(text_form->TextForm, FL_PLACE_CENTER|FL_FREE_SIZE,
		FL_FULLBORDER, "Channel List");
    }
}

void irc_who(int item, char *dat)
{
    switch(item) {
	case RPL_WHOREPLY: {
	    int i;
	    char em1[80], em2[80], nd[9];

	    bzero(em1, 80);
	    bzero(em2, 80);
	    bzero(nd, 9);

	    sscanf(dat, "%*[^ ] %*[^ ] %[^ ] %[^ ] %*[^ ] %[^ ] %*[^ ] :%*[^\r\n]\r\n",
		em1, em2, nd);

	    for(i = 0; i < 500; i++) {
		if (chanusers[i].email1[0] == 0) {
		    char disp[80];

		    bzero(disp, 80);
		    strcpy(chanusers[i].email1, em1);
		    strcpy(chanusers[i].email2, em2);
		    strcpy(chanusers[i].nick, nd);
		    break;
		}
	    }
	} break;
    }
}
