/*
 * XIRC
 * Options routines
 */

#include "xirc.h"

LOADDATATABLE irc_data;

int get_style_by_name(char *str)
{
    if (!strncasecmp(str, "boldface", 8))
	return(1);
    if (!strncasecmp(str, "italic", 6))
	return(2);
    if (!strncasecmp(str, "bolditalic", 10))
	return(3);
    if (!strncasecmp(str, "fixed", 5))
	return(4);

    return(0);
}

int get_color_by_name(char *str)
{
    if (!strncasecmp(str, "black", 5))
	return(1);
    if (!strncasecmp(str, "red", 3))
	return(2);
    if (!strncasecmp(str, "green", 5))
	return(3);
    if (!strncasecmp(str, "yellow", 6))
	return(4);
    if (!strncasecmp(str, "blue", 4))
	return(5);
    if (!strncasecmp(str, "purple", 6))
	return(6);
    if (!strncasecmp(str, "sky", 3))
	return(7);
    if (!strncasecmp(str, "white", 5))
	return(8);

    return(0);
}

void set_style_obj(FL_OBJECT *obj, int dat)
{
    switch(dat) {
	case 1:		fl_set_object_label(obj, "Boldface");	break;
	case 2:		fl_set_object_label(obj, "Italic");	break;
	case 3:		fl_set_object_label(obj, "BoldItalic");	break;
	case 4:		fl_set_object_label(obj, "Fixed");	break;
	default:	fl_set_object_label(obj, "Normal");	break;
    }
}

void set_color_obj(FL_OBJECT *obj, int dat)
{
    switch(dat) {
	case 1:		fl_set_object_label(obj, "Black");	break;
	case 2:		fl_set_object_label(obj, "Red");	break;
	case 3:		fl_set_object_label(obj, "Green");	break;
	case 4:		fl_set_object_label(obj, "Yellow");	break;
	case 5:		fl_set_object_label(obj, "Blue");	break;
	case 6:		fl_set_object_label(obj, "Purple");	break;
	case 7:		fl_set_object_label(obj, "Sky");	break;
	case 8:		fl_set_object_label(obj, "White");	break;
    }
}

char *get_pair_name(FONTPAIR pair)
{
    char *ret;

    ret = (char *) malloc(80);
    switch(pair.style) {
	case 0:	sprintf(ret, "normal");	break;
	case 1: sprintf(ret, "boldface"); break;
	case 2: sprintf(ret, "italic"); break;
	case 3: sprintf(ret, "bolditalic"); break;
	case 4: sprintf(ret, "fixed"); break;
    }

    switch(pair.color) {
	case 1: strcat(ret, ":black");		break;
	case 2: strcat(ret, ":red");		break;
	case 3: strcat(ret, ":green");		break;
	case 4: strcat(ret, ":yellow");		break;
	case 5: strcat(ret, ":blue");		break;
	case 6: strcat(ret, ":purple");		break;
	case 7: strcat(ret, ":sky");		break;
	case 8: strcat(ret, ":white");		break;
    }

    return(ret);
}

int split_get_style(char *str)
{
    char style[80];

    bzero(style, 80);
    sscanf(str, "%[^:\r\n]:%*[^\r\n]\r\n", style);

    return(get_style_by_name(style));
}

int split_get_color(char *str)
{
    char color[80];

    bzero(color, 80);
    sscanf(str, "%*[^:\r\n]:%[^\r\n]\r\n", color);

    return(get_color_by_name(color));
}

void load_options(void)
{
    FILE *file;

    bzero(irc_data.nick, 9);
    bzero(irc_data.realname, 80);
    bzero(irc_data.username, 80);
    bzero(irc_data.server, 80);
    irc_data.fontsize = 12;
    irc_data.config1 = 0;

    file = fopen("xirc.conf", "r");
    if (file) {
	char str[160], opt[160], line[320];

	while(!feof(file)) {
	    bzero(str, 160);
	    bzero(opt, 160);
	    bzero(line, 320);

	    fgets(line, 320, file);
	    sscanf(line, "%[^:\r\n]: %[^\r\n]\r\n", opt, str);

	    if (!strncasecmp(opt, "nick", 4)) {
		strcpy(irc_data.nick, str);
		bzero(nick, 9);
		strcpy(nick, str);
	    } else if (!strncasecmp(opt, "realname", 8))
		strcpy(irc_data.realname, str);
	    else if (!strncasecmp(opt, "username", 8))
		strcpy(irc_data.username, str);
	    else if (!strncasecmp(opt, "ircserver", 9)) {
		strcpy(irc_data.server, str);
		strcpy(server, irc_data.server);
	    } else if (!strncasecmp(opt, "fontsize", 8)) {
		if (atoi(str) != 0) {
		    fl_set_browser_fontsize(main_form->MainBrowser, atoi(str));
		    irc_data.fontsize = atoi(str);
		} else {
		    fl_set_browser_fontsize(main_form->MainBrowser, 12);
		    irc_data.fontsize = 12;
		}
	    } else if (!strncasecmp(opt, "servermsg", 9)) {
		char style[80], color[80];

		bzero(style, 80);
		bzero(color, 80);
		sscanf(str, "%[^:]:%[^\r\n]\r\n", style, color);
		irc_data.servermsg.style = get_style_by_name(style);
		irc_data.servermsg.color = get_color_by_name(color);
	    } else if (!strncasecmp(opt, "privatemsg", 10)) {
		char style[80], color[80];

		bzero(style, 80);
		bzero(color, 80);
		sscanf(str, "%[^:]:%[^\r\n]\r\n", style, color);
		irc_data.privatemsg.style = get_style_by_name(style);
		irc_data.privatemsg.color = get_color_by_name(color);
	    } else if (!strncasecmp(opt, "critservmsg", 11)) {
		char style[80], color[80];

		bzero(style, 80);
		bzero(color, 80);
		sscanf(str, "%[^:]:%[^\r\n]\r\n", style, color);
		irc_data.critservmsg.style = get_style_by_name(style);
		irc_data.critservmsg.color = get_color_by_name(color);
	    } else if (!strncasecmp(opt, "ctcpmsg", 7)) {
		char style[80], color[80];

		bzero(style, 80);
		bzero(color, 80);
		sscanf(str, "%[^:]:%[^\r\n]\r\n", style, color);
		irc_data.ctcpmsg.style = get_style_by_name(style);
		irc_data.ctcpmsg.color = get_color_by_name(color);
	    } else if (!strncasecmp(opt, "bantype", 7)) {
		if (!strncasecmp(str, "standard", 7))
		    irc_data.bantype = BAN_STANDARD;
		else if (!strncasecmp(str, "site", 4))
		    irc_data.bantype = BAN_SITE;
		else if (!strncasecmp(str, "screw", 5))
		    irc_data.bantype = BAN_SCREW;
		else if (!strncasecmp(str, "nick", 4))
		    irc_data.bantype = BAN_NICK;
	    } else if (!strncasecmp(opt, "locknst", 7)) {
		if (!strncasecmp(str, "on", 2))
		    irc_data.config1 |= CONFIG_LOCKNST;
	    } else if (!strncasecmp(opt, "autogetdcc", 10)) {
		if (!strncasecmp(str, "on", 2))
		    irc_data.config1 |= CONFIG_AUTOGETDCC;
	    } else if (!strncasecmp(opt, "autojoinonkick", 14)) {
		if (!strncasecmp(str, "on", 2))
		    irc_data.config1 |= CONFIG_AUTOJOINONKICK;
	    }
	}

	fclose(file);
    } else {
	fprintf(stderr, "No XIRC.CONF file in this directory.  Using defaults.\n");
	strcpy(irc_data.nick, DEFAULT_NICK);
	strcpy(nick, DEFAULT_NICK);
	strcpy(irc_data.realname, DEFAULT_REALNAME);
	strcpy(irc_data.username, DEFAULT_USERNAME);
	irc_data.servermsg.style = split_get_style(DEFAULT_SERVERMSG);
	irc_data.servermsg.color = split_get_color(DEFAULT_SERVERMSG);
	irc_data.privatemsg.style = split_get_style(DEFAULT_PRIVATEMSG);
	irc_data.privatemsg.color = split_get_color(DEFAULT_PRIVATEMSG);
	irc_data.critservmsg.style = split_get_style(DEFAULT_CRITSERVMSG);
	irc_data.critservmsg.color = split_get_color(DEFAULT_CRITSERVMSG);
	irc_data.ctcpmsg.style = split_get_style(DEFAULT_CTCPMSG);
	irc_data.ctcpmsg.color = split_get_color(DEFAULT_CTCPMSG);
	irc_data.bantype = BAN_STANDARD;
	irc_data.config1 = (CONFIG_LOCKNST | CONFIG_AUTOGETDCC);
    }
}

void save_options(void)
{
    FILE *file;

    file = fopen("xirc.conf", "w");

    if (file) {
	fprintf(file, "nick: %s\n", irc_data.nick);
	fprintf(file, "realname: %s\n", irc_data.realname);
	fprintf(file, "username: %s\n", irc_data.username);
	fprintf(file, "ircserver: %s\n", irc_data.server);
	fprintf(file, "fontsize: %d\n", irc_data.fontsize);
	fprintf(file, "servermsg: %s\n", get_pair_name(irc_data.servermsg));
	fprintf(file, "privatemsg: %s\n", get_pair_name(irc_data.privatemsg));
	fprintf(file, "critservmsg: %s\n", get_pair_name(irc_data.critservmsg));
	fprintf(file, "ctcpmsg: %s\n", get_pair_name(irc_data.ctcpmsg));
	fprintf(file, "bantype: ");

	if (bansel == 1) {
	    fprintf(file, "site\n");
	    irc_data.bantype = BAN_SITE;
	} else if (bansel == 2) {
	    fprintf(file, "nick\n");
	    irc_data.bantype = BAN_NICK;
	} else if (bansel == 3) {
	    fprintf(file, "screw\n");
	    irc_data.bantype = BAN_SCREW;
	} else if (bansel == 4) {
	    fprintf(file, "standard\n");
	    irc_data.bantype = BAN_STANDARD;
	} else if (bansel == 0) {
	    if (irc_data.bantype == BAN_STANDARD)
		fprintf(file, "standard\n");
	    else if (irc_data.bantype == BAN_SITE)
		fprintf(file, "site\n");
	    else if (irc_data.bantype == BAN_SCREW)
		fprintf(file, "screw\n");
	    else if (irc_data.bantype == BAN_NICK)
		fprintf(file, "nick\n");
	}

	fprintf(file, "locknst: %s\n", (irc_data.config1 & CONFIG_LOCKNST) ? "on" : "off");
	fprintf(file, "autogetdcc: %s\n", (irc_data.config1 & CONFIG_AUTOGETDCC) ? "on" : "off");
	fprintf(file, "autojoinonkick: %s\n", (irc_data.config1 & CONFIG_AUTOJOINONKICK) ? "on" : "off");
    }

    fclose(file);
}

void set_options(void)
{
    set_style_obj(mc_form->ServStyle, irc_data.servermsg.style);
    set_style_obj(mc_form->PrivStyle, irc_data.privatemsg.style);
    set_style_obj(mc_form->CServStyle, irc_data.critservmsg.style);
    set_style_obj(mc_form->CTCPStyle, irc_data.ctcpmsg.style);
    set_color_obj(mc_form->ServCol, irc_data.servermsg.color);
    set_color_obj(mc_form->PrvCol, irc_data.privatemsg.color);
    set_color_obj(mc_form->CSCol, irc_data.critservmsg.color);
    set_color_obj(mc_form->CTCPCol, irc_data.ctcpmsg.color);
}
