/*
 * XIRC
 * Main Routines
 */

#include "xirc.h"

FD_AboutForm *about_form;
FD_MainForm *main_form;
FD_StartForm *start_form;
FD_IRCForm *irc_form;
FD_FontForm *font_form;
FD_ULISTForm *ul_form;
FD_DCCForm *dcc_form;
FD_BanForm *ban_form;

int irc, is_connecting, bantype, dsmode, bansel;
time_t prevtime;
char channel[80], nick[80], server[80];
fd_set readset, activeset;
WHOLIST chanusers[500];
DCC_FILES files[3];

void set_chanstatus_info(void)
{
    time_t curtime = time(NULL);

    if (prevtime != curtime) {
	char formatted[80];
	char str[160];

	prevtime = curtime;
	bzero(formatted, 80);
	bzero(str, 160);

	strftime(formatted, 80, "%I:%M%p", localtime(&curtime));

	if (channel[0] == 0) {
	    sprintf(str, "[%s] %s - No channel", formatted, nick);
	    fl_set_object_label(main_form->ChanStatus, str);
	} else {
	    sprintf(str, "[%s] %s - %.10s", formatted, nick, channel);
	    fl_set_object_label(main_form->ChanStatus, str);
	}
    }
}

void clear_vars(void)
{
    int i;

    irc = -1;
    bzero(channel, 80);
    bzero(nick, 80);
    bzero(server, 80);

    is_connecting = 0;
    bantype = BAN_STANDARD;

    for (i = 0; i < 500; i++) {
	bzero(chanusers[i].email1, 80);
	bzero(chanusers[i].email2, 80);
	bzero(chanusers[i].nick, 9);
    }
}

void AboutCallback(FL_OBJECT *obj, long data)
{
    fl_hide_form(about_form->AboutForm);
}

void FileCallback(FL_OBJECT *obj, long data)
{
    int object = fl_get_menu(obj);

    switch(object) {
	case 3:		exit(0);

	default:	break;
    }
}

void set_personal(void)
{
    char serv[80], port[80];

    bzero(serv, 80);
    bzero(port, 80);

    fl_set_input(irc_form->IRCNickname, irc_data.nick);
    fl_set_input(irc_form->IRCUsername, irc_data.username);
    fl_set_input(irc_form->IRCRealname, irc_data.realname);

    if (irc_data.config1 & CONFIG_LOCKNST)
	fl_set_button(irc_form->LockModeNST, TRUE);
    if (irc_data.config1 & CONFIG_AUTOGETDCC)
	fl_set_button(irc_form->AutoDCCGet, TRUE);
    if (irc_data.config1 & CONFIG_AUTOJOINONKICK)
	fl_set_button(irc_form->AutoJoinOnKick, TRUE);

    sscanf(irc_data.server, "%[^:\r\n]:%[^\r\n]\r\n", serv, port);

    if (port[0] == 0)
	sprintf(port, "6667");

    fl_set_input(irc_form->IRCServerName, serv);
    fl_set_input(irc_form->IRCPortName, port);
}

void SettingsCallback(FL_OBJECT *obj, long data)
{
    int object = fl_get_menu(obj);

    switch(object) {
	case 1:		set_personal();
			fl_show_form(irc_form->IRCForm, FL_PLACE_MOUSE,
				FL_FULLBORDER, "XIRC Personal Settings");
			break;

	case 2:		fl_show_form(font_form->FontForm, FL_PLACE_MOUSE,
				FL_FULLBORDER, "XIRC Main Font Settings");
			break;

	case 3:		set_options();
			fl_show_form(mc_form->MCForm, FL_PLACE_MOUSE,
				FL_FULLBORDER, "XIRC Font/Color Settings");
			break;

	case 4:		set_ban_form();
			fl_show_form(ban_form->BanForm, FL_PLACE_MOUSE,
				FL_FULLBORDER, "XIRC Ban Settings");
			break;

	default:	break;
    }
}

void IRCInputCallback(FL_OBJECT *obj, long data)
{
    char bleah[600];

    bzero(bleah, 600);
    if (irc != -1) {
	strcpy(bleah, fl_get_input(main_form->IRCInputBox));
	handle_input(bleah);
    } else {
	fl_freeze_form(main_form->MainForm);
	add_line(main_form->MainBrowser, "*** You are not connected to a server.", 3);
	fl_unfreeze_form(main_form->MainForm);
    }

    fl_set_input(main_form->IRCInputBox, "");
}

void MainBrowserCallback(FL_OBJECT *obj, long data)
{
}

void WindowsCallback(FL_OBJECT *obj, long data)
{
    int object = fl_get_menu(obj);

    switch(object) {
	case 1:    fl_show_form(start_form->StartForm, FL_PLACE_MOUSE|FL_FREE_SIZE,
			FL_FULLBORDER, "XIRC 0.0.1 Main Window");
		   break;

	case 2:    fl_set_object_label(dcc_form->OptionType, " ");   
		   fl_show_form(dcc_form->DCCForm, FL_PLACE_MOUSE,
			FL_FULLBORDER, "DCC Status Window");
		   break;

	case 3:	   fl_show_form(ul_form->ULISTForm, FL_PLACE_MOUSE, 
			FL_FULLBORDER, "Names Listing");
		   break;

	default:   break;
    }
}

void DismissDebugCallback(FL_OBJECT *obj, long data)
{
    fl_hide_form(start_form->StartForm);
}

void HelpCallback(FL_OBJECT *obj, long data)
{
    int object = fl_get_menu(obj);

    switch(object) {
	case 1:		fl_show_form(about_form->AboutForm, FL_PLACE_CENTER,
				FL_FULLBORDER, "About XIRC");
			break;

	case 2:		fl_clear_browser(text_form->TextBrowser);
			fl_load_browser(text_form->TextBrowser, "UPDATES");
			fl_set_browser_fontsize(text_form->TextBrowser, 12);

			fl_show_form(text_form->TextForm, FL_PLACE_CENTER|FL_FREE_SIZE,
				FL_FULLBORDER, "Update Information");
			break;

	case 3:		fl_clear_browser(text_form->TextBrowser);
			fl_load_browser(text_form->TextBrowser, "xirc.userman");
			fl_set_browser_fontsize(text_form->TextBrowser, 12);

			fl_show_form(text_form->TextForm, FL_PLACE_CENTER|FL_FREE_SIZE,
				FL_FULLBORDER, "XIRC User Manual");
			break;

	case 4:		fl_clear_browser(text_form->TextBrowser);
			fl_load_browser(text_form->TextBrowser, "xfhelp.dat");
			fl_set_browser_fontsize(text_form->TextBrowser, 12);

			fl_show_form(text_form->TextForm, FL_PLACE_CENTER|FL_FREE_SIZE,
				FL_FULLBORDER, "XForms Help Manual");
			break;

	default:	break;
    }
}

void DebugCallback(FL_OBJECT *obj, long data)
{
}

void DismissIRCFormCallback(FL_OBJECT *obj, long data)
{
    fl_hide_form(irc_form->IRCForm);
}

void AcceptIRCFormCallback(FL_OBJECT *obj, long data)
{
    bzero(irc_data.nick, 9);
    bzero(irc_data.realname, 80);
    bzero(irc_data.username, 80);
    bzero(irc_data.server, 80);

    strcpy(irc_data.nick, fl_get_input(irc_form->IRCNickname));
    strcpy(irc_data.username, fl_get_input(irc_form->IRCUsername));
    strcpy(irc_data.realname, fl_get_input(irc_form->IRCRealname));
    sprintf(irc_data.server, "%s:%s", fl_get_input(irc_form->IRCServerName),
	fl_get_input(irc_form->IRCPortName));

    irc_data.config1 = 0;
    if (fl_get_button(irc_form->AutoJoinOnKick))
	irc_data.config1 |= CONFIG_AUTOJOINONKICK;
    if (fl_get_button(irc_form->LockModeNST))
	irc_data.config1 |= CONFIG_LOCKNST;
    if (fl_get_button(irc_form->AutoDCCGet))
	irc_data.config1 |= CONFIG_AUTOGETDCC;

    fl_hide_form(irc_form->IRCForm);
}

void do_connect(void)
{
    struct protoent *p;
    struct sockaddr_in sin;
    struct hostent *hostent = NULL;
    char srv[80], port[80];
    int flag = 1, prt;

    fl_addto_browser(start_form->DebugList, "XIRC: Starting do_connect");

    is_connecting = 0;

    bzero((char *) &sin, sizeof(sin));
    sin.sin_family = AF_INET;

    if (server[0] != 0)
	sscanf(server, "%[^:\r\n]:%[^\r\n]\r\n", srv, port);

    if (atoi(port) == 0) {
	sin.sin_port = htons(6667);
	prt = 6667;
    } else {
	sin.sin_port = htons(atoi(port));
	prt = atoi(port);
    }

    if (srv[0] == 0) {
	strcpy(srv, DEFAULT_IRC_SRV);
	fprintf(stderr, "No server specified.  Using the default: %s\n", DEFAULT_IRC_SRV);
    }

    hostent = gethostbyname(srv);
    if (hostent != NULL)
	bcopy(hostent->h_addr, (char *) &sin.sin_addr, hostent->h_length);

    if (hostent == NULL) {
	fl_show_alert("Hostent", "Can't get hostname for IRC server!", "", 0);
	clear_vars();
	return;
    } else
	fl_addto_browser(start_form->DebugList, "XIRC: Got name for IRC server");

    p = getprotobyname("tcp");
    if ((irc = socket(AF_INET, SOCK_STREAM, p->p_proto)) < 0) {
	fl_show_alert("Socket", "Can't get a socket!", "", 0);
	clear_vars();
	return;
    } else
	fl_addto_browser(start_form->DebugList, "XIRC: Socket open");

    /* Set IRC connection to non-blocking so we don't block up the
       stratosphere */
    (void) ioctl(irc, FIONBIO, &flag);

    if (connect(irc, (struct sockaddr *) &sin, sizeof(sin)) < 0) {
	if ((errno != EINPROGRESS) && (errno != EWOULDBLOCK)) {
	    fl_show_alert("Connect", "Couldn't connect!", "", 0);
	    add_line(main_form->MainBrowser, "*** Connection refused to server.", 3);
	    clear_vars();
	    return;
	} else {
	    char rep[80];

	    bzero(rep, 80);
	    fl_addto_browser(start_form->DebugList, "XIRC: Waiting for connection");
	    sprintf(rep, "*** Connecting to server %s port %d", srv, prt);
	    add_line(main_form->MainBrowser, rep, 1);
	    is_connecting = 1;
	}
    } else {
	char rep[80];

	/* Although it's EXTREMELY unlikely that a connection will happen
	   immediately unless the server is local, this precaution should
	   be taken, nonetheless. */

	bzero(rep, 80);
	sprintf(rep, "*** Connected to server %s port %d", srv, prt);
	add_line(main_form->MainBrowser, rep, 1);
	is_connecting = 2;
    }

    tcdrain(irc);
}

void stripnl(char *Str)
{
   char *P=(char *)rindex(Str,'\n');

   if (P) *P='\0';
   P=(char *)rindex(Str,'\r');
   if (P) *P='\0';
}

void input_handle(char *data)
{
    char site[80], status[80], otherdata[7800];
    int stat;

    bzero(site, 80);
    bzero(status, 80);
    bzero(otherdata, 7800);

    fl_addto_browser(start_form->DebugList, data);

    if (data[0] == ':')
	sscanf(data, ":%[^ ] %[^ ] %[^\r\n]\r\n", site, status, otherdata);
    else
	sscanf(data, "%[^ ] %[^\r\n]\r\n", status, otherdata);

    stat = atoi(status);

    if (stat == 0) {
	int x;

	if (status[0] != 0) {
	    for(x = 0; irc_names_parse[x].status != NULL; x++)
		if ((!strncasecmp(irc_names_parse[x].status, status, strlen(status))) &&
		    (irc_names_parse[x].status != NULL))
		    (irc_names_parse[x].function)(data);
	}
    } else {
	int x;

	for(x = 0; irc_nums_parse[x].status != 0; x++)
	    if (stat == irc_nums_parse[x].status)
		(irc_nums_parse[x].function)(stat, otherdata);
    }
}

void add_font(const char *str)
{
    fl_addto_browser(font_form->FontNameBox, str);
}

void clear_dcc_entries(void)
{
    int i;

    for(i = 0; i < 3; i++) {
	bzero(files[i].filename, 40);
	files[i].fd = 0;
	files[i].sock = -1;
	files[i].size = 0;
	files[i].progress = 0;
	files[i].type = 0;
	files[i].status = 0;
	files[i].start_time = (time_t) 0;
    }
}

void main(int argc, char *argv[])
{
    char hostname[80], reply[160];
    int nick_spec = 0, serv_spec = 0;

    fl_initialize(&argc, argv, "XIRC", 0, 0);

    bzero(hostname, 80);
    bzero(reply, 160);

#ifndef	FQDN
    (void) gethostname(&hostname, 1);
#else
    strcpy(hostname, FQDN);
#endif

    irc = -1;
    bzero(nick, 9);
    bzero(server, 80);

    fontsize = 10;
    fontname = 1;
    dsmode = 0;
    bansel = 0;

    about_form = create_form_AboutForm();
    main_form = create_form_MainForm();
    start_form = create_form_StartForm();
    irc_form = create_form_IRCForm();
    font_form = create_form_FontForm();
    mc_form = create_form_MCForm();
    msg_form = create_form_MSGForm();
    text_form = create_form_TextForm();
    ul_form = create_form_ULISTForm();
    dcc_form = create_form_DCCForm();
    ban_form = create_form_BanForm();

    load_options();
    fl_set_browser_fontsize(ul_form->NickBrowser, 12);

    if (argv[1] != NULL) {
	if (!strncasecmp(argv[1], "-help", 5)) {
	    give_help();
	    exit(0);
	} else if (strstr(argv[1], ".")) {
	    bzero(server, 80);
	    strcpy(server, argv[1]);
	    serv_spec = 1;
	} else {
	    bzero(irc_data.nick, 9);
	    strcpy(irc_data.nick, argv[1]);
	    nick_spec = 1;
	}
    }

    if ((argv[2] != NULL) && (argc == 2)) {
	if (strstr(argv[2], ".")) {
	    if (!serv_spec) {
		bzero(server, 80);
		strcpy(server, argv[2]);
	    } else {
		fprintf(stderr, "You cannot specify two servers on the same line!\n");
		exit(0);
	    }
	} else {
	    if (!nick_spec) {
		bzero(irc_data.nick, 9);
		strcpy(irc_data.nick, argv[2]);
	    } else {
		fprintf(stderr, "You cannot specify two nicks on the same line!\n");
		exit(0);
	    }
	}
    }

    signal(SIGPIPE, SIG_IGN);

    fl_set_menu(main_form->FileTitle,"Open|Save%l|Quit");
    fl_set_menu(main_form->SettingsTitle,"Change Personal|"
		"Change Font|Change Colors/Styles|Change Ban Definition");
    fl_set_menu(main_form->WindowsTitle,"Debugging%l|DCC Status|Names in channel");
    fl_set_menu(main_form->HelpTitle,"About|Update info%l|User manual|XForms Lib Help");

    fl_set_form_minsize(main_form->MainForm, 320, 330);
    fl_set_object_gravity(main_form->FileTitle, NorthWestGravity, NorthGravity);
    fl_set_object_gravity(main_form->SettingsTitle, NorthWestGravity, NorthGravity);
    fl_set_object_gravity(main_form->WindowsTitle, NorthWestGravity, NorthGravity);
    fl_set_object_gravity(main_form->HelpTitle, NorthEastGravity, NorthGravity);
    fl_set_object_gravity(main_form->IRCInputBox, SouthWestGravity, SouthEastGravity);
    fl_set_object_gravity(main_form->MainBrowser, NorthWestGravity, SouthEastGravity);
    fl_set_object_gravity(main_form->MenuBox, NorthWestGravity, NorthEastGravity);
    fl_set_object_gravity(main_form->ChanStatus, SouthWestGravity, SouthEastGravity);
    fl_set_object_gravity(text_form->TextBrowser, NorthWestGravity, SouthEastGravity);
    fl_set_object_gravity(text_form->DisBrowser, SouthEastGravity, SouthGravity);
    fl_set_object_gravity(text_form->DismissTextBack, SouthEastGravity, SouthGravity);

    fl_set_object_resize(main_form->FileTitle, FL_RESIZE_NONE);
    fl_set_object_resize(main_form->SettingsTitle, FL_RESIZE_NONE);
    fl_set_object_resize(main_form->WindowsTitle, FL_RESIZE_NONE);
    fl_set_object_resize(main_form->HelpTitle, FL_RESIZE_NONE);
    fl_set_object_resize(main_form->IRCInputBox, FL_RESIZE_NONE);
    fl_set_object_resize(main_form->MainBrowser, FL_RESIZE_NONE);
    fl_set_object_resize(main_form->MenuBox, FL_RESIZE_NONE);
    fl_set_object_resize(main_form->ChanStatus, FL_RESIZE_NONE);
    fl_set_object_resize(text_form->TextBrowser, FL_RESIZE_NONE);
    fl_set_object_resize(text_form->DisBrowser, FL_RESIZE_NONE);
    fl_set_object_resize(text_form->DismissTextBack, FL_RESIZE_NONE);

    fl_enumerate_fonts(add_font, 1);
    fl_addto_browser(font_form->FontSizeBox, "8  (small)");
    fl_addto_browser(font_form->FontSizeBox, "10 (normal)");
    fl_addto_browser(font_form->FontSizeBox, "11 (scaled)");
    fl_addto_browser(font_form->FontSizeBox, "12 (normal)");
    fl_addto_browser(font_form->FontSizeBox, "13 (scaled)");
    fl_addto_browser(font_form->FontSizeBox, "14 (normal)");
    fl_addto_browser(font_form->FontSizeBox, "18 (large)");
    fl_addto_browser(font_form->FontSizeBox, "24 (huge)");
    fl_addto_browser(font_form->FontSizeBox, "30 (scaled)");

    fl_select_browser_line(font_form->FontNameBox, 1);
    fl_select_browser_line(font_form->FontSizeBox, 2);

    fl_set_bitmap_file(about_form->AboutBitmap, "ryoohki.xbm");

    fl_show_form(main_form->MainForm, FL_PLACE_MOUSE|FL_FREE_SIZE,
		FL_FULLBORDER, "XIRC 0.0.1 Main Window");

    set_chanstatus_info();
    set_options();
    init_dcc();
    clear_dcc_entries();

    do_connect();

    bzero(channel, 80);

    FD_ZERO(&activeset);
    FD_SET(irc, &activeset);

    do {
	struct timeval t;
	int x;

	t.tv_sec = 0;
	t.tv_usec = 1000;

	bcopy(&activeset, &readset, sizeof(readset));

	(void) fl_check_forms();

	/* Set channel information */
	set_chanstatus_info();

	/* If the connection is active and we are listening to the
	   open IRC server (if there is one) then we go through our
	   normal select loop. */

	if (is_connecting == 3) {
	    if (irc != -1) {
		if (select(256, &readset, NULL, NULL, &t) < 0)
		    fl_addto_browser(start_form->DebugList, "Select error!");
	    }
	}

	/* Same goes here - now we're checking to see if the IRC server
	   has sent us data, and if there's data pending on the read
	   socket. */

	if ((irc != -1) && (is_connecting == 3)) {
	    if (FD_ISSET(irc, &readset)) {
		int size;
		char buffer[8000];

		bzero(buffer, 8000);
		size = read(irc, buffer, 8000);
		if (size <= 0) {
		    fl_addto_browser(start_form->DebugList, "LOOP: Connect closed");

		    (void) shutdown(irc, 2);
		    close(irc);

		    FD_CLR(irc, &readset);
		    FD_CLR(irc, &activeset);

		    clear_vars();

		    add_line(main_form->MainBrowser, "*** Your server disconnected.  Type \"/server\" to select a new one.", 3);
		} else {
		    char *EolPtr = NULL, *BolPtr = NULL;
            
		    BolPtr = buffer;
		    stripnl(buffer);

		    while(1) {
			EolPtr = (char *) index(BolPtr, (int) '\r');
			EolPtr = (char *) index(BolPtr, (int) '\n');

			if (EolPtr == NULL) {
			    input_handle(BolPtr);
			    break;
			}

			*EolPtr = '\0';
			input_handle(BolPtr);
			EolPtr++;
			BolPtr = EolPtr;
		    }
		}
	    }
	}

	/* Now, check if we have any DCC data that we're waiting on.
	   If DCC data is waiting, it gets dispatched off to the DCC
	   routine to handle data accordingly. */
	for(x = 0; x < 3; x++) {
	    if (files[x].sock != -1) {
		if (FD_ISSET(files[x].sock, &activeset)) {
		    handle_dcc(x);
		}
	    }
	}

	/* If we've received the "OK" that we are now connected, it's time
	   to send data to the server and make ourselves registered.
	   This way, the server knows who we are and doesn't ping timeout. */

	if (is_connecting == 2) {
	    sprintf(reply, "USER %s %s %s :%s\r\nNICK %s\r\n", irc_data.username,
		hostname, hostname, irc_data.realname, irc_data.nick);
	    write(irc, reply, strlen(reply));
	    bzero(nick, 9);
	    strcpy(nick, irc_data.nick);
	    is_connecting = 3;
	    add_line(main_form->MainBrowser, "*** Non-blocking connection established.", 1);
	}

	/* Check and see whether or not we're still trying to connect.
	   Simply send a SPACE to the server we're trying to connect to. */

	if (is_connecting == 1) {
	    int result;

	    if ((result = write(irc, " ", 1)) < 0) {
		if (errno == EPIPE) {
		    add_line(main_form->MainBrowser, "*** Connection dropped: broken pipe or remote closed", 3);
		    add_line(main_form->MainBrowser, "*** You are no longer connecting to a server.", 3);

		    (void) shutdown(irc, 2);
		    close(irc);

		    FD_CLR(irc, &readset);
		    FD_CLR(irc, &activeset);

		    clear_vars();
		}

		if ((errno == EINPROGRESS) || (errno == EWOULDBLOCK) ||
		    (errno == 11))
		    is_connecting = 1;
		else
		    is_connecting = 2;
	    } else
		is_connecting = 2;
	}
    } while(1);
}
