/*
 * XIRC
 * Header file
 */

#include "forms.h"

#include <stdio.h>
#include <stdlib.h> 
#include <string.h>
#include <netdb.h>
#include <unistd.h>  
#include <syslog.h>  
#include <fcntl.h>
#include <pwd.h>
#include <ctype.h>
#include <sys/signal.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/wait.h>
#include <sys/errno.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <arpa/inet.h>
#include <netinet/in.h>

#include "InternalForms.h"
#include "Config.h"

typedef struct {
    char *status;
    void (*function)();
} NAMECODETABLE;

typedef struct {
    int status;
    void (*function)();
} NUMCODETABLE;

typedef struct {
    int style, color;
} FONTPAIR;

typedef struct {
    char email1[80];
    char email2[80];
    char nick[9];
} WHOLIST;

typedef struct {
    char nick[9];
    char username[80];
    char realname[80];
    char server[80];
    int fontsize;
    int bantype;
    int config1;
    FONTPAIR servermsg;
    FONTPAIR privatemsg;
    FONTPAIR critservmsg;
    FONTPAIR ctcpmsg;
} LOADDATATABLE;

typedef struct {
    char filename[600];
    int fd;
    int sock;
    int size;
    int progress;
    int type;
    int status;
    time_t start_time;
} DCC_FILES;

extern FD_AboutForm *about_form;
extern FD_MainForm *main_form;
extern FD_StartForm *start_form;
extern FD_FontForm *font_form;
extern FD_MSGForm *msg_form;
extern FD_MCForm *mc_form;
extern FD_TextForm *text_form;
extern FD_ULISTForm *ul_form;
extern FD_DCCForm *dcc_form;
extern FD_BanForm *ban_form;

extern int irc, fontsize, fontname, bantype, dsmode, bansel;
extern char channel[80], nick[80], server[80], reqnick[80];
extern fd_set readset, activeset;
extern NAMECODETABLE irc_names_parse[];
extern NUMCODETABLE irc_nums_parse[];
extern LOADDATATABLE irc_data;
extern WHOLIST chanusers[500];
extern DCC_FILES files[3];

#define	BAN_STANDARD		0
#define	BAN_SITE		1
#define	BAN_SCREW		2
#define	BAN_NICK		3

#define	CONFIG_LOCKNST		0x0001	/* Lock channel mode with +nst */
#define	CONFIG_AUTOGETDCC	0x0002	/* Auto-get DCC offered files */
#define	CONFIG_AUTOJOINONKICK	0x0004	/* Automatically join channel on kick */

#define	DCC_SEND		1	/* Normal 2K/block Send */
#define	DCC_GET			2	/* Normal 2K/block get */
#define	DCC_TGET		3	/* Turbo 64K/block get */
#define	DCC_TSEND		4	/* Turbo 64K/block send */
#define	DCC_CHAT		5	/* Chat */

#define	RPL_IRCINFO1		001
#define	RPL_IRCINFO2		002
#define	RPL_IRCINFO3		003
#define	RPL_IRCINFO4		004

#define	RPL_TRACELINK		200
#define	RPL_TRACECONNECTING	201
#define	RPL_TRACEHANDSHAKE	202
#define	RPL_TRACEUNKNOWN	203
#define	RPL_TRACEOPERATOR	204
#define	RPL_TRACEUSER		205
#define	RPL_TRACESERVER		206
#define	RPL_TRACENEWTYPE	208
#define	RPL_TRACECLASS		209	/* Unused */
#define	RPL_STATSLINKINFO	211
#define	RPL_STATSCOMMANDS	212
#define	RPL_STATSCLINE		213
#define	RPL_STATSNLINE		214
#define	RPL_STATSILINE		215
#define	RPL_STATSKLINE		216
#define	RPL_STATSQLINE		217	/* Unused */
#define	RPL_STATSYLINE		218
#define	RPL_ENDOFSTATS		219
#define	RPL_UMODEIS		221
#define	RPL_SERVICEINFO		231	/* Unused */
#define	RPL_ENDOFSERVICES	232	/* Unused */
#define	RPL_SERVICE		233	/* Unused */
#define	RPL_SERVLIST		234	/* Unused */
#define	RPL_SERVLISTEND		235	/* Unused */
#define	RPL_STATSLLINE		241
#define	RPL_STATSUPTIME		242
#define	RPL_STATSOLINE		243
#define	RPL_STATSHLINE		244
#define	RPL_LUSERCLIENT		251
#define	RPL_LUSEROP		252
#define	RPL_LUSERUNKNOWN	253
#define	RPL_LUSERCHANNELS	254
#define	RPL_LUSERME		255
#define	RPL_ADMINME		256
#define	RPL_ADMINLOC1		257
#define	RPL_ADMINLOC2		258
#define	RPL_ADMINEMAIL		259
#define	RPL_TRACELOG		261

#define	RPL_NONE		300
#define	RPL_AWAY		301
#define	RPL_USERHOST		302
#define	RPL_ISON		303
#define	RPL_UNAWAY		305
#define	RPL_NOWAWAY		306
#define	RPL_WHOISUSER		311
#define	RPL_WHOISSERVER		312
#define	RPL_WHOISOPERATOR	313
#define	RPL_WHOWASUSER		314
#define	RPL_ENDOFWHO		315
#define	RPL_WHOISCHANOP		316
#define	RPL_WHOISIDLE		317
#define	RPL_ENDOFWHOIS		318
#define	RPL_WHOISCHANNELS	319
#define	RPL_LISTSTART		321
#define	RPL_LIST		322
#define	RPL_LISTEND		323
#define	RPL_CHANNELMODEIS	324
#define	RPL_NOTOPIC		331
#define	RPL_TOPIC		332
#define	RPL_TOPICSETBY		333	/* Modified - not in RFC */
#define	RPL_INVITING		341
#define	RPL_SUMMONING		342
#define	RPL_VERSION		351
#define	RPL_WHOREPLY		352
#define	RPL_NAMREPLY		353
#define	RPL_KILLDONE		361	/* Unused */
#define	RPL_CLOSING		362	/* Unused */
#define	RPL_CLOSEEND		363	/* Unused */
#define	RPL_LINKS		364
#define	RPL_ENDOFLINKS		365
#define	RPL_ENDOFNAMES		366
#define	RPL_BANLIST		367
#define	RPL_ENDOFBANLIST	368
#define	RPL_ENDOFWHOWAS		369
#define	RPL_INFO		371
#define	RPL_MOTD		372
#define	RPL_INFOSTART		373
#define	RPL_ENDOFINFO		374
#define	RPL_MOTDSTART		375
#define	RPL_ENDOFMOTD		376
#define	RPL_YOUREOPER		381
#define	RPL_REHASHING		382
#define	RPL_MYPORTIS		384
#define	RPL_TIME		391
#define	RPL_USERSSTART		392
#define	RPL_USERS		393
#define	RPL_ENDOFUSERS		394
#define	RPL_NOUSERS		395

#define	ERR_NOSUCHNICK		401
#define	ERR_NOSUCHSERVER	402
#define	ERR_NOSUCHCHANNEL	403
#define	ERR_CANNOTSENDTOCHAN	404
#define	ERR_TOOMANYCHANNELS	405
#define	ERR_WASNOSUCHNICK	406
#define	ERR_TOOMANYTARGETS	407
#define	ERR_NOORIGIN		409
#define	ERR_NORECIPIENT		411
#define	ERR_NOTEXTTOSEND	412
#define	ERR_NOTOPLEVEL		413
#define	ERR_WILDTOPLEVEL	414
#define	ERR_UNKNOWNCOMMAND	421
#define	ERR_NOMOTD		422
#define	ERR_NOADMININFO		423
#define	ERR_FILEERROR		424
#define	ERR_NONICKNAMEGIVEN	431
#define	ERR_ERRONEUSNICKNAME	432
#define	ERR_NICKNAMEINUSE	433
#define	ERR_NICKCOLLISION	436
#define	ERR_USERNOTINCHANNEL	441
#define	ERR_NOTONCHANNEL	442
#define	ERR_USERONCHANNEL	443
#define	ERR_NOLOGIN		444
#define	ERR_SUMMONDISABLED	445
#define	ERR_USERSDISABLED	446
#define	ERR_NOTREGISTERED	451
#define	ERR_NEEDMOREPARAMS	461
#define	ERR_ALREADYREGISTERED	462
#define	ERR_NOPERMFORHOST	463
#define	ERR_PASSWDMISMATCH	464
#define	ERR_YOUREBANNEDCREEP	465
#define	ERR_YOUWILLBEBANNED	466	/* Unused */
#define	ERR_KEYSET		467
#define	ERR_CHANNELISFULL	471
#define	ERR_UNKNOWNMODE		472
#define	ERR_INVITEONLYCHAN	473
#define	ERR_BANNEDFROMCHAN	474
#define	ERR_BADCHANNELKEY	475
#define	ERR_BADCHANMASK		476	/* Unused */
#define	ERR_NOPRIVILEGES	481
#define	ERR_CHANOPRIVSNEEDED	482
#define	ERR_CANTKILLSERVER	483
#define	ERR_NOOPERHOST		491
#define	ERR_NOSERVICEHOST	492	/* Unused */
#define	ERR_UMODEUNKNOWNFLAG	501
#define	ERR_USERSDONTMATCH	502

/* Prototypes */
void give_help(void);
void load_options(void);
void handle_input(char *);
void save_options(void);
void set_options(void);
void init_dcc(void);
void add_names_list(char *);
void add_names(char *);
void remove_names(char *);
void names_refresh(void);
void add_line(FL_OBJECT *, char *, int);
int get_style_by_name(char *);
int get_color_by_name(char *);
void clear_who(void);
void change_who(char *, char *);
void add_who(char *, char *);
void remove_who(char *);
void handle_dcc(int);
void add_dcc(int, char *, char *, char *, int, int);
void set_ban_form(void);
