/* mesh.h -- mesh handling routines header
**
** Written and Copyright (C) 1994-1996 by Michael J. Gourlay
**
** NO WARRANTEES, EXPRESS OR IMPLIED.
*/

#ifndef _MESH_H__INCLUDED_
#define _MESH_H__INCLUDED_

typedef struct {
  long nx;    /* number of mesh points in the x-direction */
  long ny;    /* number of mesh points in the y-direction */
  double *x;  /* 2D array of x-values */
  double *y;  /* 2D array of y-values */
} MeshT;



/* MP_PICK_DIST: farthest you may be from a Meshpoint to pick it */
#define MP_PICK_DIST 15




void meshFree(MeshT *meshP, char *proc);

int meshAlloc(MeshT *meshP, const char *proc);

void meshInitialize(MeshT *meshP, int nx, int ny);

void meshPrint(const MeshT *meshP);

int meshCheckCompatibility(const MeshT *mesh1, const MeshT *mesh2);

void meshInterpolate(MeshT *moP, const MeshT *m1P, const MeshT *m2P, double tween_param);

void meshReset(MeshT *mP, const int img_width, const int img_height);

int meshFunctionalize(MeshT *mP, int mw, int mh);

int meshPointNearest(const MeshT *mP, int px, int py, int *mi, int *mj, int *dx, int *dy);

int meshAddLine(MeshT *mP, const int mi, const double mt, const int type);

int meshDeleteLine(MeshT *mP, int mi, int type);

int meshRead(MeshT *meshP, const char *filename);

int meshWrite(const MeshT *meshP, char *filename);

void meshMatch(MeshT *this, const MeshT *other);




#endif /* _MESH_H__INCLUDED_ */
