/*  xxDrawable  Abstract class for all tide-predicting windows.
    Last modified 1998-04-05

    Copyright (C) 1998  David Flater.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

class xxDrawable: public xxWindow {
  friend void xxDrawableMarkCallback (PredictionValue *marklevel,
    void *in_ptr);
  friend void xxDrawableTimestampCallback (Timestamp t,
    void *in_ptr);
  friend void xxDrawableAspectCallback (double aspect, void *in_ptr);
  friend void xxDrawableGraphCallback (Widget w, XtPointer client_data, XtPointer call_data);
  friend void xxDrawableTextCallback (Widget w, XtPointer client_data, XtPointer call_data);
  friend void xxDrawableClockCallback (Widget w, XtPointer client_data, XtPointer call_data);
  friend void xxDrawableChooserCallback (Widget w, XtPointer client_data, XtPointer call_data);
  friend void xxDrawableControlPanelCallback (Widget w, XtPointer client_data, XtPointer call_data);

public:
  xxDrawable (xxTideContext *in_tidecontext, xxContext *in_xxcontext,
    Station *in_station, int needbox = 1);
  virtual ~xxDrawable();

  virtual void help() = 0;
  virtual void save() = 0;
  virtual void redraw() = 0;

  // Global_redraw signals a change in global parameters (like colors).
  virtual void global_redraw();

  // These are actually the same for all drawables.
  void mark();
  void timestamp();
  void aspect();

  // This is called by subclasses to hook up the dismiss button, the help
  // button, and the options menu.  Requires needbox = 1.
  void addNormalButtons();

  // Set this to NULL if station is "stolen" for another window.
  Station *station;

  virtual int is_graph();
  virtual int is_clock();

protected:

  // This is the "starting time" or "now" that is used to calibrate
  // any given drawable in time.  This needs to be here so that it
  // is possible for one drawable to create another one with the
  // same start time.
  Timestamp t;

  xxContext *savebutton, *markbutton, *helpbutton, *dismissbutton,
    *optionsbutton, *optionsmenu, *graphbutton, *textbutton,
    *aspectbutton, *clockbutton, *chooserbutton, *timestampbutton,
    *unitsbutton, *rootbutton;
};
